# PowerShell Windows ps1 Script to build and install the c/c++ core-API ComMoDE.dll library and its tests
# Use the cmake project directive first to create the sln/vcxproj Visual Studio projects for building, then use 
# msbuild.exe command to automatically build them (ComMoDe.sln ALL_BUILD -> library and its related test executables).
# msbuild is used also to provide installation of the library in the custom user specified prefix path (INSTALL.vcxproj).  
# Please note, a Visual Studio suite plus its c/c++ development tools is needed to run safely this script.
# As a side note, once the cmake created the *.sln/vcxproj, you can skip the msbuild part, and build/install directly the 
# project using the Visual Studio GUI (simply double clicking on the icon of the sln project)   

# arguments: 
# -G <cmake Visual Studio Generator string> (MANDATORY)
# -prefix <path to installation prefix> (OPTIONAL)
# -python <path to target python interpreter> (OPTIONAL)
# -enableDoc (OPTIONAL)

#BEWARE: python version supported by the current ComMoDE distribution must be 3.8 <= x <= 3.10

# the script will do the following: 
# 1) check python in use, if no python interpreter was provided use the "python.exe" default command of the powershell 
# 2) download and install boost (boost-python module only) for ComMoDE internal usage.
# 3) build ComMoDE.dll c++ core-API and its python binding PyComMoDE.pyd
# 4) Prepare ComMoDe python enviroment in <installation prefix>\bin folder alongside the other 
#    python packages required: pip, venv, setuptools, wheel and PySide2
# 5) install ComMoDe.dll in bin folder alongside its c++ native executables
# 6) install PyComMoDe inside the ComMoDE python environment in <installation prefix>/bin 
# 7) run ComMode functional tests   
#

$totargs = $args.Count

$posVSG = -1
$posPrefix = -1
$posPython = -1
$posBuildDoc = -1

for ($index_ = 0; $index_ -lt $args.Count; $index_++) { 
    if ( $args[$index_] -eq "-prefix" ) {
        $posPrefix = $index_
    }
    elseif ($args[$index_] -eq "-python" ) {
        $posPython = $index_
    }
    elseif ($args[$index_] -eq "-G" ) { 
        $posVSG = $index_
    }
    elseif ($args[$index_] -eq "-enableDoc" ) { 
        $posBuildDoc = $index_
    }

}
$posArgPrefix = $posPrefix + 1
$posArgVSG = $posVSG + 1
$posArgPython = $posPython + 1

if ( ( $posArgVSG -gt 0 ) -and ($posArgVSG -lt $totargs) -and ($posArgVSG -ne $posBuildDoc) -and ($posArgVSG -ne $posPrefix) -and ($posArgVSG -ne $posPython)) {
    $CMAKE_VS_GEN = -join ("-G ", '"', ($args[$posArgVSG]), '"')
}
else {
    $CMAKE_VS_GEN = -join ("-G ", '"Visual Studio 17 2022"')
    Write-Host "Cannot retrieve valid Generator string: using a default CMake generator for Visual Studio $CMAKE_VS_GEN !"
    Write-Host "Beware cmake generator must match the Visual Studio version actually in use!"
}

if ( ( $posArgPrefix -gt 0 ) -and ($posArgPrefix -lt $totargs) -and ($posArgPrefix -ne $posBuildDoc) -and ($posArgPrefix -ne $posVSG) -and ($posArgPrefix -ne $posPython)) {
    $INPUT_PREFIX = $args[$posArgPrefix]
}
else {
    $INPUT_PREFIX = -join ((Resolve-Path ~), "\ESA-ComMoDE_v1.0")
    Write-Host "Cannot retrieve a prefix path: using default installation prefix $INPUT_PREFIX !"
    Write-Host "You can always customize installation path adding to argument of the script -prefix <pathLocationToInstall>"
}   
Write-Host "Installing on prefix path $INPUT_PREFIX"

if ( ( $posArgPython -gt 0 ) -and ($posArgPython -lt $totargs) -and ($posArgPython -ne $posPrefix) -and ($posArgPython -ne $posVSG) -and ($posArgPython -ne $posBuildDoc)) {
    $PYTHON_COMMAND = $args[$posArgPython]
    Write-Host "User specified python interpreter : " $PYTHON_COMMAND
}
else {
    Write-Host "Error : the full path to valid python interpreter command must be provided.
    For example, for a python 3.10.x : 
    .\build.ps1 ... -python C:\Users\johndoe\AppData\Local\Programs\Python\Python310\python.exe ..."
    exit(1)}   

if ( $posBuildDoc -gt -1 ) {
    $ENABLE_DOXY = "ON"
}
else {
    $ENABLE_DOXY = "OFF"
}   

#verify its admissible python
$expr = -join ($PYTHON_COMMAND, " --version")
$version = (Invoke-Expression $expr)
if ( $version -like "*Python*") {
    $numstringarray = ([regex]::Matches($version, '[0-9]+').value).split(" ")
    $numstring = [float]::Parse($numstringarray[1])
}
else { 
    Write-Host "Error $PYTHON_COMMAND is not recognized as a valid python interpreter"
    exit(1)
}

if ($numstring -gt 10 -or $numstring -lt 8) {
    Write-Host "Error : $PYTHON_COMMAND command does not match any of 
    recommendend version >= 3.8.x and <= 3.10.x. 
    If you have a suitable version of python installed please re-run this script 
    providing the full path to python.exe interpreter command as argument. 
    For example, for a python 3.10.x : 
    .\build.ps1 ... -python C:\Users\johndoe\AppData\Local\Programs\Python\Python310\python.exe ..."
    exit(1)
}

$CURRENT_DIR = $pwd
$BUILD_DIR = -join ($CURRENT_DIR, "\build")

##creating the build directory and access it
if ( Test-Path -Path $BUILD_DIR ) {
    Write-Host "Building locally in already existent build dir : $BUILD_DIR"
}
else { 
    (New-Item -ItemType "directory" -Path $BUILD_DIR)  
    Write-Host "Building locally in newly created build dir : $BUILD_DIR"
}

Set-Location $BUILD_DIR

$scriptboost = -join ($CURRENT_DIR, "\external\buildBoost.ps1")
$boostinstall = -join ($BUILD_DIR, "\boost-ESAComMoDE")

$rebuild = $true

if (Test-Path -Path $boostinstall ) {
    $libbostdir = -join ($boostinstall, "\lib")
    Set-Location $libbostdir
    $sofiles = ( Get-ChildItem *boost_python*.dll*).name
    $nfiles = $sofiles.Count
    if ( $nfiles -gt 0 ) { 
        $numsoarray = ([regex]::Matches($sofiles[0], '[0-9]+').value).split(" ")
        $soboostver=$numsoarray.substring(1) #remove the first 3   
        if ( $soboostver -eq $numstring ) { 
            Write-Host ""
            Write-Host ""
            Write-Host "*********************************************************"
            Write-Host "Boost 1.78 already installed in $BUILD_DIR"
            Write-Host "*********************************************************"
            $rebuild=$false
        }
    }
    Set-Location $BUILD_DIR
}

if ($true -eq $rebuild) { 
    Write-Host ""
    Write-Host ""
    Write-Host "*********************************************************"
    Write-Host "INSTALLING Boost 1.78 in $BUILD_DIR"
    Write-Host "redirecting output to $BUILD_DIR/boost_install.log"
    Write-Host "*********************************************************"
    $scriptboost = -join ($scriptboost, " ", $PYTHON_COMMAND)
    $scriptboost = -join ($scriptboost, " > boost_install.log 2>&1")
    Invoke-Expression($scriptboost)
    
    if (1 -eq $LASTEXITCODE ){
        Write-Host "boost installation failed. See details in $BUILD_DIR/boost_install.log" 
        exit(1)
    }elseif (2 -eq $LASTEXITCODE ) {
        Write-Host "Error while downloading boost_1_78_0.zip with Web-Request: the archive may be corrupted."
        Write-Host "Please download it manually from https://boostorg.jfrog.io/artifactory/main/release/1.78.0/source ."
        Write-Host "Then, copy it into $BUILD_DIR and try again to run the build script" 
        exit(1)
    }
}

if (Test-Path -Path $boostinstall ) {
    $libbostdir = -join ($boostinstall, "\lib")
    Set-Location $libbostdir
    $sofiles = ( Get-ChildItem *boost_python*.dll*).name
    $nfiles = $sofiles.Count
    if ( $nfiles -eq 0 ) { 
        Write-Host "Error building boost_1_78_0: no suitable boost python dlls were created."
        Write-Host "A very common cause is having the current $PYTHON_COMMAND exported into the Windows Path environment variable. Please erase python paths from Path and try again."
        Write-Host "If it's not the case, please check the boost_install.log file for errors." 
        exit(1)
    }
    Set-Location $BUILD_DIR
}

$INPUT_BOOST = -join ($BUILD_DIR, "\boost-ESAComMoDE")

#configure the cmake project
$cmake_command = -join ("cmake.exe ", ($CMAKE_VS_GEN))
$cmake_command = -join (($cmake_command), " -DCMAKE_INSTALL_PREFIX=", ($INPUT_PREFIX))
$cmake_command = -join (($cmake_command), " -DCMAKE_BUILD_TYPE=Release")
$cmake_command = -join (($cmake_command), " -DENABLE_CORE_FUNCTIONAL_TEST=ON")
$cmake_command = -join (($cmake_command), " -DENABLE_INTERFACE=ON")
$cmake_command = -join (($cmake_command), " -DVERBOSE_MAKE=OFF")
$cmake_command = -join (($cmake_command), " -DBUILD_DOCUMENTATION=", ($ENABLE_DOXY))
$cmake_command = -join (($cmake_command), " -DBoost_DIR=", ($INPUT_BOOST), "\lib\cmake\Boost-1.78.0")
$cmake_command = -join (($cmake_command), " ..")

Write-Host ""
Write-Host ""
Write-Host "*********************************************************"
Write-Host "CONFIGURE CMAKE PROJECT       ***************************"
Write-Host "*********************************************************"

Invoke-Expression $cmake_command

#build the coreAPI
$build_command = "MSBuild.exe ComMoDE.sln /t:ALL_BUILD /p:Configuration=Release > coreAPI-build.log"

Write-Host ""
Write-Host ""
Write-Host "*********************************************************"
Write-Host "BUILDING CORE_API "
Write-Host "redirecting output to $BUILD_DIR/coreAPI-build.log"
Write-Host "*********************************************************"
Invoke-Expression $build_command 

$BINFOLDER = -join ($INPUT_PREFIX, "\bin")

#install core-API, if successfull it will regenerate the BINFOLDER
Write-Host ""
Write-Host ""
Write-Host "*********************************************************"
Write-Host "INSTALL CORE_API   **************************************"
Write-Host "redirecting output to $BUILD_DIR/coreAPI-install.log"

Write-Host "*********************************************************"
$install_command = "MSBuild.exe INSTALL.vcxproj /p:Configuration=Release > coreAPI-install.log"
Invoke-Expression $install_command 

$LIBFILE = -join($BINFOLDER, "\ComMoDE.dll")
if  (-Not(Test-Path -Path $LIBFILE)){
    Write-Host "Error: coreAPI building/installation failed."
    exit(1)
}

#install wheel for python bindings
Write-Host ""
Write-Host ""
Write-Host "*********************************************************"
Write-Host "INSTALL PYTHON BINDED CORE_API   ************************"
Write-Host "*********************************************************"
Write-Host ""
Write-Host ""


Write-Host "prepare the PyComMoDE python environment"
Set-Location $BINFOLDER

$ENVPYTHONVERSION=""
$ENVACTIVATEFILE =".\ComMoDE-env\Scripts\activate"
if (Test-Path -Path $ENVACTIVATEFILE) {
    $activate = ".\ComMoDE-env\Scripts\activate"
    Invoke-Expression $activate

    $expr = -join ("python", " --version")
    $ENVPYTHONVERSION = (Invoke-Expression $expr)

    $deactivate = "deactivate"
    Invoke-Expression $deactivate
}

if( $version -ne $ENVPYTHONVERSION) {
    Write-Host "re-creating ComMoDE environment from scratch"
    if ((Test-Path -Path "./ComMoDE-env") ) {
        Remove-Item "./ComMoDE-env" -Recurse
    }

    $command = -join ($PYTHON_COMMAND, ' -m venv ./ComMoDE-env')  
    Invoke-Expression $command
} else {
    Write-Host "found already suitable ComMoDE environment"
}

$activateEnvCommand = ".\ComMoDE-env\Scripts\activate"
Invoke-Expression $activateEnvCommand 

#installing required packages
Write-Host "installing required python packages: setuptools wheel numpy matplolib PySide2 pandas"
$install_prereq_command = "pip install setuptools wheel numpy matplotlib PySide2 pandas"
Invoke-Expression $install_prereq_command

Write-Host "prepare the wheel folder"
$WHEEL_FOLDER = -join ($BINFOLDER, "\py-wheel")
Set-Location $WHEEL_FOLDER
$pycommodeFolder = -join (".\", "PyComMoDE")
if (-Not(Test-Path -Path $pycommodeFolder)) {
    New-Item -ItemType "directory" -Path $pycommodeFolder 
}
Copy-Item $BINFOLDER\*.dll -Destination $pycommodeFolder 
Move-Item $BINFOLDER\*.pyd -Destination $pycommodeFolder -force
Move-Item __init__.py -Destination $pycommodeFolder -force
#hard copy dll from boost
$DLLBOOSTPATH = -join ($INPUT_BOOST, "\lib")
Copy-Item $DLLBOOSTPATH\*.dll -Destination $pycommodeFolder 
Copy-Item $DLLBOOSTPATH\*.dll -Destination $BINFOLDER
 
Write-Host "installing the ComMoDE wheel"
$wheel_setup_command = "python setup.py bdist_wheel"
Invoke-Expression $wheel_setup_command

$WHEELDISTFILE = -join (".\dist", "\PyComMoDE-1.0.0-py3-none-any.whl")
if (Test-Path -Path $WHEELDISTFILE) {
    $wheel_install_command = -join ("pip install --force ", $WHEELDISTFILE)
    Invoke-Expression $wheel_install_command
}
else {
    Write-Host "cannot install PyComMoDE python wheel!"
    exit(1)
}

#deactivate the environment
$deactivateEnvCommand = "deactivate"

Invoke-Expression $deactivateEnvCommand

#run core-API functional tests
Write-Host ""
Write-Host ""
Write-Host "*********************************************************"
Write-Host "RUN CORE_API FUNCTIONAL TESTS   *************************"
Write-Host "*********************************************************"

Set-Location $BUILD_DIR 
Set-Location core-API
#BEWARE - hard coded for Release only. If you change build type you need to change folder
#basically i'm copying the ComMoDE dll into the functional test folder.
$copydll_command = "cp ./src/Release/*dll ./functional_test/Release"
Invoke-Expression $copydll_command 
# here i'm running the pool of tests.
$ctest_command = "ctest.exe -C Release"
Invoke-Expression $ctest_command 


#return to the mother directory
set-Location $CURRENT_DIR

#run interface functional tests
Write-Host ""
Write-Host ""
Write-Host "*********************************************************"
Write-Host "RUN INTERFACE FUNCTIONAL TESTS   *************************"
Write-Host "*********************************************************"

#activate the environment
Set-Location $BINFOLDER
$activateEnvCommand = ".\ComMoDE-env\Scripts\activate"
Invoke-Expression $activateEnvCommand 

#move to functional test folder
$TESTFOLDER = -join($INPUT_PREFIX, "\gui\functional_test")
Set-Location $TESTFOLDER

Write-Host "Run test 001: ------------------------>"
$command = "python interface_functional_test_001.py"
Invoke-Expression $command
Write-Host ""

Write-Host "Run test 002: ------------------------>"
$command = "python interface_functional_test_002.py"
Invoke-Expression $command
Write-Host ""

Write-Host "Run test 003: ------------------------>"
$command = "python interface_functional_test_003.py"
Invoke-Expression $command

# $deactivateEnvCommand = "deactivate"
Invoke-Expression $deactivateEnvCommand

#return to the mother directory
Set-Location $CURRENT_DIR
