#!/bin/bash

# Linux bash sh script to build and install the c/c++ core-API libComMoDE.so library and its tests
# Use the cmake project directive first to configure the building, then use make and make install
# commands to build the code and install it in a custom user specified prefix path. 

# arguments (all) optional: 
# -prefix <path to installation prefix>
# -python <path to target python interpreter>
# -enableDoc

#BEWARE: python version supported by the current ComMoDE distribution must be 3.8 <= x <=3.10

# the script will do the following: 
# 1) check python in use, if no python interpreter was provided use the "python3" default command of the shell 
# 2) download and install boost (boost-python module only) for ComMoDE internal usage.
# 3) build libComMoDE.so c++ core-API and its python binding PyComMoDE.so
# 4) Prepare ComMoDe python enviroment in <installation prefix>/bin folder alongside the other 
#    python packages required: pip, venv, setuptools, wheel and PySide2
# 5) install libComMoDe.so in bin folder alongside its c++ native executables
# 6) install PyComMoDe inside the ComMoDE python environment in <installation prefix>/bin 
# 7) run ComMode functional tests   
#

declare -a v=("$@")
totargs=$#
posPrefix=-1
posPython=-1
posBuildDoc=-1
index_=0

for var in ${v[@]} 
do
    if [ "$var" == "-prefix" ]; then
        posPrefix=$index_
    fi 
    if [ "$var" == "-python" ]; then
        posPython=$index_
    fi 
    if [ "$var" == "-enableDoc" ]; then
        posBuildDoc=$index_
    fi 
    index_=$((index_+1))
done 

posArgPrefix=$((posPrefix+1))
posArgPython=$((posPython+1))

if [[ $posArgPrefix > 0 && $posArgPrefix != totargs && $posArgPrefix != $posPython && $posArgPrefix != $posBuildDoc ]]; then
    INPUT_PREFIX=${v[$posArgPrefix]}
else
    INPUT_PREFIX="$HOME/ESA-ComMoDE_v1.0"
    echo "Using default installation prefix $INPUT_PREFIX !"
    echo "You can always customize installation path adding to argument of the script -prefix <pathLocationToInstall>"
fi
echo "Installing on prefix path $INPUT_PREFIX"

if [[ $posArgPython > 0 && $posArgPython != totargs && $posArgPython != $posPrefix && $posArgPython != $posBuildDoc ]]; then
    PYTHON_COMMAND=${v[$posArgPython]}
else
    PYTHON_COMMAND="python3"
    echo "Using default shell python3 interpreter"
    echo "You can always pick a custom installed python interpreter by -python <fullPathToCustomPython>"
fi
echo "Using python : $PYTHON_COMMAND"

if [[ $posBuildDoc == -1 ]]; then
    ENABLE_DOXY="OFF"
else
    ENABLE_DOXY="ON"
fi

#verify it's an admissible python
version=$($PYTHON_COMMAND --version)
if [[ $version == *"Python"* ]]; then
    numstringarray=( $(echo ${version} | grep -o -E "[0-9]+") )
    numstring=${numstringarray[1]}
    PYMAJOR=${numstringarray[0]}
    PYMINOR=${numstringarray[1]}
else 
    echo "Error "$PYTHON_COMMAND" is not recognized as a valid python interpreter"
    exit 1 
fi

if [[ $numstring -gt "10" || $numstring -lt "8" ]]; then 
    echo "Error : $PYTHON_COMMAND command does not match any of 
    recommended version >= 3.8.x and <= 3.10.x. 
    If you have a suitable version of python installed please re-run this script 
    providing the full path to a python3 interpreter command as argument. 
    For example, for a python 3.8.x : 
    sh build.sh -python /usr/bin/python3.8 -prefix ..."
    exit 1
fi

CURRENT_DIR="$(pwd)"
BUILD_DIR="$CURRENT_DIR/build"

##creating the build directory and access it
if [ -d "$BUILD_DIR" ] ; then
    echo "Building locally in already existent build dir : $BUILD_DIR"
else 
    mkdir $BUILD_DIR 
    echo "Building locally in newly created build dir : $BUILD_DIR"
fi

cd $BUILD_DIR

scriptboost="$CURRENT_DIR/external/buildBoost.sh"
chmod a+x $scriptboost
boostinstall="$BUILD_DIR/boost-ESAComMoDE"
rebuild=true

if [ -d "$boostinstall" ]; then
    cd $boostinstall/lib
    sofiles=( $(find . -type f -name "*boost_python*.so*") )
    nfiles=${#sofiles[@]}
    if [[ $nfiles -gt "0" ]]; then 
        numsoarray=( $(echo ${sofiles[0]} | grep -o -E "[0-9]+") )
        temp=${numsoarray[0]}
        soboostver=${temp:1}   
        if [[ $soboostver == $numstring ]]; then 
            echo ""
            echo ""
            echo "*********************************************************"
            echo "Boost 1.78 already installed in $BUILD_DIR"
            echo "*********************************************************"
            rebuild=false
        fi
    fi
    cd $BUILD_DIR
fi

if [[ "$rebuild" == true ]]; then 
    echo ""
    echo ""
    echo "*********************************************************"
    echo "INSTALLING Boost 1.78 in $BUILD_DIR"
    echo "redirecting output to $BUILD_DIR/boost_install.log"
    echo "*********************************************************"
    $scriptboost $PYTHON_COMMAND > boost_install.log 2>&1 

    status=$?
    if [ $status -eq 1 ]; then 
        echo "boost installation failed. See details in $BUILD_DIR/boost_install.log" 
        exit 1
    elif [ $status -eq 2 ]; then 
        echo "Error while downloading boost_1_78_0.tar.gz with wget: the archive may be corrupted."
        echo "Please download it manually from https://boostorg.jfrog.io/artifactory/main/release/1.78.0/source"
        echo "Then, copy it into $BUILD_DIR and try again to run the build script" 
        exit 1
    fi
fi

#configure the cmake project
echo ""
echo ""
echo "*********************************************************"
echo "CONFIGURE CMAKE PROJECT"
echo "*********************************************************"
command_cmake="cmake -DCMAKE_BUILD_TYPE=Release -DCMAKE_INSTALL_PREFIX=$INPUT_PREFIX -DENABLE_CORE_FUNCTIONAL_TEST=ON -DBoost_DIR=$boostinstall/lib/cmake/Boost-1.78.0 -DENABLE_INTERFACE=ON -DVERBOSE_MAKE=OFF -DBUILD_DOCUMENTATION=$ENABLE_DOXY ../"
$command_cmake

#run make 
echo ""
echo ""
echo "*********************************************************"
echo "BUILDING CORE_API"
echo "*********************************************************"
make 

BINFOLDER=$INPUT_PREFIX/bin
LIBFOLDER=$INPUT_PREFIX/lib

#run make install, if successfull it will regenerate the BINFOLDER
echo ""
echo ""
echo "*********************************************************"
echo "INSTALL CORE_API"   
echo "redirecting output to $BUILD_DIR/coreAPI-install.log"
echo "*********************************************************"
make install > coreAPI-install.log

### CHECK INSTALLATION OF CORE API.
if [ ! -f "$LIBFOLDER/libComMoDE.so" ] ; then
    echo "Error: coreAPI building/installation failed."
    exit 1
fi

#install wheel for python bindings
echo ""
echo ""
echo "*********************************************************"
echo "INSTALL PYTHON BINDED CORE_API "
echo "*********************************************************"
echo ""
echo ""

#copy boost libraries in LIBFOLDER/boost
DLLBOOSTPATH="$BUILD_DIR/boost-ESAComMoDE/lib"
if [ ! -d "$LIBFOLDER/boost" ] ; then
    mkdir $LIBFOLDER/boost
fi
cp $DLLBOOSTPATH/*.so* $LIBFOLDER/boost 

echo "prepare the PyComMoDE python environment"
cd $BINFOLDER

ENVPYTHONVERSION=""
if [[ -f "./ComMoDE-env/bin/activate" ]]; then 
    source ./ComMoDE-env/bin/activate
    ENVPYTHONVERSION=$(python --version)
    deactivate
fi

if [[ "$version" != "$ENVPYTHONVERSION" ]]; then    
    echo "re-creating ComMoDE environment from scratch"
    if [[ -d ./ComMoDE-env ]]; then
        rm -r ./ComMoDE-env
    fi
    $PYTHON_COMMAND -m venv ./ComMoDE-env
    #alter the activate source in case Qt5 is already in the linux system. In that case it will hide 
    # the Qt suite binded to the PySide2 package in the python environment
    #added also the manually built boost to LD_LIBRARY_PATH

    RESTOREKEY='    # reset old environment variables'
    RESTORELDP='    # reset old environment variables'"\n"
    RESTORELDP+='    if [ -n "${_OLD_VIRTUAL_LDLIBRARYPATH:-}" ] ; then'"\n"
    RESTORELDP+='        LD_LIBRARY_PATH=\"${_OLD_VIRTUAL_LDLIBRARYPATH:-}"'"\n"
    RESTORELDP+='        export LD_LIBRARY_PATH'"\n"
    RESTORELDP+='        unset _OLD_VIRTUAL_LD_LIBRARY_PATH'"\n"
    RESTORELDP+='    fi'"\n"


    ALTERKEY='_OLD_VIRTUAL_PATH="$PATH"'

    ALTERLDP='_OLD_VIRTUAL_LDLIBRARYPATH="$LD_LIBRARY_PATH"'"\n"
    ALTERLDP+='LD_LIBRARY_PATH="$VIRTUAL_ENV/lib64/python'"$PYMAJOR.$PYMINOR"
    ALTERLDP+='/site-packages/PySide2/Qt/lib:'$LIBFOLDER'/boost:$LD_LIBRARY_PATH"'"\n"
    ALTERLDP+='export LD_LIBRARY_PATH'"\n"
    ALTERLDP+='_OLD_VIRTUAL_PATH="$PATH"'

    sed -i 's|'"$RESTOREKEY"'|'"$RESTORELDP"'|g' ./ComMoDE-env/bin/activate
    sed -i 's|'"$ALTERKEY"'|'"$ALTERLDP"'|g' ./ComMoDE-env/bin/activate

else 
    echo "found already suitable ComMoDE environment"
fi

source ComMoDE-env/bin/activate

#installing required packages
echo "installing required python packages: setuptools wheel numpy matplolib PySide2 pandas"
pip install setuptools wheel numpy matplotlib
pip install PySide2
pip install pandas

echo "prepare the wheel folder"
cd $BINFOLDER/py-wheel
if [[ -d "./PyComMoDE" ]]; then
    rm -r ./PyComMoDE
fi

cp -r $LIBFOLDER ./PyComMoDE
mv __init__.py PyComMoDE

echo "installing the ComMoDE wheel"
python setup.py bdist_wheel
if [[ -f "./dist/PyComMoDE-1.0.0-py3-none-any.whl" ]]; then
    pip install --force ./dist/PyComMoDE-1.0.0-py3-none-any.whl
else 
    echo "Error : cannot install PyComMoDe python wheel"
    exit 1
fi

deactivate

#run core-API functional tests
echo ""
echo ""
echo "*********************************************************"
echo "RUN CORE_API FUNCTIONAL TESTS   *************************"
echo "*********************************************************"
cd "$BUILD_DIR/core-API"
ctest

#return to the mother directory
cd $CURRENT_DIR

#run interface functional tests
echo ""
echo ""
echo "*********************************************************"
echo "RUN INTERFACE FUNCTIONAL TESTS   *************************"
echo "*********************************************************"

#activate the environment
source $BINFOLDER/ComMoDE-env/bin/activate

#move to functional test folder
cd $INPUT_PREFIX/gui/functional_test

echo "Run test 001: ------------------------>"
python interface_functional_test_001.py
echo ""

echo "Run test 002: ------------------------>"
python interface_functional_test_002.py
echo ""

echo "Run test 003: ------------------------>"
python interface_functional_test_003.py

deactivate

#return to the mother directory
cd $CURRENT_DIR
