#include "run_models.h"
#include "utils.h"
#include "utils_fountain.h"
#include <chrono>
#include <fstream>
#include <iostream>
#include <random>
#include <tuple>


using namespace std;

int main() {

  // inputs - in general, if not stated otherwise, all inputs should be written
  // to a file

  /*
   * default_size_switch = 0 -> default values are used
   * default_size_switch = 1 -> values are user inputs
   * size limits are the smallest and largest possible exponents of the particle
   * size bins lower_size_limit specifies the largest particle cut-off
   * higher_size_limit specifies the smallest particle cut-off
   */
  int default_size_switch = 1;
  int number_of_bins = 10;
  double largest_particle_size = pow(10.0, -3);
  double smallest_particle_size = pow(10.0, -7);
  double power_law_index = 3.1;

  /*
   * exponents calculated from input size cut-offs, which are passed to the
   * model
   */
  double lower_size_limit = log10(largest_particle_size);
  double higher_size_limit = log10(smallest_particle_size);
  double power_law_index_passed = -power_law_index;

  // rotation_switch = 0 means no rotation -> default parameters amplitude_rot =
  // 0 and period_rot = 0 are passed rotation_switch = 1 -> amplitude_rot
  // between 0 and 1 and period_rot (passed on in seconds) are user inputs
  int rotation_switch = 0;
  double amplitude_rot = 0;
  double period_rot_in_hours = 0;
  // conversion to seconds
  double period_rot = period_rot_in_hours * 3600.0;

  // qd_estimation_switch = 0 -> qd is user input
  // qd_estimation_switch = 1 -> afrho is user input (qd is estimated by model)
  int qd_estimation_switch = 0;
  double afrho_input = 25.0;
  double qd_input = 1000.0;

  // in [AU]
  double heliocentric_distance = 0.5;

  /* species index is used to decide which file is read in
   * 0 for water ice (n = 1.31 + 0i)
   * 1 for silicate (enstatite n = 1.63 + 4.4e-5 i)
   * 2 for astronomical silicate (n = 1.81 + 0.1012i)
   * 3 for Troilite (FeS n = 1.40 + 1.25i)
   */
  int species_index = 0;
  int phase_angle = 0.0;

  double Q_gas = 1000;
  double radius_N = 1000.0;
  double particle_density = 400.0;
  double nucleus_temp = 280.0;

  // 0 means no asymmetry -> fountain model
  // 1 means asymmetry -> numerical model
  int day_night_asym_switch = 1;

  // default values of angular_dependency_index and cosine_scaling are passed on
  // even if there is no angular dependence (or no cosine-law) inputs are
  // 0 - no dependency (only used for comparison with the fountain model - i.e.
  // verification, validation purposes) 1 - cosine-law (this also needs a
  // cosine_scaling that is between 0 and 1) 2 - gaussian 3 - half-maxwell
  int angular_dependency_index = 2;
  double cosine_scaling = 0;

  // input grid parameters (in [m])
  const int max_distance = 10000000;
  const int cellsize = 10000;

  // gridsize does not need to be written to file as this information is only a
  // calculation to get the final grid_sizes
  const int gridsize = round(double(max_distance) / double(cellsize));

  // needs to be discussed if gridsize_apex should be a user input - now it is
  // based on a good estimate of how far the coma extends for astro-silicate also
  // describes where the nucleus is located in the grid
  const int gridsize_apex = round(60000000.0 / double(cellsize));

  const int gridsize_x = gridsize + gridsize_apex;
  const int gridsize_y = gridsize;

  // this variable here is a placeholder to connect progress of the function to
  // an external progressBar in the interface. Here it does nothing for now.
  int counterOfProcessedBins = 0;

  core::run_model_all_sizes(
      gridsize_x, gridsize_y, cellsize, gridsize_apex, number_of_bins,
      particle_density, lower_size_limit, higher_size_limit,
      power_law_index_passed, Q_gas, radius_N, nucleus_temp, species_index,
      phase_angle, heliocentric_distance, day_night_asym_switch,
      angular_dependency_index, cosine_scaling, rotation_switch, amplitude_rot,
      period_rot, qd_estimation_switch, qd_input, afrho_input, ".",
      counterOfProcessedBins);

  // ORIGINAL ISSUE OF FREDRIK (github frejon) doesn't arrive to define any
  // trajectory.
  // the problem is in the simulation comet code hanging.

  //     /*
  //  * sc parameters
  //  * d_ca = distance at closest approach in [km]
  //  * phase_angle_of_approach [°]
  //  * theta_sc = Azimuth angle wrt projection of Sun-comet line in the target
  //  plane [°]
  //  * magnitude of velocity = speed of spacecraft [km/s]
  //  */
  //     double d_ca = 600.0;
  //     double phase_angle_of_approach = 45.0;
  //     double theta_sc = 10.0;
  //     double magnitude_of_velocity = 20.0;
  //     double spacecraft_frontal_area = 1000.0;

  //     core::run_trajectory_module(d_ca,phase_angle_of_approach,theta_sc,magnitude_of_velocity,spacecraft_frontal_area,
  //     ".", ".");

  return 0;
}
