//
// Created by nh18u590 on 29.06.2022.
// Modified by RArpa_OPT on 26.2.2023
//

#include "utils_fountain.h"
#include <iostream>
#include <tuple>
#include <math.h>
#include <fstream>
#include <random>
#include <string>
#include <vector>
#include "utils.h"

using namespace std;
namespace core{

//RA 1/8 - fine grain utilities composing to split down run numerical model - to be python binded
/*!
    Sampling with trajectories to evaluate the coma values before the smoothing steps
    \param[in] gridsize_x total number of elements in x for grid
    \param[in] gridsize_y total number of elements in y for grid
    \param[in] gridsize_apex total number of elements composing the apex in x direction
    \param[in] cellsize size of the grid cell
    \param[in] beta value
    \param[in] size_index positional index of the current bin size to be processed
    \param[in] particle_size size of the dust particle
    \param[in] v0 initial outflow velocity
    \param[in] particle_density density of the dust particle
    \param[in] size_dependent_qd prod rate depending from size
    \param[in] Q_sca from scattering properties calculation
    \param[in] phase_function from scattering properties calculation
    \param[in] amplitude_rot amplitude of rotation 
    \param[in] period_rot period of rotation 
    \param[in] angular_dependency_index positional index to identify the function used to model angular dependency
    \param[in] cosine_scaling scaling factor applied only in case of cosine angular dependency function
    \param[out] grid matrix of coma values evaluated. Directly resized inside the method
*/
void numerical_model_sampling(
    int gridsize_x, int gridsize_y, int gridsize_apex, int cellsize, 
    double beta, int size_index, double particle_size, double v0, 
    double particle_density, double size_dependent_qd, double Q_sca, double phase_function, 
    double amplitude_rot, double period_rot, int angular_dependency_index, double cosine_scaling,
    std::vector<std::vector<double>>& grid)
{


    int x = 0;
    int y = 0;
    double emission_angle = 1.0;
    double max_time = 1000.0;

    int gridsize = gridsize_x-gridsize_apex;

    double v = 0;
    double volume = 0.0;
    double scaling_rotational;
    double scaling_qd;

    double v_temp = 0.0;
    double time_step = 1.0;

    //particle mass of each size bin should probably also be written to file
    //potentially this could be calculated in a seperate function
    double particle_mass = 4./3.*pi*particle_size*particle_size*particle_size*particle_density;
    int count_steps = 0;

    //needs to be fine-tuned possibly in relation to the resolution
    int number_of_trajectories = 5000;

    double solid_angle_scaling = 1.0;
    double qd_distribution = 0.0;


    calc_endtime(v0,&max_time, beta, cellsize);

    scale_to_qd(&scaling_qd,size_dependent_qd,&particle_mass);

    for (int i = 0; i <= number_of_trajectories; i++){
        //monte_carlo(&emission_angle);
        get_emission_angle(&emission_angle,&solid_angle_scaling,i,number_of_trajectories);

        velocity_distribution(v0, &v,emission_angle,angular_dependency_index,cosine_scaling);
        get_qd_distribution(&qd_distribution, emission_angle,angular_dependency_index,cosine_scaling);

        count_steps = 0;

        for (double time = 0; time < max_time; time += time_step){

            trajectory(&x,&y,&time,&emission_angle,&beta,&v, &v_temp,cellsize);
            get_time_step(&time_step, & v_temp, cellsize);
            rotational_variablility(amplitude_rot,period_rot,&scaling_rotational, &time);

            count_steps++;

            if( x+gridsize_apex >= 0 && abs(x+gridsize_apex) < gridsize_x && y < gridsize_y){
                cell_volume(&volume, y, cellsize);
                //std::vector<double> &row = grid.at(x+gridsize_apex);
                grid[x+gridsize_apex][y] += (scaling_qd*time_step*scaling_rotational*solid_angle_scaling*qd_distribution/volume);

            }
            else if(x+gridsize_apex > gridsize_x || y > gridsize_y){
                break;
            }

        }
    }

}

//return the afrho calculated. grid container for coma result is resized directly inside the method.
double run_numerical_model(std::vector<std::vector<double>> &grid, int gridsize_x, int gridsize_y, int cellsize, int gridsize_apex, 
                            double beta, int size_index, double particle_size, double v0, double particle_density,
                            double size_dependent_qd, double Q_sca, double phase_function, 
                           double amplitude_rot, double period_rot, int angular_dependency_index, double cosine_scaling,
                           const std::string &dumpModelDirectoryPath)
{                       
    
    numerical_model_sampling(gridsize_x, gridsize_y, gridsize_apex, cellsize, beta, size_index,
        particle_size, v0, particle_density, size_dependent_qd, Q_sca, phase_function, amplitude_rot,
        period_rot, angular_dependency_index, cosine_scaling, grid);

    double afrho = 0.0;
    std::vector<std::vector<double>> grid_smooth =
        perform_smoothing(grid, gridsize_x,gridsize_y, cellsize, gridsize_apex, size_index,
        particle_size, particle_density, afrho, Q_sca, phase_function, dumpModelDirectoryPath);
    grid = grid_smooth;

    return afrho;
}

//return afrho calculated
double run_fountain(std::vector<std::vector<double>> &grid, int gridsize_x, int gridsize_y, int cellsize, int gridsize_apex, 
                  double beta, double qd, double particle_size, double v0, int size_index, double particle_density, 
                  double Q_sca, double phase_function, const std::string & dumpModelDirectoryPath)
{
    grid.resize(gridsize_x, std::vector<double>(gridsize_y,0.0));

    double S;
    double R;

    double E;

    double eta;

    //times of flight
    double t1;
    double t2;

    //emission_angles
    double theta1;
    double theta2;

    //Jacobians
    double jac1;
    double jac2;

    //directional distribution
    double fs1;
    double fs2;

    double dust_density;

    double particle_mass = 4./3.*pi*particle_size*particle_size*particle_size*particle_density;


    E = calc_E(beta, v0);

    for (int x = 0; x < gridsize_x; x++){
        std::vector<double> &row = grid.at(x);

        for (int y = 0; y < gridsize_y; y++) {
            R = sqrt((1.0*x-gridsize_apex)*(1.0*x-gridsize_apex)+1.0*y*y)*1.0*cellsize;
            S = 1.0*cellsize*(x-gridsize_apex);
            row[y] = 0.0;

            if((2*E+S)*(2*E+S) - R*R > 0){
                calc_eta(E,S,R,&eta);
                if(R == 0){
                    t1 = 0;
                    t2 = 0;
                    theta1 = 0;
                    theta2 = 0;
                }
                else{
                    calc_time_of_flight(R,E,eta,&t1, &t2, v0);

                    emission_direction(R,S,t1,beta,&theta1, v0);
                    emission_direction(R,S,t2,beta,&theta2, v0);
                }

                if(R == 0){
                    jac1 = 1;
                    jac2 = 1;
                }
                else{
                    calc_jacobian(R,eta,&jac1, &jac2, v0);
                }
                directional_distribution(theta1,theta2,&fs1,&fs2);
                calc_dust_density(qd,fs1,fs2,jac1,jac2,&dust_density);

                row[y] = dust_density/particle_mass;
            }

        }
    }

    double afrho = calc_afrho(grid,gridsize_x,gridsize_y,cellsize, gridsize_apex,Q_sca,phase_function,particle_size);

    std::string filename = dumpModelDirectoryPath + "/dust_density_grid" + to_string(size_index) + ".dat";
    write_file(grid , filename);
    return afrho;
}

void run_model_all_sizes(int gridsize_x, int gridsize_y,int cellsize,int gridsize_apex,
                         int number_of_bins, double particle_density, double lower_size_limit, double higher_size_limit, double power_law_index,
                         double Q_gas, double radius_N,double nucleus_temp, int species_index, int phase_angle, double heliocentric_distance,
                         int day_night_asym_switch, int angular_dependency_index, double cosine_scaling,
                         int rotational_switch, double amplitude_rot, double period_rot,
                         int qd_estimation_switch, double qd_input, double afrho_input,
                         const std::string & dumpModelDirectoryPath, int &counterOfProcessedBins)
{
    // should be written to file
    std::vector<double> size_bins = get_size_bins(number_of_bins,lower_size_limit,higher_size_limit);

    std::vector<double> particle_densities = calc_particle_density(size_bins, particle_density);

    // should be written to file
    std::vector<double> terminal_velocities = get_terminal_v(size_bins, particle_densities,Q_gas,radius_N, nucleus_temp);
    //
    // Ph Reynier 28/07/23
    // add grid_test for functional test 10

    // Q_pr,Q_abs,Q_sca,phase_function should all be written to file
    // QData is a structure 4 x number of bins collecting Q_pr(QData[0]), Q_abs(QData[1]), Q_sca(QData[2]), phase_function(QData[3])
    std::vector<std::vector<double>> QData = get_scattering_properties(size_bins, phase_angle, species_index);

    // qd (dust production rate) should be written to file
    double qd = 0.0;
    if (qd_estimation_switch==1){
        qd = estimate_production_rate(afrho_input, power_law_index, size_bins, particle_densities, terminal_velocities, QData[2], QData[3]);
    }else{
        qd = qd_input;
    }

    // should be written to file
    std::vector<double> beta = calculate_beta(heliocentric_distance, size_bins, particle_densities, QData[2], QData[3]);

    // should be written to file
    std::vector<double> size_dependent_qd = get_size_dependent_qd(qd, power_law_index, size_bins, particle_densities,terminal_velocities);

    double afrho_calculated = 0.0;
    double total_afrho_calculated = 0.0;

    double estimate_afrho2 = 0.0;

    for (int i = 0; i < number_of_bins; i++){

        vector<vector<double>> grid(gridsize_x, vector<double>(gridsize_y, 0));


        if(day_night_asym_switch==1){


            afrho_calculated = run_numerical_model(grid,gridsize_x,gridsize_y,cellsize, gridsize_apex,
                        beta[i],i,size_bins[i],terminal_velocities[i],
                        particle_densities[i], size_dependent_qd[i], QData[2][i],
                        QData[3][i],amplitude_rot,period_rot, angular_dependency_index,
                        cosine_scaling, dumpModelDirectoryPath);

        }

        else if(day_night_asym_switch==0){
            afrho_calculated = run_fountain(grid,gridsize_x,gridsize_y,cellsize,gridsize_apex,
                         beta[i],size_dependent_qd[i],size_bins[i], terminal_velocities[i],i,
                         particle_densities[i],QData[2][i],QData[3][i], dumpModelDirectoryPath);

        }

        // for testing purposes - runs both models and sets the angular_dependency to 0 (so that they should give the same/very similar result)
        else{
           angular_dependency_index = 0;
           cosine_scaling = 0;

            afrho_calculated = run_numerical_model(grid,gridsize_x,gridsize_y,cellsize,gridsize_apex,
                            beta[i],i,size_bins[i],terminal_velocities[i],
                            particle_densities[i], size_dependent_qd[i], QData[2][i],
                            QData[3][i],amplitude_rot,period_rot,angular_dependency_index,
                            cosine_scaling, dumpModelDirectoryPath);

            //careful with afrho_calculated (fountain overwrites numerical model)
            afrho_calculated = run_fountain(grid,gridsize_x,gridsize_y,cellsize,gridsize_apex,
                         beta[i],size_dependent_qd[i],size_bins[i], terminal_velocities[i],i,
                         particle_densities[i],QData[2][i],QData[3][i], dumpModelDirectoryPath);
            

        }

        //incrementing the output counterOfProcessedBins for log purposes.
        ++counterOfProcessedBins;
        //summing up Afrho
        total_afrho_calculated += afrho_calculated;
    }

write_header(gridsize_x, gridsize_y, cellsize, gridsize_apex, particle_density, lower_size_limit,  
             higher_size_limit, power_law_index, Q_gas, radius_N, nucleus_temp,species_index, heliocentric_distance,
             day_night_asym_switch, angular_dependency_index, cosine_scaling, rotational_switch,  amplitude_rot, period_rot,
             qd_estimation_switch, qd,afrho_input,size_bins,terminal_velocities,beta, QData,phase_angle,
             size_dependent_qd, total_afrho_calculated, dumpModelDirectoryPath);

}

void run_trajectory_module(double d_ca, double phase_angle_of_approach, double theta_sc, 
                           double magnitude_of_velocity, double spacecraft_frontal_area,
                           const std::string &restoreModelDirectoryPath, 
                           const std::string &resultsOutputDirectoryPath)
{
    ComaHeader chunk = read_header(restoreModelDirectoryPath);
    std::vector<double> size_bins = get_size_bins(chunk.number_of_bins,chunk.lower_size_limit,chunk.higher_size_limit);

    //initialize total impact mass
    double total_mass = 0.0;
    std::vector<double> impact_mass_per_bin(chunk.number_of_bins,0.0);

    /*
     * all values below should be written to file
     * that is:
     * impact mass of each size bin [kg]
     * total impact mass (sum of all bins) [kg]
     */
    for(int i=0; i < chunk.number_of_bins; i++){

        std::vector<std::vector<double>> grid = read_grid(chunk.gridsize_x,chunk.gridsize_y, i, restoreModelDirectoryPath);
        impact_mass_per_bin[i] = mass_along_trajectory(grid, chunk.gridsize_x, chunk.gridsize_y, chunk.particle_density,
                                                       size_bins[i], chunk.cellsize, chunk.gridsize_apex, i, d_ca, 
                                                       phase_angle_of_approach, theta_sc, 
                                                       magnitude_of_velocity, spacecraft_frontal_area, resultsOutputDirectoryPath);
        total_mass += impact_mass_per_bin[i];

    }

    write_header_spacecraft(d_ca,phase_angle_of_approach,theta_sc, magnitude_of_velocity, 
                            spacecraft_frontal_area, chunk.number_of_bins, size_bins, impact_mass_per_bin, 
                            total_mass, resultsOutputDirectoryPath);

}

} //end namespace core
