//
// Created by nh18u590 on 17.02.2022.
// Modified by RArpa_OPT on 11.1.2023
//

#include "utils.h"
#include <iostream>
#include <tuple>
#include <math.h>
#include <fstream>
#include <random>
#include <string>
#include <iomanip>
#include <limits>
#include <vector>

using namespace std;

namespace core{

// Utility to allocate the matrix of coma results per bin on xy grid and initialize to zero.
// It's used in GUI python wrapper context
// x,y are int fimension for cells in x and y direction. 
// data is the output resized matrix structure, should be declared outside (the method will resize and initialize it.)

void initializeGrid(int x, int y, std::vector<std::vector<double>> & data){
    data.clear();
    data.resize(x, std::vector<double>(y, 0.0));
}

//calculates end_time
//Only needed to get a time that is longer than the time needed for the test particle to move out of scope
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void calc_endtime(double v0, double *max_time, double beta, int cellsize){
    double g = beta * G * M_sun / (dist_sun*dist_sun);
    *max_time = 2.0*(v0+sqrt(v0*v0+2.0*g*double(cellsize)*20000.0))/g;
}

// calculates the position of the test particle at a given time
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void trajectory(int *x, int *y, double *time, double *emission_angle, double *beta, double *v, double *v_temp, int cellsize) {
    //double g = (*beta) * G * M_sun / (dist_sun*dist_sun);
    double g = (*beta) * g_const;
    *x = floor((-(*v) * (*time) * cos(*emission_angle) + 0.5 * g * (*time)*(*time))/(1.0*cellsize));
    *y = (int)(*v * (*time) * sin(*emission_angle)/(1.0*cellsize));
    *v_temp = sqrt(g*g*(*time)*(*time) - 2.0*g*(*time)*(*v)*cos(*emission_angle) + (*v)*(*v));

}

//calculates time step between two calculations of the position of the test particle
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void get_time_step(double *time_step, double *v_temp, int cellsize){
    *time_step = double(cellsize)/(4.0*(abs(*v_temp)));
    }

//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void velocity_distribution(double v0, double *v, double emission_angle, int angular_function_index, double cosine_scaling) {

    if (angular_function_index == 0){
        *v = v0;
    }

    else if (angular_function_index == 1){
        //cosine
        *v = v0*(1.0+cosine_scaling*cos(emission_angle));
    }

    else if (angular_function_index == 2) {
        //gaussian
        double norm = 3.19174;
        double sigma = pi / 4.0;
        *v = v0*norm * exp(-emission_angle*emission_angle / (2.0 * sigma * sigma));
    }

    else if (angular_function_index == 3){
        //half-maxwellian
        double norm = 4.38001;
        *v = v0*norm*(emission_angle + sqrt2)*(emission_angle + sqrt2)*exp(-(emission_angle+sqrt2)*(emission_angle + sqrt2)/2.0);
    }


}

//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void get_qd_distribution(double *qd_distribution, double emission_angle, int angular_function_index, double cosine_scaling) {

    if (angular_function_index == 0){
        *qd_distribution = 1.0;
    }

    else if (angular_function_index == 1){
        //cosine
        *qd_distribution = 1.0 + cosine_scaling*cos(emission_angle);
    }

    else if (angular_function_index == 2) {
        //gaussian
        double norm = 3.19174;
        double sigma = pi / 4.0;
        *qd_distribution = norm * exp(-emission_angle*emission_angle / (2.0 * sigma * sigma));
    }

    else if (angular_function_index == 3){
        //half-maxwellian
        double norm = 4.38001;
        *qd_distribution = norm*(emission_angle + sqrt2)*(emission_angle + sqrt2)*exp(-(emission_angle+sqrt2)*(emission_angle + sqrt2)/2.0);
    }

}

/*
 * gives emission angle of test particle and a scaling factor accounting for the corresponding solid angle
 * emission angles are between 0° and 180°
 *
 */
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void get_emission_angle(double* emission_angle,double* solid_angle_scaling, int i, int number_of_trajectories){

    *emission_angle = double(i)/(1.0*number_of_trajectories)*pi;
    double dtheta = pi/double(number_of_trajectories);
    *solid_angle_scaling = abs(sin(*emission_angle)*sin(0.5*dtheta));
}

void write_file(const vector<vector<double>> &grid, const std::string &filename) {
    ofstream file(filename.c_str(), ofstream::binary);

    for(const std::vector<double> & chunk : grid) {
        if (!chunk.empty()) {
            const char *buffer = reinterpret_cast<const char*>(chunk.data());
            file.write(buffer, chunk.size()*sizeof(double));
        }
    }

    file.close();
}

void write_sc_output(const vector<double> &dust_density_of_time, const vector<double> &flux ,
                     const vector<double> & fluence, const vector<double> &R_coordinates, 
                     const vector<double> &S_coordinates, const vector<double> &time, const std::string & filename) 
{
    ofstream file(filename.c_str(), ofstream::binary);
    std::size_t nentries =dust_density_of_time.size(); 
    for(size_t i = 0; i < nentries; i++ ) {

        file << fixed << setprecision(20) << dust_density_of_time[i];
        file << ", ";
        file << fixed << setprecision(16) << flux[i];
        file << ", ";
        file << fixed << setprecision(16) << fluence[i];
        file << ", ";
        file << fixed << setprecision(16) << R_coordinates[i];
        file << ", ";
        file << fixed << setprecision(16) << S_coordinates[i];
        file << ", ";
        file << fixed << setprecision(3) << time[i];
        file << "\n";
    }
    file.close();
}

// calculates the volume of a grid cell.
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void cell_volume(double* volume, int y, int cellsize){

    double c = double(cellsize);
    *volume = 1.*c*c*c*(2.0*double(y)+1.0)*pi;

}

//calculates a scaling factor to account for variability of the activity due to the rotation of the nucleus
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void rotational_variablility(double amplitude, double period, double *scaling_rotational, double *time){
    if (period >= 0.01){
        *scaling_rotational = 1+amplitude*cos(2.*(*time)*pi/period);
    }
    else{
        *scaling_rotational = 1;
    }

}

//scaling factor that scales the test particles to the actual number of particles
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void scale_to_qd(double* scaling_qd, double qd_per_size, double* particle_mass){
    *scaling_qd = qd_per_size/(*particle_mass);
    //*scaling_qd = 1;
}

//pre-calculates the weights used for the smoothing to optimize the running time
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
double weights_smoothing(int i, int j, double sigma){
    double id = double(i);
    double jd = double(j);
    double weight = exp(-(id*id+jd*jd)/(2.0*sigma*sigma));
    return weight;
}

/*
 * calculates the trajectory of the spacecraft
 * d = distance to closest approach
 * delta_l = integration step
 * d_ca = distance at closest approach
 * s_ca = S-coordinate at closest approach
 */
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void trajectory_sc(int i, double theta, double phase_angle_of_approach, double d_ca, double s_ca, double start_dist, double delta_l, double* d, double* R, double* S){
    *d = start_dist + double(i)*delta_l;
    *R = sqrt((*d)*(*d) + d_ca*d_ca);
    *S = sin(phase_angle_of_approach)*(*d) + s_ca;

}

//calculates to which grid coordinates a given R and S correspond
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
void get_dust_density_index(int* x_position, int* y_position,double delta_l, double R, double S, int cellsize, int gridsize_apex){
    *x_position = round(S/double(cellsize)) + gridsize_apex;
    *y_position = round(sqrt(R*R - S*S)/double(cellsize));
}

//calculates S (which in our grid corresponds to the x-coordinate) at closest approach
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
double calculate_s_ca(double theta, double phase_angle_of_approach, double d_ca){
    double s_ca = sin(theta)*cos(phase_angle_of_approach)*d_ca;

    return s_ca;

}

// iv describes the ratio of the gas mass in a volume with the cross section of the particle and a characteristic length to the particle mass
// it is needed to calculate the terminal velocity
// an approximation to the numerical solution is used
// see Marschall et. al. (2022) for more details
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
double calculate_iv(double Q_gas, double radius_N, double particle_density, double particle_size, double nucleus_temp){
    double Iv;
    double molecular_weight = 0.018;
    double qgas_giotto = 5.5*pow(10,29);
    Iv = 4.038*pow(10.0,-29)*Q_gas*avogadro/(molecular_weight*radius_N*particle_density*particle_size*sqrt(nucleus_temp));

    return Iv;
}

// calculates the terminal dust velocity (see Marschall et. al. (2022) for theory behind it)
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
double calculate_vd(double Q_gas, double radius_N, double particle_density, double particle_size, double nucleus_temp){
    double vd;
    double iv = calculate_iv(Q_gas,radius_N,particle_density,particle_size,nucleus_temp);
    vd = 58.903*sqrt(nucleus_temp)/(1.0+0.6*sqrt(1.5/iv));

    return vd;
}

// gets the terminal dust velocities for all size bins
std::vector<double> get_terminal_v(const std::vector<double> &size_bins, const std::vector<double> &particle_densities, double Q_gas, double radius_N,double nucleus_temp){
    int number_of_bins = size_bins.size();
    std::vector<double> terminal_velocities(number_of_bins);
    for (int i = 0; i < number_of_bins; ++i){
        terminal_velocities[i]= calculate_vd(Q_gas,radius_N,particle_densities[i],size_bins[i],nucleus_temp);
    }
    return terminal_velocities;
}


//return the 2D matrix of grid smoothed.
std::vector<std::vector<double>>
perform_smoothing(const vector<vector<double>>&grid, int gridsize_x, int gridsize_y, int cellsize, int gridsize_apex, 
                  int size_index, double particle_size, double particle_density, double &afrho_calculated, double Q_sca, 
                  double phase_function, const std::string & dumpModelDirectoryPath)
{
    /*
     * as close to the nucleus the particles are closer to each other, but the dust density variations are larger a smaller smoothing radius is used close to the nucleus
     * in total 4 smoothing radii in relation to the distance are defined
     * 0-200 km
     * 200 - 2000 km
     * 2000 - 20000 km
     * further than 20000 km
     */


    std::vector<std::vector<double>> weights1, weights2, weights3, weights4;
    std::vector<int> smoothing_radius = prepareSmoothingWeights(cellsize, weights1, weights2, weights3, weights4);

    std::vector<std::vector<double>> grid_smooth(gridsize_x, std::vector<double> (gridsize_y,0));
 
    // loops over the whole grid
    for (int x = 0; x < gridsize_x; x++){
        for (int y = 0; y < gridsize_y; y++){
            grid_smooth[x][y] = smoothingValue(x,y,gridsize_x, gridsize_y, gridsize_apex, cellsize, 
                grid, smoothing_radius, weights1, weights2, weights3, weights4);
            if (isnan(grid_smooth[x][y])){
                cout << "There is a NaN value in the smoothed grid" << endl;
            }
        }
    }

    afrho_calculated = calc_afrho(grid_smooth,gridsize_x,gridsize_y,cellsize,gridsize_apex,Q_sca,phase_function, particle_size);

    std::string filename = dumpModelDirectoryPath + "/dust_density_grid" + to_string(size_index) + ".dat";
    write_file(grid_smooth,filename);

    return grid_smooth;
}

//RA 1/8 - fine grain utilities composing the perform smoothing - to be python binded

/*!
    Prepare the smoothing weights pool 
    \param[in] cellsize size of the cell actually in use in the computational grid
    \param[out] weights1 first level of weights (target dist 40000)
    \param[out] weights2 first level of weights (target dist 80000)
    \param[out] weights3 first level of weights (target dist 180000)
    \param[out] weights4 first level of weights (target dist 400000)
    \return numbers of smoothing radius for weights1, 2,3 and 4 respectively.
*/
std::vector<int> prepareSmoothingWeights(
    int cellsize,
    std::vector<std::vector<double>> & weights1, 
    std::vector<std::vector<double>> & weights2, 
    std::vector<std::vector<double>> & weights3, 
    std::vector<std::vector<double>> & weights4)
{

    /*
     * as close to the nucleus the particles are closer to each other, but the dust density variations are larger a smaller smoothing radius is used close to the nucleus
     * in total 4 smoothing radii in relation to the distance are defined
     * 0-200 km
     * 200 - 2000 km
     * 2000 - 20000 km
     * further than 20000 km
     COMMENT PROBABLY  OUTDATED (1/8/2023)   
     */

    std::vector<int> smoothing_radius(4);
    smoothing_radius[0] = round(40000.0/double(cellsize));
    double sigma1 = double(smoothing_radius[0])/1.5;

    smoothing_radius[1] = round(80000.0/double(cellsize));
    double sigma2 = double(smoothing_radius[1])/1.75;

    smoothing_radius[2] = round(180000.0/double(cellsize));
    double sigma3 = double(smoothing_radius[2])/1.75;
    
    smoothing_radius[3] = round(400000.0/double(cellsize));
    double sigma4 = double(smoothing_radius[3])/2.0;

    // this part initializes & pre-computes the weights of the smoothing
    weights1 = std::vector<std::vector<double>>(smoothing_radius[0], std::vector<double>(smoothing_radius[0], 0.0));
    for (int i = 0; i < smoothing_radius[0]; i++){
        for (int j = 0; j < smoothing_radius[0]; j++) {
            weights1[i][j] = weights_smoothing(i,j,sigma1);
        }
    }

    weights2 = std::vector<std::vector<double>>(smoothing_radius[1], std::vector<double>(smoothing_radius[1], 0.0));
    for (int i = 0; i < smoothing_radius[1]; i++){
        for (int j = 0; j < smoothing_radius[1]; j++) {
            weights2[i][j] = weights_smoothing(i,j,sigma2);
        }
    }

    weights3 = std::vector<std::vector<double>>(smoothing_radius[2], std::vector<double>(smoothing_radius[2], 0.0));
    for (int i = 0; i < smoothing_radius[2]; i++){
        for (int j = 0; j < smoothing_radius[2]; j++) {
            weights3[i][j] = weights_smoothing(i,j,sigma3);
        }
    }

    weights4=std::vector<std::vector<double>> (smoothing_radius[3], std::vector<double>(smoothing_radius[3], 0.0));
    for (int i = 0; i < smoothing_radius[3]; i++){
        for (int j = 0; j < smoothing_radius[3]; j++) {
            weights4[i][j] = weights_smoothing(i,j,sigma4);
        }
    }

    return smoothing_radius;
}

/*!
    Smooth the values in a particular position of grid x, y
    \param[in] x positional index on grid
    \param[in] y positional index on grid
    \param[in] gridsize_x total number of elements in x for grid
    \param[in] gridsize_y total number of elements in y for grid
    \param[in] gridsize_apex total number of elements composing the apex in x direction
    \param[in] cellsize size of the grid cell
    \param[in] grid matrix of values to be smoothed.
    \param[in] smoothing_radius vector fo 4 values of smoothing radii for weights 1,2,3,4 in this order
    \param[in] weights1 matrix of weights referring to smoothing_radius[0]
    \param[in] weights2 matrix of weights referring to smoothing_radius[1]
    \param[in] weights3 matrix of weights referring to smoothing_radius[2]
    \param[in] weights4 matrix of weights referring to smoothing_radius[3]
    \return local smoothed value of grid in x,y position
    
*/
double smoothingValue(
    int x, int y, int gridsize_x, int gridsize_y, int gridsize_apex, int cellsize,
    const std::vector<std::vector<double>> & grid,
    std::vector<int> & smoothing_radius,
    std::vector<std::vector<double>> & weights1, 
    std::vector<std::vector<double>> & weights2, 
    std::vector<std::vector<double>> & weights3, 
    std::vector<std::vector<double>> & weights4 )
{
    double normalization = 0.0;
    int smoothing_y = 0;

    double smoothvalue  = 0.0;
    //smoothing_value = (grid[x-1][y-1]+grid[x][y-1]+grid[x+1][y-1]+grid[x-1][y]+grid[x][y]+grid[x+1][y]+grid[x-1][y+1]+grid[x][y+1]+grid[x+1][y+1])/9;

    if(double(x-gridsize_apex)*double(x-gridsize_apex)+double(y)*double(y) < (200000.0/double(cellsize))*(200000.0/double(cellsize))){
        // loops over the cells relevant for the smoothing of the current cell
        if (smoothing_radius[0] != 0) {
            for (int i = -smoothing_radius[0]+1; i < smoothing_radius[0]; i++) {

                smoothing_y = round(sqrt(double(smoothing_radius[0] * smoothing_radius[0]) - double(i * i)));
                for (int j = -smoothing_y+1; j < smoothing_y; j++) {
                    if (((x + i) >= 0) && ((x + i) < gridsize_x) && ((y + j) >= 0) && ((y + j) < gridsize_y)) {

                        smoothvalue += grid[x + i][y + j] * weights1[abs(i)][abs(j)];
                        normalization += weights1[abs(i)][abs(j)];
                    }
                }
            }
        }
        else{
            smoothvalue = grid[x][y];
            normalization = 1.0;
        }
    }
    else if(double(x-gridsize_apex)*double(x-gridsize_apex)+double(y)*double(y) < (2000000.0/double(cellsize))*(2000000.0/double(cellsize))){
        // loops over the cells relevant for the smoothing of the current cell
        for (int i = -smoothing_radius[1]+1; i < smoothing_radius[1]; i++){

            smoothing_y = round(sqrt(double(smoothing_radius[1]*smoothing_radius[1]) - double(i*i)));
            for (int j = -smoothing_y+1; j < smoothing_y; j++){
                if (((x+i) >= 0) && ((x+i) < gridsize_x) && ((y+j) >= 0) && ((y+j) < gridsize_y)){

                    smoothvalue += grid[x+i][y+j]*weights2[abs(i)][abs(j)];
                    normalization += weights2[abs(i)][abs(j)];
                }
            }
        }
    }
    else if(double(x-gridsize_apex)*double(x-gridsize_apex)+double(y)*double(y) < (20000000.0/double(cellsize))*(20000000.0/double(cellsize))){

        // loops over the cells relevant for the smoothing of the current cell
        for (int i = -smoothing_radius[2]+1; i < smoothing_radius[2]; i++){

            smoothing_y = round(sqrt(double(smoothing_radius[2]*smoothing_radius[2]) - double(i*i)));
            for (int j = -smoothing_y+1; j < smoothing_y; j++){
                if (((x+i) >= 0) && ((x+i) < gridsize_x) && ((y+j) >= 0) && ((y+j) < gridsize_y)){

                    smoothvalue += grid[x+i][y+j]*weights3[abs(i)][abs(j)];
                    normalization += weights3[abs(i)][abs(j)];
                }
            }
        }
    }
    else if(double(x-gridsize_apex)*double(x-gridsize_apex)+double(y)*double(y) >= (20000000.0/double(cellsize))*(20000000.0/double(cellsize))){
        for (int i = -smoothing_radius[3]+1; i < smoothing_radius[3]; i++){

            smoothing_y = round(sqrt(double(smoothing_radius[3]*smoothing_radius[3]) - double(i*i)));
            for (int j = -smoothing_y+1; j < smoothing_y; j++){
                if (((x+i) >= 0) && ((x+i) < gridsize_x) && ((y+j) >= 0) && ((y+j) < gridsize_y)){

                    smoothvalue += grid[x+i][y+j]*weights4[abs(i)][abs(j)];
                    normalization += weights4[abs(i)][abs(j)];
                }
            }
        }
    }

    smoothvalue/=normalization;

    return smoothvalue;
}

// calculates afrho
// gets f*rho/geometric_cross_section from function calculate_frho
double calc_afrho(const vector<vector<double>> &grid_smooth , int gridsize_x, int gridsize_y,int cellsize,int gridsize_apex, double Q_sca, double phase_function, double particle_size){

    double afrho;
    double f_rho;

    f_rho = calculate_frho(grid_smooth,gridsize_x,gridsize_y,cellsize, gridsize_apex);

    afrho = f_rho*Q_sca*phase_function*particle_size*particle_size*pi;


    return afrho;
}

// calculates f*rho/geometric_cross_section, but for the sake of "shorter" variable names in the function this is called frho
// the filling factor f is calculated for a column at a given distance (about 150km) from the nucleus, which optimizes for the running time
// this distance is called rho

double calculate_frho(const vector<vector<double>> &grid_smooth , int gridsize_x, int gridsize_y, int cellsize, int gridsize_apex){

    int y_temp;
    int rho;

    //rho = int(500 * 1000 / cellsize);

    vector<vector<double>> grid_column(gridsize_x, vector<double> (gridsize_y,0));

    double integration_step = 100.0;

    double frho = 0.0;

    //rho is set as 150km, but rounded to the precision of the grid
    rho = round(150000.0/double(cellsize));


    double x;
    double y;

    /*
     * loops over and averages column density of 4 points in the coma in the distance defined above
     * a point in anti-sun direction, a point in sun direction and on both sides
     * points on the side are only calculated once for symmetry reasons
     */
    for (int j = 0; j < 3; j++) {

        double angle = pi*double(j)/2;

        x = gridsize_apex + round(double(rho)*cos(angle));
        y = round(abs(double(rho)*sin(angle)));


        // performs integration to get the column density

        for (int i = 1; i < gridsize_y*int(integration_step); i++) {
            y_temp = round(sqrt(y * y + 1.0 * (double(i) * double(i)) / (integration_step * integration_step)));
            if (y_temp < gridsize_y - 1) {
                grid_column[x][y] += grid_smooth[x][y_temp] * (double(cellsize) / (integration_step));
            }


        }

        if (x==gridsize_apex){
            grid_column[x][y] = 2.0*(2.0*grid_column[x][y] + double(cellsize)/integration_step*grid_smooth[x][y]);
        }
        else{
            grid_column[x][y] = 2.0*grid_column[x][y] + double(cellsize)/integration_step*grid_smooth[x][y];
        }

        /*
         * factor 2.0 comes from 2*pi*r = circumference
         * factor 1/4.0 comes from averaging over the 4 points
        */
        frho += (2.0/4.0)* grid_column[x][y] * double(rho) * double(cellsize);

    }

    return frho;
}

// function calculates particle densites
// at the moment it just returns a constant value, however this could be changed at a later date
std::vector<double> calc_particle_density(const std::vector<double> &particle_sizes, double particle_density){
    int number_of_bins = particle_sizes.size();
    std::vector<double> particle_densities(number_of_bins);
    for(int i = 0; i < number_of_bins; i++){
        // double rho0 = 3.0;
        // double rho1 = 2.65;
        // double a0 = 2.0*pow(10,-6);
        // particle_density[i] = 1000*(rho0-rho1*(particle_sizes[i]/(particle_sizes[i]+a0)));
        particle_densities[i] = particle_density;
    }
    return particle_densities;
}

// gets scattering properties of all size bins as a matrix QData 4 x number_of_bins. Return Q_pr in QData[2],
// Q_sca in QData[0], Q_abs in QData[1] and phase_function in QData[3]
// calls interpolate function.
std::vector<std::vector<double>> 
get_scattering_properties(const std::vector<double> &size_bins,int phase_angle, int species_index){
    int number_of_bins = size_bins.size();
    std::vector<std::vector<double>> QData(4, std::vector<double>(number_of_bins));
    for(int i= 0; i< number_of_bins; ++i) {
        interpolate(QData[0][i],QData[1][i],QData[2][i],QData[3][i],phase_angle,size_bins[i],species_index);
    }
    return QData;
}

// calculates beta parameter for all size bins
std::vector<double> calculate_beta(double heliocentric_distance, const std::vector<double> &size_bins, const std::vector<double> &particle_densities, 
                                   const std::vector<double> &Q_pr, const std::vector<double> &phase_function)
{
    int number_of_bins = size_bins.size();
    std::vector<double> beta(number_of_bins);
    for (int j = 0; j < number_of_bins; j++) {
        beta[j] = 3.0 * (L_sun/(heliocentric_distance*heliocentric_distance)) * Q_pr[j] / (16.0 * pi * lightspeed * G * M_sun * particle_densities[j] * size_bins[j]);
    }
    return beta;
}


// calculates size bins
std::vector<double> get_size_bins(int number_of_bins, double lower_limit_exp, double higher_limit_exp){
    std::vector<double> size_bins(number_of_bins);
    for (int i = 0; i < number_of_bins; i++){
        size_bins[i]= pow(10,lower_limit_exp - (lower_limit_exp - higher_limit_exp)*i*1.0/(1.0*number_of_bins-1.0));
    }
    return size_bins;
}

//estimates production rate based on afrho
//this function is highly sensitive to the inputs and can create unphysical results if not used with care

double estimate_production_rate(double afrho, double power_law_index, const std::vector<double> &size_bins, 
                                const std::vector<double> &particle_densities, const std::vector<double> &terminal_velocities,
                                const std::vector<double> & Q_sca, const std::vector<double> &phase_function)
{

    // see Fink and Rubin (2012) for theory

    double denom = 0.0;
    int number_of_bins = size_bins.size();
    // double a1 = size_bins[number_of_bins-1]/sqrt(size_bins[number_of_bins-2]/size_bins[number_of_bins-1]);
    // double a2 = size_bins[0]*sqrt(size_bins[0]/size_bins[1]);

    double a1 = size_bins[number_of_bins-1]/pow(10.0,(6.0/(1.0*number_of_bins)));
    double a2 = size_bins[0];
    
    double normalisation = integrate_bin_vol(a1,a2,power_law_index);
    for (int i = 0; i < number_of_bins; i++) {

        if ((i != 0) && (i != (number_of_bins-1))){
            a2 = size_bins[i];
            a1 = size_bins[i+1];
        }

        else if(i == 0){
            a2 = size_bins[i];
            a1 = size_bins[i+1];
        }

        else{
            a2 = size_bins[i];
            a1 = size_bins[i]/pow(10.0,(6.0/(1.0*number_of_bins)));
        }


        denom += 0.5*Q_sca[i]*phase_function[i]*integrate_bin_vol(a1,a2,(power_law_index))/((4.0/3.0)*particle_densities[i]*size_bins[i]*normalisation*terminal_velocities[i]);
    }
    return afrho/denom;

}


/* calculates bin size dependent dust production rate in kg/s
* checker sums over all production rates and should be equal to the global dust production rate
 * qd is the dust production rate
 * a1, a2 is the integration interval
*/
std::vector<double> get_size_dependent_qd( double qd, double power_law_index, const std::vector<double> &size_bins, 
                                           const std::vector<double> &particle_densities, const std::vector<double> &terminal_velocities)
{

    int number_of_bins = size_bins.size();
    std::vector<double> size_dependent_qd(number_of_bins);

    //double a1 = size_bins[number_of_bins-1]/sqrt(size_bins[number_of_bins-2]/size_bins[number_of_bins-1]);
    //double a2 = size_bins[0]*sqrt(size_bins[0]/size_bins[1]);

    double span = abs(log10(size_bins[0])-log10(size_bins[number_of_bins-1]));
    

    double a1 = size_bins[number_of_bins-1]/pow(10.0,(span/(1.0*number_of_bins)));
    double a2 = size_bins[0];

//    double a1 = size_bins[number_of_bins-1]/pow(10,(6.0/(1.0*number_of_bins)));
//    double a2 = size_bins[0];

    double normalisation = integrate_bin_vol(a1,a2,power_law_index);

    // see below
    //double checker = 0.0;



    for (int i = 0; i < number_of_bins; i++) {

//        if ((i != 0) and (i != (number_of_bins-1))){
//            a2 = size_bins[i]*sqrt(size_bins[i-1]/size_bins[i]);
//            a1 = size_bins[i]/sqrt(size_bins[i]/size_bins[i+1]);
//        }
//
//        else if(i == 0){
//            //a2 = size_bins[i]*sqrt(size_bins[i]/size_bins[i+1]);
//            a1 = size_bins[i]/sqrt(size_bins[i]/size_bins[i+1]);
//            a2 = size_bins[0];
//
//        }
//
//        else{
//            a2 = size_bins[i]*sqrt(size_bins[i-1]/size_bins[i]);
//            //a1 = size_bins[i]/sqrt(size_bins[i-1]/size_bins[i]);
//            a1 = size_bins[i];
//        }

        if ((i != 0) && (i != (number_of_bins-1))){
            a2 = size_bins[i];
            a1 = size_bins[i+1];
        }

        else if(i == 0){
            a2 = size_bins[i];
            a1 = size_bins[i+1];
        }

        else{
            a2 = size_bins[i];
            a1 = size_bins[i]/pow(10.0,(span/(1.0*number_of_bins)));
        }



        size_dependent_qd[i] = qd*integrate_bin_vol(a1,a2,power_law_index)/normalisation;


        // could potentially be used as test
        //checker += size_dependent_qd[i];



    }

    return size_dependent_qd;

}

//calculates mass along spacecraft trajectory
double mass_along_trajectory(vector<vector<double>>&grid_smooth, int gridsize_x, int gridsize_y, double particle_density, 
                             double particle_size, int cellsize, int gridsize_apex,int size_index ,double d_ca, 
                             double phase_angle_of_approach, double theta, double speed, double scarea, const std::string &outputResultsDirPath)
{
    d_ca = d_ca*1000.0;
    phase_angle_of_approach = phase_angle_of_approach*pi/180.0;
    theta = theta*pi/180.0;

    speed = speed*1000.0;

    double spacecraft_frontal_area = scarea;

    double s_ca = calculate_s_ca(theta, phase_angle_of_approach, d_ca);
    int max_distance = (gridsize_x-gridsize_apex)*cellsize;
    double start_dist = -max_distance;
    double delta_l = cellsize/20.0;

    int number_of_steps = abs(round(start_dist*2.0/delta_l));


    double delta_t = delta_l/speed;
    double start_time = -double(max_distance)/speed;

    vector<double> time(number_of_steps,0);
    vector<double> dust_density_of_time(number_of_steps,0);
    vector<double> fluence(number_of_steps,0);
    vector<double> flux(number_of_steps,0);
    vector<double> R_coordinates(number_of_steps,0);
    vector<double> S_coordinates(number_of_steps,0);


    int x_position = -1;
    int y_position = -1;

    double dust_density = 0.0;

    double d = 0.;
    double R = 0.;
    double S = 0.;

    for (int i = 0; i < number_of_steps; i++){
        trajectory_sc(i,theta,phase_angle_of_approach,d_ca,s_ca,start_dist, delta_l, &d, &R, &S);
        R_coordinates[i] = R;
        S_coordinates[i] = S;
        time[i] = start_time + double(i)*delta_t;

        get_dust_density_index(&x_position, &y_position, delta_l, R, S, cellsize, gridsize_apex);

        if ((x_position >= 0) && (x_position <= gridsize_x-1) && (y_position >= 0) && (y_position <= gridsize_y-1)){
            flux[i] = grid_smooth[x_position][y_position]*speed;
            dust_density += grid_smooth[x_position][y_position]*delta_l;
            fluence[i] = dust_density;
            dust_density_of_time[i] = grid_smooth[x_position][y_position];

            /*
            * for debug/testing purposes only - can be used to "draw" a line corresponding to which values are taken from the grid
            *
             * grid_smooth[x_position][y_position] = 10;
            * grid_smooth[x_position+1][y_position] = 10;
            * */

        }

        // returns NaN whenever trajectory is not in the range of the grid
        else{
            flux[i] = numeric_limits<double>::quiet_NaN();
            fluence[i] = numeric_limits<double>::quiet_NaN();
            dust_density_of_time[i] = numeric_limits<double>::quiet_NaN();
        }
    }


    std::string filename = outputResultsDirPath + "/dust_density_along_trajectory" + to_string(size_index) + ".txt";
    write_sc_output(dust_density_of_time,flux, fluence, R_coordinates,S_coordinates, time, filename);

    double impact_mass;
    impact_mass = 4.0/3.0*pi*particle_size*particle_size*particle_size*particle_density*dust_density*spacecraft_frontal_area;

    return impact_mass;

}

/*
 * reads in scattering files
 * species index is used to decide which file is read in
 * 0 for water ice (n = 1.31 + 0i)
 * 1 for silicate (enstatite n = 1.63 + 4.4e-5 i)
 * 2 for astronomical silicate (n = 1.81 + 0.1012i)
 * 3 for Troilite (FeS n = 1.40 + 1.25i)
 *
 * input files need to have the following form:
 * Q_sca
 * Q_abs
 * Q_pr
 * Phi_s(0°)
 * Phi_s(1°)
 * etc.
 * Phi_s(180°)
 */

void read_files(double *q_abs,double *q_sca, double *q_pr, double *phase_function, int phase_angle, double exponent, int species_index){
    double a[181]= {0.0};
    double q_ext;
    string str;
    string species;
    str = to_string(exponent);
    str = str.substr(0,4);

    if (species_index==0){
        species = "./databases/H2O/";
    }
    else if (species_index==1){
        species = "./databases/Silicate/";
    }
    else if (species_index==2){
        species = "./databases/astro_silicate/";
    }
    else if (species_index==3){
        species = "./databases/Troilite/";
    }
    std::string filename = species + str + ".txt"; 
    ifstream data(filename);
    if(!data.is_open()){
        std::cout<<"CoreAPI:utils:read_files: missing missing databases folder! cannot open file: " <<filename<<std::endl;
        throw std::runtime_error("missing databases folder! cannot open file: " + filename);
    }

    int i = 0;
    while(!data.eof()){
        if (i==0){
            data >> *q_sca;
        }
        else if(i == 1){
            data >> *q_abs;
        }
        else if(i == 2){
            data >> *q_pr;
        }

        else{
            data >> a[i-3] ;

        }

        i++;

    }
    data.close();
    *phase_function = a[phase_angle];

}

//interpolates scattering parameters. Data read from databases folder. See read_files. 
//linear interpolation in y = log(particle_size), x = scattering_parameter
//exponents in scattering files are given in steps of 0.05 precision
void interpolate(double &Q_abs, double &Q_sca, double &Q_pr, double &phase_function, int phase_angle, double particle_size, int species_index){

    double exponent1;
    double exponent2;

    double Q_abs1;
    double Q_abs2;

    double Q_sca1;
    double Q_sca2;

    double Q_pr1;
    double Q_pr2;

    double phase_function1;
    double phase_function2;

    double exponent = -log10(particle_size);

    //as exponents in scattering files are specified in steps of 0.05 precision this gets the exponents adjacent on both sides on the input value
    exponent1 = floor(20.0*exponent)/20.0;
    exponent2 = ceil(20.0*exponent)/20.0;

    if ((exponent1-exponent2)!=0){
        read_files(&Q_abs1,&Q_sca1,&Q_pr1,&phase_function1,phase_angle,exponent1,species_index);
        read_files(&Q_abs2,&Q_sca2,&Q_pr2,&phase_function2,phase_angle,exponent2,species_index);

        Q_abs = Q_abs1*(exponent2-exponent)/(exponent2-exponent1)+Q_abs2*(exponent-exponent1)/(exponent2-exponent1);
        Q_sca = Q_sca1*(exponent2-exponent)/(exponent2-exponent1)+Q_sca2*(exponent-exponent1)/(exponent2-exponent1);
        Q_pr = Q_pr1*(exponent2-exponent)/(exponent2-exponent1)+Q_pr2*(exponent-exponent1)/(exponent2-exponent1);
        phase_function = phase_function1*(exponent2-exponent)/(exponent2-exponent1)+phase_function2*(exponent-exponent1)/(exponent2-exponent1);

    }

    else{
        read_files(&Q_abs1,&Q_sca1,&Q_pr1,&phase_function1,phase_angle, exponent,species_index);
        Q_abs = Q_abs1;
        Q_sca = Q_sca1;
        Q_pr = Q_pr1;
        phase_function = phase_function1;

    }


}

void write_header(int gridsize_x, int gridsize_y, int cellsize, int gridsize_apex,double particle_density, 
                  int lower_size_limit, int higher_size_limit,double power_law_index_passed, double Q_gas,double radius_N,
                  double nucleus_temp,int species_index,double heliocentric_distance,int day_night_asym_switch,
                  int angular_dependency_index,double cosine_scaling, int rotation_switch, double amplitude_rot, 
                  double period_rot, int qd_estimation_switch, double qd,double afrho_input, const vector<double> &size_bins, 
                  const vector<double> &terminal_velocities, const vector<double> &beta, const std::vector<vector<double>> &Qdata, int phase_angle, 
                  const vector<double> &size_dependent_qd,double total_afrho_calculated, const std::string &storeDirectoryPath)
{

    std::string filename = storeDirectoryPath + "/header.txt";

    ofstream file(filename, ofstream::binary);

    int number_of_bins = size_bins.size();

    file << "gridsize_x: ";
    file << gridsize_x;
    file << "\n";
    file << "gridsize_y: ";
    file << gridsize_y;
    file << "\n";
    file << "number_of_bins: ";
    file << number_of_bins;
    file << "\n";
    file << "Nucleus_position: ";
    file << gridsize_apex;
    file << "\n";
    file << "cellsize[m]: ";
    file << cellsize;
    file << "\n";

    file << "power_law_index: ";
    file << -power_law_index_passed;
    file << "\n";
    file << "lower_size_limit(10^)[m]: ";
    file << lower_size_limit;
    file << "\n";
    file << "higher_size_limit(10^)[m]: ";
    file << higher_size_limit;
    file << "\n";

    file << "Particle_density[kg/m^3]: ";
    file << particle_density;
    file << "\n";
    file << "Gas_production_rate[kg/s]: ";
    file <<  Q_gas;
    file << "\n";
    file << "Radius_nucleus[m]: ";
    file << radius_N;
    file << "\n";
    file << "Nucleus_surface_temperature[K]: ";
    file << nucleus_temp;
    file << "\n";

    file << "Material_choice: ";
    if (species_index==0){
        file << "Water_ice";
    }
    else if (species_index==1){
        file << "Enstatite";
    }
    else if (species_index==2){
        file << "Astro_silicate";
    }
    else if (species_index==3){
        file << "Troilite";
    }
    file << "\n";

    file << "Helio_centric_distance[AU]: ";
    file << heliocentric_distance;
    file << "\n";

    file << "Day_night_asymmetry: ";
    if (day_night_asym_switch==0){
        file << "No";
    }
    else if (day_night_asym_switch==1){
        file << "Yes";
    }
    file << "\n";

    file << "Angular_dependency: ";

    if (day_night_asym_switch==0){
        file << "-";
    }
    else if (angular_dependency_index==1){
        file << "Cosine";
    }
    else if (angular_dependency_index==2){
        file << "Gaussian";
    }
    else if (angular_dependency_index==3){
        file << "Half_Maxwellian";
    }
    file << "\n";

    file << "Cosine_scaling: ";
    if(angular_dependency_index==1){
        file << cosine_scaling;
    }
    else{
        file << "-";
    }
    file << "\n";

    file << "Rotational_variability: ";
    if (rotation_switch==0){
        file << "No";
        file << "\n";
        file << "Rotational_period[s]: ";
        file << "-";
        file << "\n";
        file << "Amplitude_of_rotational_variability: ";
        file << "-";
    }
    else if (rotation_switch==1){
        file << "Yes";
        file << "\n";
        file << "Rotational_period[s]: ";
        file << period_rot;
        file << "\n";
        file << "Amplitude_of_rotational_variability: ";
        file << amplitude_rot;
    }
    file << "\n";

    file << "Dust_production_rate_estimation: ";
    if (qd_estimation_switch==0){
        file << "No";
        file << "\n";
        file << "Afrho_input: ";
        file << "-";
        file << "\n";
        file << "Dust_production_rate[kg/s]: ";
        file << qd;
        file << "\n";
    }
    else if(qd_estimation_switch==1){
        file << "Yes";
        file << "\n";
        file << "Afrho_input[m]: ";
        file << afrho_input;
        file << "\n";
        file << "Dust_production_rate[kg/s]: ";
        file << qd;
        file << "\n";
    }

    file << "particle_size_bins[m]: ";
    for (int i=0; i<number_of_bins; i++){
        file << size_bins[i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";


    file << "terminal_velocities[m/s]: ";
    for (int i=0; i<number_of_bins; i++){
        file << terminal_velocities[i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "beta: ";
    for (int i=0; i<number_of_bins; i++){
        file << beta[i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "Q_pr: ";
    for (int i=0; i<number_of_bins; i++){
        file << Qdata[2][i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "Q_abs: ";
    for (int i=0; i<number_of_bins; i++){
        file << Qdata[1][i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "Q_sca: ";
    for (int i=0; i<number_of_bins; i++){
        file << Qdata[0][i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "phase_angle: ";
    file << phase_angle;
    file << "\n";

    file << "phase_function: ";
    for (int i=0; i<number_of_bins; i++){
        file <<Qdata[3][i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "production_rate_per_size_bin[kg/s]: ";
    for (int i=0; i<number_of_bins; i++){
        file << size_dependent_qd[i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "calculated_afrho[m]: ";
    file << total_afrho_calculated;
    file << "\n";



    file.close();

}

void write_header_spacecraft(double d_ca, double phase_angle_of_approach, double theta_sc, double magnitude_of_velocity, 
                             double spacecraft_frontal_area,int number_of_bins, const vector<double> &size_bins, 
                            const vector<double> &impact_mass_per_bin, double total_mass, const std::string &outputResultsDirPath)
{

    std::string filename = outputResultsDirPath + "/header_trajectory.txt";

    ofstream file(filename, ofstream::binary);


    file << "distance_at_closest_approach[km]: ";
    file << d_ca;
    file << "\n";

    file << "phase_angle_of_approach[°]: ";
    file << phase_angle_of_approach;
    file << "\n";

    file << "altitude_angle[°]: ";
    file << theta_sc;
    file << "\n";

    file << "spacecraft_velocity[km/s]: ";
    file << magnitude_of_velocity;
    file << "\n";

    file << "spacecraft_frontal_area[m^2]: ";
    file << spacecraft_frontal_area;
    file << "\n";

    file << "size_bins[m]: ";
    for (int i=0; i<number_of_bins; i++){
        file << size_bins[i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "impact_mass_per_bin[kg]: ";
    for (int i=0; i<number_of_bins; i++){
        file << impact_mass_per_bin[i];
        if(i!= number_of_bins-1){
            file << ", ";
        }
    }
    file << "\n";

    file << "total_impact_mass[kg]: ";
    file << total_mass;
    file << "\n";


    file.close();

}

ComaHeader read_header(const std::string &headerDirPath)
{                 
    std::string filename = headerDirPath+ "/header.txt";
    ifstream headerfile(filename);
    if(!headerfile.is_open()){
        throw std::runtime_error("failed opening the coma header file "+ filename);
    }
    ComaHeader chunk;
    std::string buffer;

    headerfile >> buffer>> chunk.gridsize_x;
    headerfile >> buffer>> chunk.gridsize_y;
    headerfile >> buffer>> chunk.number_of_bins;
    headerfile >> buffer>> chunk.gridsize_apex;
    headerfile >> buffer>> chunk.cellsize;
    headerfile >> buffer>> chunk.power_law_index_used;
    chunk.power_law_index_passed = -1.0*chunk.power_law_index_used;
    headerfile >> buffer>> chunk.lower_size_limit;
    headerfile >> buffer>> chunk.higher_size_limit;
    headerfile >> buffer>> chunk.particle_density;

    return chunk;
}

std::vector<std::vector<double>> read_grid(int gridsize_x, int gridsize_y, int size_index, const std::string &modelDataDirPath)
{

    std::string filename = modelDataDirPath+ "/dust_density_grid" + to_string(size_index) + ".dat";
    std::vector<std::vector<double>> grid(gridsize_x, vector<double> (gridsize_y,0));

    ifstream file(filename.c_str(), ofstream::binary);

    for(std::vector<double> &chunk : grid) {
        if (!chunk.empty()) {
            file.read(reinterpret_cast<char*>(chunk.data()), sizeof(double)*gridsize_y);
        }
    }

    file.close();

    return grid;

}

/*
 * calculates volume (mass/density) of particles in a bin
 * input:
 * particle_size_min - start point of interval of integration
 * particle_size_max - end point of interval of integration
 * power_law_index - power_law_index
 * output:
 * total volume of particles in a size range of particles
 */
//YOU CAN MAKE THIS INLINE see utils_fountain (RA)
double integrate_bin_vol(double particle_size_min, double particle_size_max, double power_law_index){

    double total_vol;

    if (abs(power_law_index + 4.0) >= 0.001){
        total_vol = (pow(particle_size_max,(4.0+power_law_index)) - pow(particle_size_min,(4.0+power_law_index)))/(4.0+power_law_index);
    }

    else{
        total_vol = log(particle_size_max) - log(particle_size_min);
    }

    return total_vol;

}

} // end namespace core

