import sys
import os
from PySide2.QtWidgets import (
    QAction,
    QApplication,
    QWidget,
    QMainWindow,
    QStatusBar,
    QHBoxLayout,
    QVBoxLayout,
    QHBoxLayout,
    QFileDialog,
    QLabel,
    QFrame,
    QMessageBox,
    QProgressBar,
    QInputDialog,
    QLineEdit
)

from PySide2.QtGui import QIcon
from PySide2.QtCore import QThread, Qt

from frontEnd.StatusProject import StatusProject
from frontEnd.Visualization import PlotCanvas2D
from frontEnd.ParametersTab import ParamTabWidget
from frontEnd.Logger import LoggerWidget
import frontEnd.resources
import backEnd.runFunctions as BERun
import backEnd.fileFunctions as BEFile

import xml.etree.ElementTree as ET
import xml.dom.minidom as XDOM
import time
import glob
import re
import shutil


class MainWindowApp(QMainWindow):
    '''!
        ComMode application main window 
        Create the main widget and draw inside objects provided by internal buildBasicUI method
    '''

    def __init__(self, applicationRoot):
        '''!
            Constructor.
            @param applicationRoot : path where the main file using the current object is installed.
        '''
        super(MainWindowApp, self).__init__()
        ## store the position of the root invoking the current object
        self.applicationFolder = applicationRoot

        self.setWindowTitle("ComMoDE")
        ## istance to store the path of the stylesheet css information.
        self.stylesheet_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "frontEnd/css/customstyle.qss")
        with open(self.stylesheet_path, 'r') as f:
            self.setStyleSheet(f.read())
        ## instance to store the position of icon assets
        self.iconAddress = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), 'frontEnd/assets/')

        self.setIcon()
        self.buildBasicUI()

        # private utility to track down the status of the project
        self.__status = StatusProject()

        # useful connections to control the front-end appearance/back-end action according to the status of the project
        self.__status.comaChanged.connect(self.__comaChangedActions)
        self.__status.flyByChanged.connect(self.__flyByChangedActions)

        self.__status.updateComaStatus(False)
        self.__status.updateFlyByStatus(False)

        ## dedicated thread private instance to run coma simulations
        self.__comaRunner = QThread()
        ## private utility to run back-end coma simulation jobs
        self.__comaWorker = BERun.ComaSimulationWorker()
        self.__comaWorker.moveToThread(self.__comaRunner)

        #edit connections
        self.__comaRunner.started.connect(self.__comaWorker.runProcessComa) 
        self.__comaWorker.finishedComa.connect(self.__comaRunner.quit, Qt.BlockingQueuedConnection) 
        self.__comaRunner.finished.connect(self.__finalizeComaAction)

        # connect stopAction and stop button to terminated slot action of the comaWorker
        self.paramTabWidget.comaParameters.stopComaButton.clicked.connect(self.__comaWorker.terminate)

        #connect logger and progress bar to comaWorker
        self.__comaWorker.messageRaisedComa.connect(self.logger.appendPlainText, Qt.BlockingQueuedConnection)
        self.__comaWorker.processAdvancesComa.connect(self.progressBar.setValue, Qt.BlockingQueuedConnection)

        ## dedicated thread private instance to run trajectories
        self.__flybyRunner = QThread()
        ## private utility to run back-end flyby trajectories jobs
        self.__flybyWorker = BERun.FlyByProcessWorker()
        self.__flybyWorker.moveToThread(self.__flybyRunner)


        #edit connections
        self.__flybyRunner.started.connect(self.__flybyWorker.processFlyBy) 
        self.__flybyWorker.finishedFlyBy.connect(self.__flybyRunner.quit, Qt.BlockingQueuedConnection) 
        self.__flybyRunner.finished.connect(self.__finalizeFlyByAction)

        # connect stopAction and stop button to terminated slot action of the flybyWorker
        self.paramTabWidget.flybyParameters.stopFlyByButton.clicked.connect(self.__flybyWorker.terminate) 
        #connect logger and progress bar to comaWorker
        self.__flybyWorker.messageRaisedFlyBy.connect(self.logger.appendPlainText, Qt.BlockingQueuedConnection) 
        self.__flybyWorker.processAdvancesFlyBy.connect(self.progressBar.setValue, Qt.BlockingQueuedConnection) 


        self.__deactivateCurrentProject()

    def setIcon(self):
        '''!
        Attach a representative icon to the main window.
        '''
        appIcon = QIcon(self.iconAddress + "comet.png")
        self.setWindowIcon(appIcon)

    def buildBasicUI(self):
        '''!
        Create the basic layout of the app - MenuBar/StatusBar, ProjectLabel, InputSection, 
        Viz canvas, logger and progress bar.
        '''
        # main widget
        ## Central widget of the application
        self.centralWidget = QWidget()
        self.setCentralWidget(self.centralWidget)

        ## Progressbar of the application
        self.progressBar = QProgressBar()
        self.progressBar.setRange(0, 100)
        self.progressBar.setValue(0)

        ## Project label: when an active project will be loaded in the application, its name will appear in this widget.
        self.projectLabel = QLabel()
        self.projectLabel.setWordWrap(True)
        self.projectLabel.setText("")
        self.projectLabel.setFrameStyle(QFrame.Sunken)
        self.projectLabel.setFixedHeight(30)

        ## Project Folder path: when an active project will be loaded, its full path will be stored into this variable.
        self.activeProjectFolder = ""

        ## Top wrapper Widget for collecting coma and trajectories User inputs forms
        self.paramTabWidget = ParamTabWidget(self.iconAddress)

        ## Top plot2D widget carrying utilities to visualize results
        self.plotCanvas2D = PlotCanvas2D(self.iconAddress)

        ## Logger widget for collecting feedback messages from the application
        self.logger = LoggerWidget("ComMoDE Logger:\n")

        # fix the layout

        ## Abstract layout widget for positionining (on the Left) the User inputs forms (paramTabWidget)
        self.LeftWidget = QWidget()
        layout2 = QVBoxLayout()
        layout2.setContentsMargins(2, 2, 2, 2)
        layout2.addWidget(self.projectLabel)
        layout2.addWidget(self.paramTabWidget)
        self.LeftWidget.setLayout(layout2)

        ## Abstract layout widget for positionining (on the Right) logger and visualizator (plotCanvas2S, log)
        self.RightWidget = QWidget()
        layout3 = QVBoxLayout()
        layout3.setContentsMargins(2, 2, 2, 2)
        layout3.addWidget(self.plotCanvas2D, stretch=3)
        layout3.addWidget(self.logger, stretch=1)
        self.RightWidget.setLayout(layout3)

        ## Abstract layout to host Left and Right widgets in unique one.
        self.focusWidget = QWidget()
        layout1 = QHBoxLayout()
        layout1.setContentsMargins(2, 2, 2, 2)
        layout1.addWidget(self.LeftWidget, stretch=2)
        layout1.addWidget(self.RightWidget, stretch=3)

        self.focusWidget.setLayout(layout1)

        # add it in a vertical layout and send them finally to the central widget
        layout0 = QVBoxLayout()
        layout0.setContentsMargins(2, 2, 2, 2)
        layout0.addWidget(self.focusWidget)
        layout0.addWidget(self.progressBar)

        self.centralWidget.setLayout(layout0)

        # creating the menu bar at the end (need all other objects to be defined, so don't move from here)
        self.create_menu()

    # creating the main menu
    def create_menu(self):
        '''!
            Defining the menu bar of the application, adding the 3 topics File, Run, Help and list all the actions that can
            be done for menus. 
        '''
        mainMenu = self.menuBar()

        # assing  menus
        fileMenu = mainMenu.addMenu("File")
        runMenu = mainMenu.addMenu('Run')
        helpMenu = mainMenu.addMenu('Help')

        # Actions (Sub Items)
        iconAddress = self.iconAddress

        ## hosting the QAction object that handle the creation of a new project
        self.file_newProjectAction = QAction(
            QIcon(iconAddress+"file-new.png"), "New Project", self)
        ## hosting the QAction object that handle the opening of a pre-existent project
        self.file_openProjectAction = QAction(
            QIcon(iconAddress+"reboot.png"), "Open Project", self)
        ## hosting the QAction object that handle the saving of a new project
        self.file_saveProjectAction = QAction(
            QIcon(iconAddress+"save.png"), "Save Project", self)
        ## hosting the QAction object that handle the cloning of a project into an independent one.
        self.file_cloneProjectAction = QAction(
            QIcon(iconAddress+"edit-copy.png"), "Clone Project into", self)
        ## hosting the QAction object that handle the clearing of a project and restore its default state
        self.file_eraseProjectAction = QAction(
            QIcon(iconAddress+"reset.png"), "Clear Project", self)
        ## hosting the QAction object that exit the application
        self.file_exitAction = QAction(
            QIcon(iconAddress+"exit.png"), "Exit", self)

        ## hosting the QAction that run the cometary simulation
        self.run_runComa = QAction(
            QIcon(iconAddress+"start.png"), "Run Coma simulation", self)
        ## hosting the QAction that run the trajectories post-processing
        self.run_flybymission = QAction(
            QIcon(iconAddress+"satellite2.png"), "Run S/C Fly-By mission", self)

        ## hosting the QAction that access ComMode doxygen documentation
        self.help_viewdoc = QAction(
            QIcon(iconAddress+"doxy.png"), "Access Documentation", self)
        ## hosting the QAction that print useful information on what ComMoDE is
        self.help_whatis = QAction(
            QIcon(iconAddress+"whatis.png"), "What is ComMoDE...", self)

        # adding Actions to the menu(sub Items)
        fileMenu.addAction(self.file_newProjectAction)
        fileMenu.addAction(self.file_openProjectAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.file_cloneProjectAction)
        fileMenu.addAction(self.file_saveProjectAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.file_eraseProjectAction)
        fileMenu.addSeparator()
        fileMenu.addAction(self.file_exitAction)

        runMenu.addAction(self.run_runComa)
        runMenu.addSeparator()
        runMenu.addAction(self.run_flybymission)


        helpMenu.addAction(self.help_viewdoc)
        helpMenu.addSeparator()
        helpMenu.addAction(self.help_whatis)

        # tooltips
        self.file_newProjectAction.setStatusTip("Create a brand New Project ")
        self.file_openProjectAction.setStatusTip(
            "Load an already existent Project ")
        self.file_saveProjectAction.setStatusTip(
            "Save current Project modifications ")
        self.file_cloneProjectAction.setStatusTip(
            "Copy/Backup current Project into another independent project")
        self.file_eraseProjectAction.setStatusTip(
            "Clear all contents of the current project and restore defaults")
        self.file_exitAction.setStatusTip("Exit the program")

        self.run_runComa.setStatusTip("Run the Cometary simulation")
        self.run_flybymission.setStatusTip("Run the Spacecraft fly-by mission (trajectories) once the Cometary simulation is performed")

        self.help_viewdoc.setStatusTip(
            "Open html ComMoDe Documentation using browser")
        self.help_whatis.setStatusTip("General information on ComMoDe")

        self.setStatusBar(QStatusBar(self))

        # to be done, coding slot activated by the action (except for exit)
        self.file_newProjectAction.triggered.connect(self.newProjectAction)
        self.file_openProjectAction.triggered.connect(self.openProjectAction)
        self.file_saveProjectAction.triggered.connect(self.saveProjectAction)
        self.file_cloneProjectAction.triggered.connect(self.cloneProjectAction)
        self.file_eraseProjectAction.triggered.connect(self.eraseProjectAction)
        self.file_exitAction.triggered.connect(self.exitApp)

        self.run_runComa.triggered.connect(self.runComaAction)
        # connect also redundancy runComaButton in comaParameters form
        self.paramTabWidget.comaParameters.runComaButton.clicked.connect(self.runComaAction)

        self.run_flybymission.triggered.connect(self.runFlyByAction)
        # connect also redundancy runComaButton in flybyparameters form
        self.paramTabWidget.flybyParameters.runFlyByButton.clicked.connect(self.runFlyByAction)

        # run_stopComa and comaParameters stopComaButton signal connection
        # is done internally in run Coma Action. . same for flyby in run FlyByAction
        #
        self.help_viewdoc.triggered.connect(self.accessDoc)
        self.help_whatis.triggered.connect(self.whatIsComMoDE)

        # to be done coding shortcuts if needed
        # fileMenu.grabShortcut("Ctrl+F")
        # openAction.setShortcut("Ctrl+O")
        self.file_exitAction.setShortcut("Ctrl+X")

    def closeEvent(self, event):
        '''!
            Overriding QMainWindow::closeEvent to let perform kill of ongoing processes and save of the input forms.
        '''
        if self.__comaRunner.isRunning():
            self.paramTabWidget.comaParameters.stopComaButton.clicked()

        if self.__flybyRunner.isRunning():
            self.paramTabWidget.flybyParameters.stopFlyByButton.clicked()

        ## add flyby process here too.
        if self.activeProjectFolder:
            self.saveProjectAction()
        QMainWindow.closeEvent(self, event)

    def exitApp(self):
        '''!
            Slot to close the application main window.
        '''
        # understand if coma simulation is running.
        if self.__comaRunner.isRunning():
            ret = QMessageBox.warning(self, "Found coma simulation running",
                                            "All results will be lost. Do you want to terminate the ongoing simulation?", QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
            if ret == QMessageBox.Yes:
                # triggering the stop action will terminate the coma running thread and activate the finalization.
                self.paramTabWidget.comaParameters.stopComaButton.clicked()
            else:
                return
        elif self.__flybyRunner.isRunning():
            ret = QMessageBox.warning(self, "Found flyby processing running",
                                            "Some trajectories results may be not completed. Do you want to terminate the ongoing job?", QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
            if ret == QMessageBox.Yes:
                # triggering the stop action will terminate the coma running thread and activate the finalization.
                self.paramTabWidget.flybyParameters.stopFlyByButton.clicked()
            else:
                return
        else:
            self.close()

    def writePrettifiedXMLDocument(self, xmlElement, filepath):
        '''!
            Internal utility to write a xml document in nice readable human format
            @param xmlElement instance of ElementTree.Element filled elsewhere
            @param filepath path of xml file to write. Format tag must be mandatorily *.xml
        '''
        rstring = ET.tostring(xmlElement, 'utf-8')
        xdom = XDOM.parseString(rstring)
        towritestring = xdom.toprettyxml(indent="    ")
        with open(filepath, "w") as file:
            file.write(towritestring)

    def __activateNewProject(self, dirpath):
        '''!
            Private utility to make a project folder named as name and located in parent root dir as active project
            folder of the application. The method will enable automatically the paramTabWidget forms
            @param dirpath full path to the project folder 
        '''
        self.projectLabel.setText(os.path.basename(dirpath))
        self.activeProjectFolder = dirpath
        self.paramTabWidget.setEnabled(True)
        self.plotCanvas2D.setEnabled(True)
        self.run_runComa.setEnabled(True)
        self.run_flybymission.setEnabled(True)
        self.plotCanvas2D.projectFolder = self.activeProjectFolder
 
    def __deactivateCurrentProject(self):
        '''!
            Private utility that deactivate the current project folder and the input forms related.
        '''
        self.projectLabel.setText("")
        self.activeProjectFolder = ""
        self.paramTabWidget.setEnabled(False)
        self.plotCanvas2D.setEnabled(False)
        self.run_runComa.setEnabled(False)
        self.run_flybymission.setEnabled(False)
        self.plotCanvas2D.projectFolder = ''
 
    def __comaChangedActions(self):
        '''!
            If the coma status changed in the project the following actions are performed
            1) if coma status is false, fly by status if forcefully set to false. Otherwise do nothing
            2) if coma status is false/true, fly-by front end form is disabled/enabled.
            3) if coma status is false/true, fly-by actions are disabled/enabled.
        '''
        flag = self.__status.getComaStatus()
        self.paramTabWidget.flybyParameters.setEnabled(flag)
        self.run_flybymission.setEnabled(flag)
        if not flag:
            self.__status.updateFlyByStatus(flag)

    def __flyByChangedActions(self):
        '''!
            If the fly by status changed in the project the following actions are performed
            1) if fly-by status is false/true, visualization canvas is disabled/enabled.
            2) if fly-by status is true, the plotCanvas2D will update information on available 
            trajectories (call to parseTrajectories)
        '''
        if self.__status.getFlyByStatus() :
            self.plotCanvas2D.setEnabled(True)
            self.plotCanvas2D.parseTrajectories()
        else:
            self.plotCanvas2D.setEnabled(False)
            self.plotCanvas2D.cleanTrajectories()

        self.plotCanvas2D.updateFigure()

    def accessDoc(self):
        '''!
            Method to open a dialog widget to expose the User usuful info on ComMoDE documentation (where it can be found, 
            how he can have access to it, etc ...)
            Documentation should be built during the ComMoDE installation stage. If not, please consult the Install 
            guide coming along with ComMoDe sources
        '''
        installationPath = os.path.normpath(
            os.path.join(self.applicationFolder, ".."))
        docpath = os.path.join(installationPath, "doc",
                               "ComMoDE", "html", "index.html")

        if not os.path.exists(docpath):
            text = "Cannot find html documentation at {}: please consider to reinstall ComMoDE using the -enableDoc option".format(
                docpath)
            QMessageBox.warning(self, "Cannot find Documentation", text)
        else:
            text = "Html documentation available at {}: Please open it with your preferred browser (chrome, firefox, etc...)".format(
                docpath)
            QMessageBox.information(self, "Documentation available", text)

    def whatIsComMoDE(self):
        '''!
            Method to open a dialog widget to expose ComMoDE project basic information
        '''
        text = "Cometary Model of Dust Enviroments (ComMoDE) is a Comet Dust Emission calculator coming with a practical Graphic User Interface.\n"
        text += "Project funded by ESA (contract 4000136014/21/NL/CRS), and lead by ISASpace(FR) - University of Bern (CH) - Optimad (IT) consortium, 2022-2023.\n"
        text += "\n"
        text += "ESA referent : Fabrice Cipriani (Fabrice.Cipriani@esa.int)\n"
        text += "\n"
        text += "Consortium referent : Philippe Reynier (Philippe.Reynier@isa-space.eu)"

        QMessageBox.information(self, "what is ComMoDE", text)

    def newProjectAction(self):
        '''!
            Slot to create a new project. It delegates folders check and creation to backend file method createNewProject.
            The front-end will backup the User inputs into two xml files (one dedicated to the coma simuation inputComa.xml, 
            the other to trajectories post-processing inputFlyBy.xml). Such files will be stored inside the brand-new created folder.
            Internally invokes the backEnd.fileFunction.createNewProject
        '''
        self.__deactivateCurrentProject()

        # open a dialog to get the Root directory of the New Project Directory
        dir = QFileDialog.getExistingDirectory(self, "Select Root Path directory", os.path.expanduser(
            "~"), QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks)

        if not dir:
            QMessageBox.warning(self, "No valid root directory selected",
                                "Cannot create a new project folder without selecting a root path first.")
            return

        text = "New project folder name: (new folder will be created in {})".format(dir)
        name, ok = QInputDialog.getText(
            self, "Get Project Folder Name", text, QLineEdit.Normal, "... myNewProject ...")

        if not ok or not name:
            QMessageBox.warning(self, "No valid project name inserted",
                                "Try again inserting a valid project name.")
            return

        check = BEFile.createNewProject(dir, name, self.applicationFolder, self.logger)
        if not check:
            QMessageBox.warning(self, "Project folder creation failed",
                                "Project folder cannot be created. See logger widget for further details.")
            return

        self.__activateNewProject(os.path.join(dir, name))

        # set all forms of paramTabWidget to defaults
        self.paramTabWidget.setDefaults()

        # dump current coma form content to inputComa.xml
        comatree = ET.Element('root')
        self.paramTabWidget.dumpComaToXML(comatree)
        self.writePrettifiedXMLDocument(comatree, os.path.join(
            self.activeProjectFolder, "inputComa.xml"))
        # dump current fly by form content to inputFlyBy.xml
        flybytree = ET.Element('root')
        self.paramTabWidget.dumpFlyByToXML(flybytree)
        self.writePrettifiedXMLDocument(flybytree, os.path.join(
            self.activeProjectFolder, "inputFlyBy.xml"))

        # update and dump status of the project
        self.__status.updateComaStatus(False)
        self.__status.updateFlyByStatus(False)
        self.__status.writeOnFile(self.activeProjectFolder)

    def saveProjectAction(self):
        '''!
            Save the input form data from coma and flyby to xml files (inputComa.xml and inputFlyBy.xml) 
            and update the status of the project. 
            It requires a valid project folder to be created/loaded first.
        '''
        if not self.activeProjectFolder:
            self.logger.appendPlainText("Open a Project or Create a new one first. Nothing to save")
            return

        # dump current coma form content to inputComa.xml
        comatree = ET.Element('root')
        self.paramTabWidget.dumpComaToXML(comatree)
        self.writePrettifiedXMLDocument(comatree, os.path.join(
            self.activeProjectFolder, "inputComa.xml"))
        # dump current fly by form content to inputFlyBy.xml
        flybytree = ET.Element('root')
        self.paramTabWidget.dumpFlyByToXML(flybytree)
        self.writePrettifiedXMLDocument(flybytree, os.path.join(
            self.activeProjectFolder, "inputFlyBy.xml"))
        # write status file
        self.__status.writeOnFile(self.activeProjectFolder)

    def openProjectAction(self):
        '''!
            Open a pre-existent ComMoDe project folder. 
            Let the user choose the path to target project with a dialog.
            Internally invoke backEnd.fileFunctions.checkValidProject to verify it.
            Activate the project in the GUI if everything is ok
        '''
        # deactivate the current project

        self.__deactivateCurrentProject()

        # open a dialog to get the project directory
        dir = QFileDialog.getExistingDirectory(self, "Select An Existent Project", os.path.expanduser(
            "~"), QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks)

        if not dir:
            QMessageBox.warning(self, "No valid folder selected",
                                "Please try again selecting a folder.")
            return

        check = BEFile.checkValidProject(dir, self.logger)
        if not check:
            QMessageBox.warning(self, "Unvalid Project folder detected",
                                "Cannot upload {}. See logger widget for further details.".format(dir))
            return

        self.__activateNewProject(dir)

        # restore status of the project from file
        self.__status.restoreFromFile(self.activeProjectFolder)
        # restore current coma form contents from inputComa.xml
        comatreeroot = ET.parse(os.path.join(self.activeProjectFolder, "inputComa.xml")).getroot()
        self.paramTabWidget.restoreComaFromXML(comatreeroot)

        # restore current flyby form contents from inputFlyBy.xml
        flybytreeroot = ET.parse(os.path.join(self.activeProjectFolder, "inputFlyBy.xml")).getroot()
        self.paramTabWidget.restoreFlyByFromXML(flybytreeroot)

        # last step check if valid coma/flyby and compare current forms with backup
        # information if any

        # find a backupInputComa
        comabackupfile = None
        if str(self.__status.getComaStatus()) == 'True':
            comabackups = sorted(glob.glob(os.path.join(self.activeProjectFolder, "backupInputComa*.xml")), reverse=True)
            if comabackups:
                comabackupfile = comabackups[0]
        
        if comabackupfile:
            backuptreeroot = ET.parse(comabackupfile).getroot()
            if not self.paramTabWidget.compareComaWithXML(backuptreeroot):
                message = "Latest valid coma simulation results were generated with a set of "
                message = message + "inputs different from those reported in inputComa.xml.\n"
                message = message + "Do you want to restore coherent inputs in the Coma form?"
                ret = QMessageBox.warning(self, "Coma inputs not coherent with latest coma simulation results",
                                          message, QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
                if ret == QMessageBox.Yes:
                    self.paramTabWidget.restoreComaFromXML(backuptreeroot)
                    inputcomapath = os.path.join(
                        self.activeProjectFolder, "inputComa.xml")
                    os.remove(inputcomapath)
                    os.rename(comabackupfile, inputcomapath)

       # find a backupInputFlyBy
        flybybackupfile = None
        if str(self.__status.getFlyByStatus()) == 'True':
            flybybackups = sorted(
                glob.glob(os.path.join(self.activeProjectFolder, "backupInputFlyBy*.xml")), reverse=True)
            if flybybackups:
                flybybackupfile = flybybackups[0]

        if flybybackupfile:
            backuptreeroot = ET.parse(flybybackupfile).getroot()
            if not self.paramTabWidget.compareFlyByWithXML(backuptreeroot):
                message = "Latest valid flyby results were generated with a set of "
                message = message + "inputs different from those reported in inputFlyBy.xml.\n"
                message = message + "Do you want to restore coherent inputs in the FlyBy form?"
                ret = QMessageBox.warning(self, "FlyBy inputs not coherent with latest flyby post-processing results",
                                          message, QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
                if ret == QMessageBox.Yes:
                    self.paramTabWidget.restoreFlyByFromXML(backuptreeroot)
                    inputflybypath = os.path.join(
                        self.activeProjectFolder, "inputFlyBy.xml")
                    os.remove(inputflybypath)
                    os.rename(flybybackupfile, inputflybypath)
        
        self.logger.appendPlainText("Successfully re-activated project {}".format(self.projectLabel.text()))

    def cloneProjectAction(self):
        '''!
            Copy/backup the current active project in a brand new one. Internally invokes the backEnd.fileFunctions.backupProject
        '''
        if not self.activeProjectFolder:
            self.logger.appendPlainText("Open a Project or Create a new one first. Nothing to copy from")
            return

        # open a dialog to get the Root directory to place the New Project Directory
        dir = QFileDialog.getExistingDirectory(self, "Select Root Path directory", os.path.expanduser(
            "~"), QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks)

        if not dir:
            QMessageBox.warning(self, "No valid root directory selected",
                                "Cannot create a backup of the project without selecting a root path first.")
            return

        text = "Backup project folder name: (new folder will be created in {})".format(dir)
        name, ok = QInputDialog.getText(self, "Get Backup Project Folder Name", text, QLineEdit.Normal, "... backupProjectName ...")

        if not ok or not name:
            QMessageBox.warning(self, "No valid project name inserted", "Try again inserting a valid project name.")
            return

        check = BEFile.backupProject(self.activeProjectFolder, os.path.join(dir, name), self.logger)
        if not check:
            QMessageBox.warning(self, "Project folder backup creation failed",
                                "A Backup of current Project cannot be created. See logger widget for further details.")
            return
        else:
            self.logger.appendPlainText("project backup created in {0}. Use Open Project to access and activate it".format(os.path.join(dir, name)))

    def eraseProjectAction(self):
        '''!
            Clear the current active project and restore its default state. Invoking backEnd.fileFunctions.clearProject.
        '''
        if not self.activeProjectFolder:
            self.logger.appendPlainText(
                "Open a Project or Create a new one first. Nothing to clear")
            return

        message = "All progress achieved in the current project will be lost. Do you want to proceed anyway?"
        ret = QMessageBox.information(
            self, "Clearing current active project", message, QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

        if ret == QMessageBox.No:
            self.logger.appendPlainText("Clear Project action aborted per User Request")

        # invoking backEnd fellow method.
        check = BEFile.clearProject(self.activeProjectFolder, self.logger)
        if not check:
            QMessageBox.critical(self, "Project folder clearing failed",
                                 "Connot restore the default state for the current project. See logger widget for further details.")

        # fix inputComa.xml and inputFlyBy
        os.remove(os.path.join(self.activeProjectFolder, "inputComa.xml"))
        os.remove(os.path.join(self.activeProjectFolder, "inputFlyBy.xml"))

        # set all forms of paramTabWidget to defaults
        self.paramTabWidget.setDefaults()

        # dump current coma form content to inputComa.xml
        comatree = ET.Element('root')
        self.paramTabWidget.dumpComaToXML(comatree)
        self.writePrettifiedXMLDocument(comatree, os.path.join(
            self.activeProjectFolder, "inputComa.xml"))
        # dump current fly by form content to inputFlyBy.xml
        flybytree = ET.Element('root')
        self.paramTabWidget.dumpFlyByToXML(flybytree)
        self.writePrettifiedXMLDocument(flybytree, os.path.join(
            self.activeProjectFolder, "inputFlyBy.xml"))

        # update and dump status of the project
        self.__status.updateComaStatus(False)
        self.__status.updateFlyByStatus(False)
        self.__status.writeOnFile(self.activeProjectFolder)

        self.logger.appendPlainText("Cleared project {} : default state restored.".format(self.projectLabel.text()))

    def runComaAction(self):
        '''!
            Slot to run a Coma simulation. 
            Requires a valid project folder to be created first.
            Action performed are: 
            1. Auto save inputs form ->call to saveProjectAction
            2. Check coma input form and if the project has valid run coma results.
            3. Run coma simulation starting a thread with internal member self.__comaRunner and the Back End worker class ComaSimulationWorker, activate "Stop simulation" action, 
            and waiting for process to properly end or to be aborted bu the User.
            4. Collect outcomes for 3.
        '''
        if not self.activeProjectFolder:
            return

        # save latest modifications on graphical form and collect them in inputComa.xml inside the project folder
        self.saveProjectAction()

        # retrieve coma status
        comaStatus = self.__status.getComaStatus()
        # retrieve latest inputComa.xml run
        backupInputComaFile = None
        backupInputComaList = sorted(
            glob.glob(os.path.join(self.activeProjectFolder, "backupInputComa_*.xml")), reverse=True)
        if backupInputComaList:
            # pick up the most recent one.
            backupInputComaFile = backupInputComaList[0]

        # redundancy checks: if validComa is true, I expect to have a backupInputComaFile, otherwise validComa is false.
        if str(comaStatus) == 'True' and not backupInputComaFile:
            self.__status.updateComaStatus(False)
            comaStatus = self.__status.getComaStatus()

        # redundancy checks: if validComa is false, I expect to have no backupInputComaFile, otherwise i need to purge it.
        if str(comaStatus) == 'False' and backupInputComaFile:
            os.remove(backupInputComaFile)
            backupInputComaFile = ""

        # if valid coma results are present, check if new input form matches the old backupped form or not.
        if str(comaStatus) == 'True':
            rootXML = ET.parse(backupInputComaFile).getroot()
            check = self.paramTabWidget.compareComaWithXML(rootXML)
            if check:
                QMessageBox.information(self, "Simulation requested already performed",
                                        "Current coma input form did not changed from the last valid coma run. Results are already available in coma-results subfolder.")
                self.logger.appendPlainText("frontEnd.runComaAction : no changes in input Coma form, request already satisfied")
                return
            else:
                message = "A new coma simulation is about to run: any previous coma simulation data"
                message = message + " will be overwritten, any flyby results related to it will be"
                message = message + " erased since no longer valid. If you don't want to lose old data "
                message = message + " consider to back-up you current project first with Clone"
                message = message + " option in File menu.\n\n Do you want to run the new simulation anyway?"

                ret = QMessageBox.warning(self, "Old valid simulation and/or flyby results found", message, QMessageBox.Yes | QMessageBox.Abort, QMessageBox.Yes)
                if ret == QMessageBox.Abort:
                    self.logger.appendPlainText("frontEnd.runComaAction : action aborted per User request")
                    return
                else:
                    # remove topic folders and re.create them from scratch
                    shutil.rmtree(os.path.join(
                        self.activeProjectFolder, "coma-results"))
                    shutil.rmtree(os.path.join(
                        self.activeProjectFolder, "flyby-results"))
                    os.mkdir(os.path.join(
                        self.activeProjectFolder, "coma-results"))
                    os.mkdir(os.path.join(
                        self.activeProjectFolder, "flyby-results"))
                    # remove all backupInput files.
                    listbackups = glob.glob(os.path.join(
                        self.activeProjectFolder, "backupInput*.xml"))
                    for item in listbackups:
                        os.remove(item)

        #update target project folder of the comaworker instance
        self.__comaWorker.setProjectFolder(self.activeProjectFolder)
        # put the progress bar to 0 value
        self.progressBar.setValue(0)

        # put the valid coma status to false (this will disable automatically the fly-by parameters too)
        self.__status.updateComaStatus(False)

        # enable stop action/button in menu and coma form
        self.paramTabWidget.comaParameters.stopComaButton.setEnabled(True)

        # disable run action/button in menu and coma form to avoid reclicking
        self.run_runComa.setEnabled(False)
        self.paramTabWidget.comaParameters.runComaButton.setEnabled(False)

        # freeze file actions, flyby form and visualization environment if the coma is running
        self.file_newProjectAction.setEnabled(False)
        self.file_openProjectAction.setEnabled(False)
        self.file_saveProjectAction.setEnabled(False)
        self.file_cloneProjectAction.setEnabled(False)
        self.file_eraseProjectAction.setEnabled(False)

        # put coma form in readonly mode
        self.paramTabWidget.comaParameters.setReadOnly(True)

        # send a message to logger:
        self.logger.appendPlainText("Started coma simulation for project {0} on {1}".format(self.projectLabel.text(), time.strftime("%Y%m%d-%H%M%S")))

        # self.__comaRunner.setStackSize(4100000000)

        # starting the thread
        self.__comaRunner.start()
        self.__comaRunner.setPriority(QThread.HighestPriority)



    def __finalizeComaAction(self):
        '''!
            Private utility, perform follow up actions after the run Coma thread end. Standard actions of finalization are
            - trigger signal of valid/ unvalid coma status in the app
            - disable stop coma button/action and re-enabling run coma button/action
            - reenable all other widgets freezed in the app by rumComaAction primer.

            The method is invoked when these particular configurations happen:
                1. Clean coma run, with successfull response
                2. Failed coma run
                3. Coma Run aborted by User.
        
        '''
        # restore widgets first
        # disable stop action/button in menu and coma form
        self.paramTabWidget.comaParameters.stopComaButton.setEnabled(False)

        # re-enable run action/button in menu and coma form
        self.run_runComa.setEnabled(True)
        self.paramTabWidget.comaParameters.runComaButton.setEnabled(True)

        # unfreeze file actions, flyby form and visualization environment if the coma is running
        self.file_newProjectAction.setEnabled(True)
        self.file_openProjectAction.setEnabled(True)
        self.file_saveProjectAction.setEnabled(True)
        self.file_cloneProjectAction.setEnabled(True)
        self.file_eraseProjectAction.setEnabled(True)

        # put the progress bar to 0
        self.progressBar.setValue(0)
        statworker = self.__comaWorker.getStatus()

        if str(statworker) == "0":
            # clean run

            #remove any pre-existent backup
            listbackups = glob.glob(os.path.join(self.activeProjectFolder, "backupInputComa*.xml"))
            for item in listbackups:
                os.remove(item)
            # copy current inputComa.xml as backupInputComa with timestamp
            newInputComaBackup = os.path.join(self.activeProjectFolder, "backupInputComa_{}.xml".format(time.strftime("%Y%m%d%H%M%S")))
            shutil.copyfile(os.path.join(self.activeProjectFolder, "inputComa.xml"), newInputComaBackup)
            # update coma status to true
            self.__status.updateComaStatus(True)
        else:
            if str(statworker) == "1":
                QMessageBox.critical(self, "Simulation failed", "Please check on logger for further information.")
            if str(statworker) == "2":
                # print message of aborted run on logger.
                self.logger.appendPlainText("Simulation aborted per User request")

            self.__status.updateComaStatus(False)

        # finally make again the Coma form widget editable.
        self.paramTabWidget.comaParameters.setReadOnly(False)
        self.logger.appendPlainText("Coma simulation terminated for project {0} on {1}".format(self.projectLabel.text(), time.strftime("%Y%m%d-%H%M%S")))

    def runFlyByAction(self):
        '''!
            Slot to run processing of one or multiple trajectories. 
            Requires a valid project folder to be created first.
            Action performed are: 
            1. Auto save inputs form ->call to saveProjectAction
            2. cross check status of last flyby process and clean up the flyby-folder where the trajectories result will be stored
            3. Run one by one trajectories and store their result in a dedicated sub-folder inside the flyby-folder. The job is done
            with backEnd.runFunctions utility FlyByProcessWorker
            The job is a plain run and cannot be customly interrupted, unless User force to exit the application.
        '''
        if not self.activeProjectFolder:
            return

        # save latest modifications on graphical form and collect them in inputFlyby.xml inside the project folder
        self.saveProjectAction()

        # retrieve flyby status
        flybyStatus = self.__status.getFlyByStatus()

        flybyFolder = os.path.join(self.activeProjectFolder, "flyby-results")
        if str(flybyStatus) != 'True':
            #remove sub contents of flybyfolder
            shutil.rmtree(flybyFolder)
            os.mkdir(flybyFolder)             

        #update target project folder of the flybyworker instance
        self.__flybyWorker.setProjectFolder(self.activeProjectFolder)
        # put the progress bar to 0 value
        self.progressBar.setValue(0)

        # put the valid flyby status to false (this will disable automatically the visualization canvas)
        self.__status.updateFlyByStatus(False)

        # enable stop action/button in menu and flyby form to perform abort
        self.paramTabWidget.flybyParameters.stopFlyByButton.setEnabled(True)

        # disable run action/button in menu and flyby form to avoid reclicking
        self.run_flybymission.setEnabled(False)
        self.paramTabWidget.flybyParameters.runFlyByButton.setEnabled(False)

        # disable coma run action/button in menu and flyby form to avoid running a coma while processing the trajectories
        self.run_runComa.setEnabled(False)
        self.paramTabWidget.comaParameters.runComaButton.setEnabled(False)

        # freeze file actions 
        self.file_newProjectAction.setEnabled(False)
        self.file_openProjectAction.setEnabled(False)
        self.file_saveProjectAction.setEnabled(False)
        self.file_cloneProjectAction.setEnabled(False)
        self.file_eraseProjectAction.setEnabled(False)

        # put coma form in readonly mode
        self.paramTabWidget.comaParameters.setReadOnly(True)
        # put flyby form in readonly mode
        self.paramTabWidget.flybyParameters.setReadOnly(True)

        # send a message to logger:
        self.logger.appendPlainText("Started trajectories processing for project {0} on {1}".format(self.projectLabel.text(), time.strftime("%Y%m%d-%H%M%S")))

        # put max stack size of the thread to 10 Gb.
        # self.__flybyRunner.setStackSize(4100000000)
        # starting the thread
        self.__flybyRunner.start()
        self.__flybyRunner.setPriority(QThread.HighestPriority)



    def __finalizeFlyByAction(self):
        '''!
            Private utility, perform follow up actions after the flyby processing thread end. Standard actions of finalization are
            - trigger signal of valid/ unvalid flyby status in the app (run is unvalid if no valid results for all trajectories is present in flyby)
            - re-enabling run coma/flyby buttons/actions
            - reenable all other widgets freezed in the app by runFlyByAction primer.

            The method is invoked when these particular configurations happen:
                1. Clean coma run, with successfull response
                2. Failed coma run
                3. Coma Run aborted by forcefully exit the application.
        
        '''
        # restore widgets first
        # disable stop action/button in  flyby form 
        self.paramTabWidget.flybyParameters.stopFlyByButton.setEnabled(False)

        # re-enable run action/button 
        self.run_flybymission.setEnabled(True)
        self.paramTabWidget.flybyParameters.runFlyByButton.setEnabled(True)

        # re-enable run action/button in coma form
        self.run_runComa.setEnabled(True)
        self.paramTabWidget.comaParameters.runComaButton.setEnabled(True)

        # unfreeze file actions
        self.file_newProjectAction.setEnabled(True)
        self.file_openProjectAction.setEnabled(True)
        self.file_saveProjectAction.setEnabled(True)
        self.file_cloneProjectAction.setEnabled(True)
        self.file_eraseProjectAction.setEnabled(True)

        # put the progress bar to 0
        self.progressBar.setValue(0)

        #restore flyby form of self.paramTabWidget.flybyParameters from worker edited inputFlyBy.xml
        flybyfolder = os.path.join(self.activeProjectFolder, "flyby-results")
        xmlfile = os.path.join(self.activeProjectFolder, "inputFlyBy.xml")
        treeXML = ET.parse(xmlfile).getroot() 
        self.paramTabWidget.restoreFlyByFromXML(treeXML)

        #clean up trajectories sub-folders who correspond to a trajectory with status false
        countSurvivor = len(self.paramTabWidget.flybyParameters.trajectories)
        for traj in self.paramTabWidget.flybyParameters.trajectories:
            if not traj.status:
                folderToRemove = os.path.join(flybyfolder, traj.name)
                if  os.path.isdir(folderToRemove) :
                    shutil.rmtree(folderToRemove)
                countSurvivor = countSurvivor - 1

        #update status
        self.__status.updateFlyByStatus(countSurvivor > 0)

        #give a message if process was aborted by the User
        statworker = self.__flybyWorker.getStatus()
        if statworker == 2: 
            self.logger.appendPlainText("FlyBy mission job aborted per User request")

        # finally make again the Coma and FlyBy forms widget editable.
        self.paramTabWidget.comaParameters.setReadOnly(False)
        self.paramTabWidget.flybyParameters.setReadOnly(False)
        self.logger.appendPlainText("FlyBy processing terminated for project {0} on {1}".format(self.projectLabel.text(), time.strftime("%Y%m%d-%H%M%S")))

if __name__ == "__main__":

    # absorb root path where the application run.
    application_folder = os.path.dirname(os.path.abspath(__file__))

    app = QApplication()

    window = MainWindowApp(application_folder)
    window.resize(1366, 768)
    window.show()

    sys.exit(app.exec_())
