import os
import shutil
import glob


def runFFPrint(message, logger):
    '''!
        Internal utility to print string messages
        @param message string to write
        @param logger valid instance of frontEnd.Logger.LoggerWidget to write messages on. If None fall back and use standard print()
    '''
    if logger:
        logger.appendPlainText(message)
    else:
        print(message)


def createNewProject(rootDir, newName, applicationRoot, logger=None):
    '''!
        Create a brand new project folder for running ComMoDe coma simulation and flyby
        spacecraft trajectories post-processing.
        The folder will act as a confined, stand-alone workspace to run app functionalities.
        Project folder will contain:
        - database subfolder (inherited from installation files)
        - coma-results subfolder
        - flyby-results subfolder
        All the needed input will be copied here, all the significant outputs will be stored here too
        (in coma-results sub-folder for simulation results and flyby-results for trajectories post-processing).
        @param rootDir where the new project folder will be created
        @param newName name of the new directory project
        @param applicationRoot path where the main ComMoDE-gui.py file is located.
        @param logger (optional) frontEnd.Logger.LoggerWidget instance to write meaningful messages.
        @return True for project folder successfully created, False otherwise
    '''
    newproject = os.path.join(rootDir, newName)
    if os.path.exists(newproject):
        message = "New Project creator: folder {} already exists in parent root {}.Try again with different name".format(
            newName, rootDir)
        runFFPrint(message, logger)
        return False

    # recover databases folder from applicationRoot
    rootpath = os.path.normpath(os.path.join(applicationRoot, ".."))
    databasepath = os.path.join(rootpath, "bin", "databases")
    if os.path.exists(databasepath) and not os.listdir(databasepath):
        message = "New Project creator: cannot find valid databases folder in {}".format(
            databasepath)
        runFFPrint(message, logger)
        return False

    # ok create the folder
    os.mkdir(newproject)
    # ok create subfolders coma-results and flyby-results
    os.mkdir(os.path.join(newproject, "coma-results"))
    os.mkdir(os.path.join(newproject, "flyby-results"))
    # copy databases in project folder
    shutil.copytree(databasepath, os.path.join(newproject, "databases"), True)
    message = "New Project creator: successfully created folder {}".format(
        newproject)
    runFFPrint(message, logger)

    return True


def checkValidProject(projectPath, logger=None):
    '''!
        Check structure of a pre-existent ComMoDE project folder.
        The method will check if the current folder is a valid project folder, that means
        it will check if the current subfolders are present:
        - database subfolder (previously inherited from installation files when creating a brand new project)
        - coma-results subfolder
        - flyby-results subfolder
        Moreover the following files must be present:
        - inputComa.xml User input collection file
        - inputFlyBy.xml User input collection file
        - .status_project internal file must be present
        @param projectPath target project folder path
        @param logger (optional) frontEnd.Logger.LoggerWidget instance to write meaningful messages
        @return True if the folder satifies the requisites, False otherwise
    '''
    if not os.path.exists(projectPath):
        message = "Check Valid Project: folder {} does not exist in your filesystem".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    if not os.path.exists(os.path.join(projectPath, "coma-results")):
        message = "Check Valid Project: coma-results folder does not exist in target project {}".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    if not os.path.exists(os.path.join(projectPath, "flyby-results")):
        message = "Check Valid Project: flyby-results folder does not exist in target project {}".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    if not os.path.exists(os.path.join(projectPath, "inputComa.xml")):
        message = "Check Valid Project: inputComa.xml User Input collection file does not exist in target project {}".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    if not os.path.exists(os.path.join(projectPath, "inputFlyBy.xml")):
        message = "Check Valid Project: inputFlyBy.xml User Input collection file does not exist in target project {}".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    if not os.path.exists(os.path.join(projectPath, ".status_project")):
        message = "Check Valid Project: internal file .status_project does not exist in target project {}".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    return True


def backupProject(sourceProjectPath, destProjectPath, logger=None):
    '''!
        Copy all contents of a source project folder into another, indipendent project folder.
        if the new destination project folder path already exists, the copy process will
        be aborted.
        @param sourceProjectPath original project folder path
        @param destProjectPath destination project folder
        @param logger (optional) frontEnd.Logger.LoggerWidget instance to write meaningful messages
        @return True for successfull backup, False otherwise
    '''
    if not os.path.exists(sourceProjectPath):
        message = "Backup Project: source project {} does not exist. Doing nothing ".format(
            sourceProjectPath)
        runFFPrint(message, logger)
        return False

    if os.path.exists(destProjectPath):
        message = "Backup Project: destination project {} already exists. Overwriting is prevented to avoid losing data. Pick up a new name ".format(
            destProjectPath)
        runFFPrint(message, logger)
        return False

    shutil.copytree(sourceProjectPath, destProjectPath, True)
    message = "Backup Project: successfully backup {} into {} ".format(os.path.basename(sourceProjectPath), destProjectPath)
    runFFPrint(message, logger)
    return True


def clearProject(projectPath, logger=None):
    '''!
        Given a project folder, clear its operational contents, namely:
        - all files inside coma-results subfolder
        - all files and sub-folders inside flyby-results
        - any *backup* files inside the folder
        The primal xml files and .status_project will remain untouched, and need to be
        reset accordingly outside this method
        @param projectPath path to ComMoDE project folder
        @param logger (optional) frontEnd.Logger.LoggerWidget instance to write meaningful messages
        @return True if clear was successfull, False otherwise
    '''
    if not os.path.exists(projectPath):
        message = "Clear Project: source project {} does not exist. Doing nothing ".format(
            projectPath)
        runFFPrint(message, logger)
        return False

    comarespath = os.path.join(projectPath, "coma-results")
    if os.path.exists(comarespath):
        shutil.rmtree(comarespath)
    os.mkdir(comarespath)

    flybyrespath = os.path.join(projectPath, "flyby-results")
    if os.path.exists(flybyrespath):
        shutil.rmtree(flybyrespath)
    os.mkdir(flybyrespath)

    listbackups = glob.glob(os.path.join(projectPath, "backupInput*.xml"))
    for item in listbackups:
        os.remove(item)

    return True
