import os
import time
import shutil

import PyComMoDE as core
import xml.etree.ElementTree as ET
import xml.dom.minidom as XDOM
from PySide2.QtCore import QObject, Signal, Slot, QCoreApplication, QEventLoop
#from PySide2.QtWidgets import QCoreApplication

class ComaSimulationWorker(QObject):
    '''!
        Defining a dedicated PySide2 QObject to wrap up calls to runComaSimulation. The class is needed for calling the 
        simulation process in a separate thread. 
        Current class exhibit a signal finishedComa to track down the runComaSimulation
    '''
    ## Public signal emitted every time the embedded runComaSimulation finishes
    finishedComa = Signal()
    ## Public signal to catch string messages 
    messageRaisedComa = Signal(str)
    ## Public signal to catch progress integer flag
    processAdvancesComa = Signal(int)
        
    def __init__(self, parent=None):
        '''!
        Class constructor
        '''
        QObject.__init__(self, parent=parent)
        ## store project folder instance passed in construction
        self.__projectFolder = None
        # Internal: store status of the thread 0-run ok, 1-run failed, 2-no response
        self.__status = 2

        ## internal flag, to customly abort  the process
        self.__abortFlag = False

    def __initializeWorker(self):
        '''!
            Put __status and __abortFlag members to their default state
        '''
        self.__status = 2
        self.__abortFlag = False

    def getStatus(self):
        '''!
            @return response of the processed coma, typically when the run has finished:  0-run ok, 1-run failed, 2-no response
        '''
        return self.__status

    def setProjectFolder(self, projectFolder):
        '''!
            Setting a valid ComMoDE project folder path. To be always done before running the simulation.
        '''
        self.__projectFolder = projectFolder

    def isAborting(self):
        '''!
            @return value of the internal __abortFlag
        '''
        return self.__abortFlag

    @Slot()
    def terminate(self):
        '''!
            class slot to customly abort an ongoing runProcessComa. Put internal variable __abortFlag to true
        '''
        self.__abortFlag = True

    def runProcessComa(self, printOnScreen=False):
        '''!
            Method to run the coma simulation.
            Require the full path of a valid project folder (organized as already explained
            in backEnd.fileFunctions.createNewProject method).
            The method will do the following:
            - absorbing primitive inputs from file "inputComa.xml" that should be located inside the projectFolder.
            - wipe out any content in projectFolder/coma-results subfolder
            - run the coma simulation performing calls to right methods of PyComMoDE core-API module. Results will be stored in
            projectFolder/coma-results
            It updates the internal member __status with an integer that can be  0-run ok, 1-run failed, 2-no response.
            Emit signals:
            - messageRaisedComa: to track usuful messages during process simulation. Suitable to be connected to an external logger 
            - processAdvancesComa: counting meaningul steps in simulation with an integer. Suitable to be connected to an external progressbar
            @param printOnScreen (optional) if set to True, force the method to write ongoing process messages on console

        '''
        # put the worker in no response mode and no abort 
        self.__initializeWorker()

        # first step check if projectFolder exists
        if not os.path.isdir(self.__projectFolder):
            message = "backEnd.ComaSimulationWorker : {} is not a valid directory".format(self.__projectFolder)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return

        #############################################################

        # second step check if inputComa.xml exists and it is readable
        inputComaFile = os.path.join(self.__projectFolder, "inputComa.xml")

        if not os.path.exists(inputComaFile):
            message = "backEnd.ComaSimulationWorker : cannot find {} ".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return
        #############################################################

        # third step - prepare coma-results
        dumpModelDirectoryPath = os.path.join(self.__projectFolder, "coma-results")
        if os.path.isdir(dumpModelDirectoryPath):
            for f in os.listdir(dumpModelDirectoryPath):
                os.remove(os.path.join(dumpModelDirectoryPath, f))
        else:
            os.mkdir(dumpModelDirectoryPath)

        message ="backEnd.ComaSimulationWorker : coma-results subfolder ready "
        self.messageRaisedComa.emit(message)
        if printOnScreen:
            print(message)

        #############################################################

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000) #without this any change of variables from terminated slot will be collect when process finishes.
        if  self.isAborting() :
            self.finishedComa.emit()
            return

        # fourth step reading input data from inputComa.xml
        inputTree = ET.parse(inputComaFile)
        inputElement = inputTree.getroot().find("ComaForm").find("ComaParameters")

        if not inputElement:
            message = "backEnd.ComaSimulationWorker : cannot read {}, missing ComaForm/ComaParameters sections".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return

        inputSizeDistroElement = inputElement.find("SizeDistribution")
        if not inputSizeDistroElement:
            message = "backEnd.ComaSimulationWorker : cannot find SizeDistribution section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return

        defaultXML = inputSizeDistroElement.findtext("default")
        NumberOfSizeBinsXML = inputSizeDistroElement.findtext("NumberOfSizeBins")
        PowerLawIndexXML = inputSizeDistroElement.findtext("PowerLawIndex")
        SmallestSizeXML = inputSizeDistroElement.findtext("SmallestSize")
        LargestSizeXML = inputSizeDistroElement.findtext("LargestSize")

        if not defaultXML or not NumberOfSizeBinsXML or not PowerLawIndexXML or not SmallestSizeXML or not LargestSizeXML:
            message = "backEnd.ComaSimulationWorker : at least one of SizeDistribution default, NumberOfSizeBins, PowerLawIndex, SmallestSize and LargestSize is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        if defaultXML.strip() == "True":
            # default distribution - Nico's feedback 13/4/2023(mail)
            number_of_bins = int(18)
            power_law_index = float(3.8)
            smallestsize = int(-7)
            largestsize = int(-1)
        else:
            number_of_bins = int(NumberOfSizeBinsXML)
            power_law_index = float(PowerLawIndexXML)
            smallestsize = int(SmallestSizeXML)
            largestsize  = int(LargestSizeXML)

        # post-processing values
        power_law_index_passed = -1.*power_law_index
        #this inversion is on purpose - from Nico integration test
        lower_size_limit = largestsize
        higher_size_limit = smallestsize

        inputScatteringElement = inputElement.find("ScatteringProperties")
        if not inputScatteringElement:
            message = "backEnd.ComaSimulationWorker : cannot find ScatteringProperties section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        MaterialXML = inputScatteringElement.findtext("Material")
        HeliocentricDistanceXML = inputScatteringElement.findtext("HeliocentricDistance")
        PhaseAngleXML = inputScatteringElement.findtext("PhaseAngle")

        if not MaterialXML or not HeliocentricDistanceXML or not PhaseAngleXML:
            message = "backEnd.ComaSimulationWorker : at least one of ScatteringProperties Material, HeliocentricDistance, PhaseAngle is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return

        # materials aka species_index
        # 0 for water ice
        # 1 for enstatite
        # 2 for astro-silicate
        # 3 for troilite
        MaterialXML = MaterialXML.strip()
        species_index = int(0)
        if MaterialXML == "enstatite":
            species_index = int(1)
        elif MaterialXML == "astro-silicate":
            species_index = int(2)
        elif MaterialXML == "troilite":
            species_index = int(3)
        heliocentric_distance = float(HeliocentricDistanceXML)
        phase_angle = int(PhaseAngleXML)

        inputDustProdElement = inputElement.find("DustProduction")
        if not inputDustProdElement:
            message = "backEnd.ComaSimulationWorker : cannot find DustProduction section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        UseAfrhoXML = inputDustProdElement.findtext("UseAfrho")
        DustProductionRateXML = inputDustProdElement.findtext("DustProductionRate")
        AfrhoXML = inputDustProdElement.findtext("Afrho")

        if not UseAfrhoXML or not DustProductionRateXML or not AfrhoXML:
            message = "backEnd.ComaSimulationWorker : at least one of DustProdution UseAfrho, DustProductionRate and Afrho is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        if UseAfrhoXML.strip() == "True":
            afrho_input = float(AfrhoXML)
            qd_estimation_switch = int(1)
            qd_input = float(0.0)
        else:
            qd_input = float(DustProductionRateXML)
            qd_estimation_switch = int(0)
            afrho_input = float(0.0)

        inputDustOutflowElement = inputElement.find("DustOutflowVelocity")
        if not inputDustOutflowElement:
            message = "backEnd.ComaSimulationWorker : cannot find DustOutflowVelocity section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        GasProductionRateXML = inputDustOutflowElement.findtext("GasProductionRate")
        DustOutflowExpertXML = inputDustOutflowElement.findtext("expert")
        NucleusRadiusXML = inputDustOutflowElement.findtext("NucleusRadius")
        NucleusSurfaceTemperatureXML = inputDustOutflowElement.findtext("NucleusSurfaceTemperature")
        DustBulkDensityXML = inputDustOutflowElement.findtext("DustBulkDensity")

        if not GasProductionRateXML or not DustOutflowExpertXML or not NucleusRadiusXML or not NucleusSurfaceTemperatureXML or not DustBulkDensityXML:
            message = "backEnd.ComaSimulationWorker : at least one of DustOutflowVelocity GasProductionRate, expert, NucleusRadius, NucleusSurfaceTemperature and DustBulkDensity is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        Q_gas = float(GasProductionRateXML)
        if DustOutflowExpertXML.strip() == "True":
            radius_N = float(NucleusRadiusXML)
            nucleus_temp = float(NucleusSurfaceTemperatureXML)
            particle_density = float(DustBulkDensityXML)
        else:
            # pick default values
            radius_N = float(5000.0)
            nucleus_temp = float(314.0)
            particle_density = float(533.0)

        inputDNAsymmetryElement = inputElement.find("DayNightAsymmetry")
        if not inputDNAsymmetryElement:
            message = "backEnd.ComaSimulationWorker : cannot find DayNightAsymmetry section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        DNAsymmetryActiveXML = inputDNAsymmetryElement.findtext("active")
        AngularAsymmetryFunctionXML = inputDNAsymmetryElement.findtext("AngularAsymmetryFunction")
        CosineAmplitudeXML = inputDNAsymmetryElement.findtext("CosineAmplitude")

        if not DNAsymmetryActiveXML or not AngularAsymmetryFunctionXML or not CosineAmplitudeXML:
            message = "backEnd.ComaSimulationWorker : at least one of DayNightAsymmetry active, AngularAsymmetryFunction, CosineAmplitude is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        if DNAsymmetryActiveXML.strip() == "True":
            # 1 means asymmetry -> numerical model
            day_night_asym_switch = int(1)
            # default values of angular_dependency_index and cosine_scaling are passed on even if there is no angular dependence (or no cosine-law)
            # inputs are
            # 1 - cosine (this also needs a cosine_scaling that is between 0 and 1)
            # 2 - gaussian
            # 3 - half-maxwell
            angular_dependency_index = int(1)
            if AngularAsymmetryFunctionXML.strip() == "gaussian":
                angular_dependency_index = int(2)
            elif AngularAsymmetryFunctionXML.strip() == "half-maxwell":
                angular_dependency_index = int(3)

            if angular_dependency_index == 0:
                cosine_scaling = float(CosineAmplitudeXML)
            else:
                cosine_scaling = float(0.0)
        else:
            # 0 means no asymmetry -> fountain model
            day_night_asym_switch = int(0)
            # default values of angular_dependency_index and cosine_scaling are passed on even if there is no angular dependence (or no cosine-law)
            # inputs are
            # 0 - no dependency (only used for comparison with the fountain model - i.e. verification, validation purposes)
            angular_dependency_index = int(0)
            cosine_scaling = float(0.0)

        inputRotationVarElement = inputElement.find("RotationalVariability")
        if not inputRotationVarElement:
            message = "backEnd.ComaSimulationWorker : cannot find RotationalVariability section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        RotationVarActiveXML = inputRotationVarElement.findtext("active")
        RotationalPeriodXML = inputRotationVarElement.findtext("RotationalPeriod")
        RotationalAmplitudeXML = inputRotationVarElement.findtext("RotationalAmplitude")

        if not RotationVarActiveXML or not RotationalPeriodXML or not RotationalAmplitudeXML:
            message = "backEnd.ComaSimulationWorker : at least one of RotationalVariability active, RotationalPeriod, RotationalAmplitude is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        if RotationVarActiveXML.strip() == "True":
            # rotation_switch = 1 -> amplitude_rot between 0 and 1 and period_rot (passed on in seconds) are user inputs
            rotational_switch = int(1)
            period_rot_in_hours = float(RotationalPeriodXML)
            amplitude_rot_percentage = int(RotationalAmplitudeXML)
        else:
            # rotation_switch = 0 means no rotation -> default parameters amplitude_rot = 0 and period_rot = 0 are passed
            rotational_switch = int(0)
            period_rot_in_hours = float(0.0)
            amplitude_rot_percentage = int(0.0)

        # post-processed variables
        period_rot = period_rot_in_hours * float(3600.0)
        amplitude_rot = float(amplitude_rot_percentage)*float(0.01)

        inputGridElement = inputElement.find("GridResolution")
        if not inputGridElement:
            message = "backEnd.ComaSimulationWorker : cannot find GridResolution section in {}".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        MaximumDistanceXML = inputGridElement.findtext("MaximumDistance")
        ResolutionXML = inputGridElement.findtext("Resolution")

        if not MaximumDistanceXML or not ResolutionXML:
            message = "backEnd.ComaSimulationWorker : at least one of GridResolution MaximumDistance and Resolution is missing in {}. Aborting".format(inputComaFile)
            self.messageRaisedComa.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        max_distance = float(MaximumDistanceXML)
        cellsize = int(ResolutionXML)
        #############################################################

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.finishedComa.emit()
            return

        # fix the total number of steps for moving progress bar: it will be  the number of total bins + 6 preprocessing steps + one final step for header writing
        totStep = 7 + number_of_bins
        countStep = 0
        #############################################################

        # calculating bins and particle densities
        messageCumulate = ""
        try:
            size_bins = core.get_size_bins(number_of_bins, lower_size_limit, higher_size_limit)
            particle_densities = core.calc_particle_density(size_bins, particle_density)

            message = "...coma module calculating bins"
            if printOnScreen :
                print(message)
            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1
        except:
            message="backEnd.ComaSimulationWorker: ...bins computation Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        #############################################################

        # calculating terminal velocities
        try:
            terminal_velocities = core.get_terminal_v(size_bins, particle_densities, Q_gas, radius_N, nucleus_temp)

            message = "...coma module calculating terminal velocities"
            if printOnScreen :
                print(message)

            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1
        except:
            message = "backEnd.ComaSimulationWorker: ...terminal velocities computation Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        #############################################################

        backupfolder = os.getcwd()
        os.chdir(self.__projectFolder)
        try:
            Qdata = core.get_scattering_properties(size_bins, phase_angle, species_index)
            message = "...coma module retrieving scattering properties data from database"
            if printOnScreen :
                print(message)

            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1
            os.chdir(backupfolder)
        except:
            message = "backEnd.ComaSimulationWorker: ...scattering properties retrieval Failed!"
            self.messageRaisedComa.emit(messageCumulate+message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
 
            os.chdir(backupfolder)
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        #############################################################
        # estimate the production rate
        try:
            qd = qd_input
            if qd_estimation_switch == 1:
                qd = core.estimate_production_rate(afrho_input, power_law_index_passed, size_bins, particle_densities, terminal_velocities, Qdata[2], Qdata[3])
                message = "...coma module performing estimation of dust production rate"
            else:
                message = "...setting User defined dust production rate"

            if printOnScreen :
                print(message)

            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1
        except:
            message = "backEnd.ComaSimulationWorker: ...dust production rate setting or computation Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        #############################################################

        # estimate beta
        try:
            beta = core.calculate_beta(heliocentric_distance, size_bins, particle_densities, Qdata[2], Qdata[3])
            message= "...coma module calculating beta"
            if printOnScreen :
                print(message)

            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1

        except:
            message = "backEnd.ComaSimulationWorker: ...beta computation Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        #############################################################

        # create array of bin-size dependent qd
        try:
            size_dependent_qd = core.get_size_dependent_qd(qd, power_law_index_passed, size_bins, particle_densities, terminal_velocities)

            message = "...coma module calculating size-dependent production rates"
            if printOnScreen :
                print(message)

            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1
 
        except:
            message = "backEnd.ComaSimulationWorker: ..size-dependent production rates computation Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        ## dump message Cumulate and update progress bar
        self.messageRaisedComa.emit(messageCumulate.strip())
        self.processAdvancesComa.emit(int(countStep*100/totStep))
        messageCumulate = ""

        #############################################################

        # loop on bins and run the full numerical model
        try:
            # initialize mesh resolution and data list on mesh
            #coreApi methods expect to have data on meters not Km.
            cellsize = cellsize*1000
            gridsize = int(round(max_distance*1000/cellsize))
            #check if this 60000000.0 is hard-coded here.
            gridsize_apex = int(round(60000000.0/cellsize))
            gridsize_x = gridsize + gridsize_apex
            gridsize_y = gridsize

            total_afrho_calculated = float(0.0)

            # instantiate python side the Matrix container of coma resulting data.
            for i in range(number_of_bins):
                afrho = float(0.0)
                #initialize the coma data chunk per bin.
                coma_data = core.DoubleMatrix()
                core.initializeGrid(gridsize_x, gridsize_y, coma_data)

                if str(day_night_asym_switch) == "1":
                    check, afrho = self.run_numerical_modelX(coma_data, gridsize_x, gridsize_y, cellsize, gridsize_apex,
                                                    beta[i], int(i), size_bins[i], terminal_velocities[i],
                                                    particle_densities[i], size_dependent_qd[i], Qdata[2][i],
                                                    Qdata[3][i], amplitude_rot, period_rot, angular_dependency_index,
                                                    cosine_scaling, dumpModelDirectoryPath,printOnScreen)
                    if check == False :
                        #check false means the sub-job was aborted by the user.
                        self.finishedComa.emit()
                        return
                
                else:
                    afrho = core.run_fountain(coma_data, gridsize_x, gridsize_y, cellsize, gridsize_apex,
                                            beta[i], size_dependent_qd[i], size_bins[i], terminal_velocities[i], int(i),
                                            particle_densities[i], Qdata[2][i], Qdata[3][i], dumpModelDirectoryPath)

                message = "...coma module simulated bin {0} ({1}/{2}) -> afrho calculated is {3}".format(i, i+1, number_of_bins, "{:e}".format(afrho))
                
                total_afrho_calculated = total_afrho_calculated + afrho

                if printOnScreen :
                    print(message)

                countStep = countStep + 1

                self.messageRaisedComa.emit(message)
                self.processAdvancesComa.emit(int(countStep*100/totStep))
 
                #abort check breakpoint
                QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
                if  self.isAborting() :
                    self.finishedComa.emit()
                    return

        ##############################################################

            message = "...total afrho calculated {}".format("{:e}".format(total_afrho_calculated))
            if printOnScreen :
                print(message)
            messageCumulate = messageCumulate + message +"\n"
            countStep = countStep + 1
            
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.processAdvancesComa.emit(int(countStep*100/totStep))
            messageCumulate = ""

        except:
            message = "backEnd.ComaSimulationWorker: ...model evaluation Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 


        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.messageRaisedComa.emit(messageCumulate.strip())
            self.finishedComa.emit()
            return

        # write input coma data on header file for restore purpose of trajectory module.
        try:
            core.write_header(gridsize_x, gridsize_y, cellsize, gridsize_apex, particle_density, lower_size_limit,
                            higher_size_limit, power_law_index_passed, Q_gas, radius_N, nucleus_temp, species_index, heliocentric_distance,
                            day_night_asym_switch, angular_dependency_index, cosine_scaling, rotational_switch,  amplitude_rot, period_rot,
                            qd_estimation_switch, qd, afrho_input, size_bins, terminal_velocities, beta, Qdata, phase_angle,
                            size_dependent_qd, total_afrho_calculated, dumpModelDirectoryPath)

            message = "...dumping coma data header header.txt in coma-results"
            if printOnScreen :
                print(message)
            messageCumulate = messageCumulate + message + "\n"
            countStep = countStep + 1

        except:
            message = "backEnd.ComaSimulationWorker: ...dumping coma data header Failed!"
            self.messageRaisedComa.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedComa
            self.finishedComa.emit()
            return 

        #############################################################

        message = "...coma simulation successfull on {}".format(time.strftime("%Y%m%d-%H%M%S"))
        if printOnScreen:
            print(message)
        messageCumulate = messageCumulate + message + "\n"
        countStep = countStep + 1

        self.messageRaisedComa.emit(messageCumulate.strip())
        self.processAdvancesComa.emit(int(countStep*100/totStep))
        messageCumulate = ""

        #changeStatus
        self.__status = 0
        # emit finishedComa
        self.finishedComa.emit()
        return 

    def run_numerical_modelX(self, coma_data, gridsize_x, gridsize_y, cellsize, gridsize_apex,
                             beta, size_index, particle_size, v0, particle_density, size_dependent_qd, Q_sca,
                            phase_function, amplitude_rot, period_rot, angular_dependency_index,
                            cosine_scaling, dumpPath, printOnScreen=False) : 
        '''!
            wrapped utility to twin method core::run_numerical_model, with the purpose to obtain
            finer grain control of the simulation process.
            For input description see the core twin function. The only thing unused here is the final coma_test, since
            useless, and a boolean to activate the printOnScreen 
            @return True is the sub-job was performed, False if aborted by the extern plus the afrho value calculated
        '''

        afrho = float(0.0)

        #start with the initial sampling to estimate the coma emission.
        core.numerical_model_sampling(gridsize_x, gridsize_y, gridsize_apex, cellsize, beta, size_index,
        particle_size, v0, particle_density, size_dependent_qd, Q_sca, phase_function, amplitude_rot,
        period_rot, angular_dependency_index, cosine_scaling, coma_data)

        message = "......sampling performed"
        if printOnScreen :
            print(message)
        self.messageRaisedComa.emit(message)
        
        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.finishedComa.emit()
            return False, afrho

        #now the expensive part: the smoothing.

        #precalculate the weights
        weights1 = core.DoubleMatrix()
        weights2 = core.DoubleMatrix()
        weights3 = core.DoubleMatrix()
        weights4 = core.DoubleMatrix()

        smoothing_radius_vec = core.prepareSmoothingWeights(cellsize, weights1,weights2, weights3, weights4)

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.finishedComa.emit()
            return False, afrho

        # looping on grid matrix dimensions
        counter = int(0)
        grid_smooth = core.DoubleMatrix()
        for x in range(gridsize_x) : 
            temp = core.DoubleVector()
            for y in range(gridsize_y) :
                value = core.smoothingValue(int(x),int(y),gridsize_x, gridsize_y, gridsize_apex, cellsize, 
                    coma_data, smoothing_radius_vec, weights1, weights2, weights3, weights4)    

                temp.append(value)
                counter = counter + 1

                if counter % 1000000 == 0 :
                    message = "......currently smoothing"
                    if printOnScreen :
                        print(message)
                    self.messageRaisedComa.emit(message)

                #abort check breakpoint
                QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
                if  self.isAborting() :
                    self.finishedComa.emit()
                    return False, afrho

            #out of y loop, time to reverse temp in the matrix
            grid_smooth.append(temp)
        # out of for loops

        message = "......smoothing performed"
        if printOnScreen :
            print(message)
        self.messageRaisedComa.emit(message)
        
        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.finishedComa.emit()
            return False, afrho

        # now calculate afrho.
        afrho = core.calc_afrho(grid_smooth,gridsize_x,gridsize_y,cellsize,gridsize_apex,Q_sca,phase_function, particle_size)
        # and write on file the coma data.

        name = "dust_density_grid"+str(size_index)+".dat"
        filename = os.path.join(dumpPath, name)
        core.write_file(grid_smooth,filename)

        return True, afrho


class TrajectoryInputStruct:
    '''!
        C-struct like container to absorb data for a single trajectory from inputFlyBy.xml
    '''
    def __init__(self):
        '''!
            Constructor
        '''
        ## hosting custom name of trajectory
        self.name = str('')
        ## hosting spacecraft velocity 
        self.velocity = float(0.0)
        ## hosting phase angle
        self.phase = float(0.0)
        ## hosting impact
        self.impact = float(0.0)
        ## hosting altitude angle
        self.altitude=float(0.0)
        ## hosting spacecraft frontal area
        self.area= float(0.0)
        ## hosting status of trajectory (already evaluated or not)
        self.status = bool(False)

class FlyByProcessWorker(QObject):
    '''!
        Defining a dedicated PySide2 QObject to wrap up calls to run fly by trajectories post-processing. 
        The class is needed for calling the processFlyBy in a separate thread. 
        Current class exhibit a signal finishedFlyBy to track down status of the process
    '''
    ## Public signal emitted every time the embedded flybyProcess finishes
    finishedFlyBy = Signal()
    ## Public signal to catch string messages 
    messageRaisedFlyBy = Signal(str)
    ## Public signal to catch progress bar integer flag
    processAdvancesFlyBy = Signal(int)
        
    def __init__(self, parent=None):
        '''!
        Class constructor
        '''
        QObject.__init__(self, parent=parent)
        ## store project folder instance passed in construction
        self.__projectFolder = None
        # Internal: store status of the thread 0-run ok, 1-run failed, 2- no response
        self.__status = 2
        #internal flag to premature abort the job.
        self.__abortFlag = False

    def __initializeWorker(self):
        '''!
            Put __status to its default state
        '''
        self.__status = 2
        self.__abortFlag = False

    def getStatus(self):
        '''!
            @return response of the processed trajectories, typically when the run has finished:  
            0-run ok, 1-run failed, 2-no response
        '''
        return self.__status

    def isAborting(self):
        '''!
            @return value of the internal __abortFlag
        '''
        return self.__abortFlag

    @Slot()
    def terminate(self):
        '''!
            class slot to customly abort an ongoing flyby process. Put internal variable __abortFlag to true
        '''
        self.__abortFlag = True

    def setProjectFolder(self, projectFolder):
        '''!
            Setting a valid ComMoDE project folder path. To be always done before running the process.
        '''
        self.__projectFolder = projectFolder

    def __printXMLTreeToFile(self, xmlElement, filepath):
        '''!
            internal utility to dump trajectory status modifications done during processFlyBy to the inputFlyBy.xml
            @param xmlElement tree of inputFlyBy.xml
            @param filepath full path to inputFlyBy.xml
        '''
        rstring = ET.tostring(xmlElement, 'utf-8')
        xdom = XDOM.parseString(rstring)
        towritestring = xdom.toprettyxml(indent="    ")
        with open(filepath, "w") as file:
            file.write(towritestring)

    def processFlyBy(self, printOnScreen=False):
        '''!
            Method to run the flyby process.
            Require the full path of a valid project folder (organized as already explained
            in backEnd.fileFunctions.createNewProject method).
            The method will do the following:
            - absorbing primitive inputs from file "inputFlyBy.xml" that should be located inside the projectFolder.
            - scan how may trajectories must be processed. 
            - Typically each trajectory should have a dedicated folder inside flyby-results subfolder. If not present, the 
            process will create it and run the processing for it. If present, the method will check the status of trajectory
            reported in the inputFlyBy.xml relative section. If status is true, the processing of such trajectory will be 
            considered already done, and it will be safely skipped   
            - Processing result data trajectory by trajectory calling the right methods of PyComMoDE core-API module. 
            Results will be stored in  projectFolder/flyby-results/current-name-of-the-trajectory-processed
            - if a processing fails for a particula trajectory, it's dedicated folder will be erased
            It updates the internal member __status with an integer that can be  0-run ok, 1-run failed, 2-no response.
            Job will be considered ok (0), if at least 1 trajectory from those prescribed in inputFlyBy.xml is successfully processed.
            Emit signals:
            - messageRaisedFlyBy: to track usuful messages during process simulation. Suitable to be connected to an external logger 
            - processAdvancesFlyBy: counting meaningul steps in simulation with an integer. Suitable to be connected to an external progressbar
            @param printOnScreen (optional) if set to True, force the method to write ongoing process messages on console

        '''
        # put the worker in no response mode
        self.__initializeWorker()

        # first step check if projectFolder exists
        if not os.path.isdir(self.__projectFolder):
            message = "backEnd.FlyByProcessWorker : {} is not a valid directory".format(self.__projectFolder)
            self.messageRaisedFlyBy.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedFlyBy
            self.finishedFlyBy.emit()
            return

        #############################################################

        # second step check if inputFlyBy.xml exists and it is readable
        inputFlyByFile = os.path.join(self.__projectFolder, "inputFlyBy.xml")

        if not os.path.exists(inputFlyByFile):
            message = "backEnd.FlyByProcessWorker : cannot find {} ".format(inputFlyByFile)
            self.messageRaisedFlyBy.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedFlyBy
            self.finishedFlyBy.emit()
            return
        #############################################################
        # third step - check flyby-results
        flybyfolder = os.path.join(self.__projectFolder, "flyby-results")
        if not os.path.isdir(flybyfolder):
            os.mkdir(flybyfolder)
        #############################################################

        # fourth step reading input data from inputFlyBy.xml
        inputTree = ET.parse(inputFlyByFile).getroot()
        inputElement = inputTree.find("FlyByForm").find("FlyByTrajectories")

        if not inputElement:
            message = "backEnd.FlyByProcessWorker : empty or missing FlyByForm/FlyByTrajectories section in {}".format(inputFlyByFile)
            self.messageRaisedFlyBy.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedFlyBy
            self.finishedFlyBy.emit()
            return

        listTrajectories = []
        for tElement in inputElement :
        
            data = TrajectoryInputStruct()
            data.name = tElement.tag.strip()
            if not data.name : 
                message = message + "backEnd.FlyByProcessWorker :  found unnamed trajectory section in xml file {}. Aborting".format(data.name, inputFlyByFile)
                self.messageRaisedFlyBy.emit(message)
                if printOnScreen:
                    print(message)
                #changeStatus
                self.__status = 1
                # emit finishedFlyBy
                self.finishedFlyBy.emit()
                return 

            velXML = tElement.findtext("SpacecraftVelocity")
            phaseXML = tElement.findtext("PhaseAngleOfApproach")
            impactXML = tElement.findtext("ImpactParameter")
            altitudeXML = tElement.findtext("AltitudeAngle")
            areaXML = tElement.findtext("SpacecraftFrontalArea")
            statusXML= tElement.findtext("Status")
            if not velXML or not phaseXML or not impactXML or not altitudeXML or not areaXML or not statusXML:
                message = "backEnd.FlyByProcessWorker : at least one of Trajectory SpacecraftVelocity, PhaseAngleOfApproach, "
                message = message + "ImpactParameter, AltitudeAngle, SpacecraftFrontalArea and Status is missing "
                message = message + "for trajectory {} in xml file {}. Aborting".format(data.name, inputFlyByFile)
                self.messageRaisedFlyBy.emit(message)
                if printOnScreen:
                    print(message)
                #changeStatus
                self.__status = 1
                # emit finishedFlyBy
                self.finishedFlyBy.emit()
                return 

            data.velocity= float(velXML)
            data.phase = float(phaseXML)
            data.impact = float(impactXML)
            data.altitude = float(altitudeXML)
            data.area = float(areaXML)
            data.status = statusXML.strip() == 'True'
            listTrajectories.append(data)
        #############################################################

        # cross check trajectories data with folder already present in flyby-results
        listValidFolders = []
        for item in listTrajectories : 
            itemfolder = os.path.join(flybyfolder, item.name)
            if os.path.isdir(itemfolder) :
                if not item.status:
                    shutil.rmtree(itemfolder)
                    os.mkdir(itemfolder)
            else:
                os.mkdir(itemfolder)
                item.status = False
                inputElement.find(item.name).find("Status").text = 'False'

            listValidFolders.append(itemfolder)

        #cross check any other folder present in flyby-results, if not in listValidFolders purge it
        for item in os.listdir(flybyfolder):
            itempath= os.path.join(flybyfolder, item)
            if os.path.isdir(itempath) and not itempath in listValidFolders: 
                shutil.rmtree(itempath)

        listValidFolders.clear()

        #abort check breakpoint
        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
        if  self.isAborting() :
            self.__printXMLTreeToFile(inputTree, inputFlyByFile)
            self.finishedFlyBy.emit()
            return

        
        #############################################################
        #restore cometary simulation data

        try: 
            comaresfolder= os.path.join(self.__projectFolder, "coma-results")
            chunk = core.read_header(comaresfolder)

            message="...restored coma simulation results"
            self.messageRaisedFlyBy.emit(message)
            if printOnScreen:
                print(message)

        except:
            message="backEnd.FlyByProcessWorker: restore of coma simulation results Failed!"
            self.messageRaisedFlyBy.emit(message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedFlyBy
            self.finishedFlyBy.emit()

        # fix the total number of steps for moving progress bar: it will be total number of trajctories * number of bins
        totStep = len(listTrajectories) * chunk.number_of_bins + 1
        countStep = 0
        #############################################################

        # loop on each trajectories->each bin and evaluate data results file for each of them
        # if a trajectory has status true, skip it and considered it done
        messageCumulate = ""

        #re-initialize the bins distribution 
        try: 
            size_bins = core.get_size_bins(chunk.number_of_bins, chunk.lower_size_limit, chunk.higher_size_limit)
            
            message="...re-initialized coma size bins distribution"
            if printOnScreen:
                print(message)
            messageCumulate= messageCumulate + message + "\n"
            countStep = countStep+1
        except:
            message="backEnd.FlyByProcessWorker: ...coma size bins re-initialization Failed!"
            self.messageRaisedFlyBy.emit(messageCumulate + message)
            if printOnScreen:
                print(message)
            #changeStatus
            self.__status = 1
            # emit finishedFlyBy
            self.finishedFlyBy.emit()
            return 

        #############################################################

        # loop on each trajectory and each bins 
        for trajectory in listTrajectories:
                
            dumpResultsFolder = os.path.join(flybyfolder, trajectory.name)
            
            if trajectory.status : 
                message = "... trajectory {} already computed. Skipping it".format(trajectory.name)
                if printOnScreen :
                    print(message)
        
                countStep = countStep + chunk.number_of_bins

                self.messageRaisedFlyBy.emit(message)
                self.processAdvancesFlyBy.emit(int(countStep*100/totStep))

            else:
                #initialize total impact mass
                total_mass = float(0.0)
                impact_mass_per_bin = core.DoubleVector()
                impact_mass_per_bin[:] =[float(0.)]*chunk.number_of_bins 
                
                try: 
                    for i in range(chunk.number_of_bins):
                        
                        ith_coma_data = core.read_grid(chunk.gridsize_x,chunk.gridsize_y, int(i), comaresfolder)
                            
                        impact_mass_per_bin[i] = core.mass_along_trajectory(
                            ith_coma_data, chunk.gridsize_x, chunk.gridsize_y, 
                            chunk.particle_density, size_bins[i], chunk.cellsize, 
                            chunk.gridsize_apex, int(i), trajectory.impact, trajectory.phase, 
                            trajectory.altitude, trajectory.velocity, trajectory.area, 
                            dumpResultsFolder)

                        message = "... {0}:  found for bin {1} an impact mass of {2} Kg along the whole trajectory".format(trajectory.name, i, "{:e}".format(impact_mass_per_bin[i]))
                        total_mass = total_mass + impact_mass_per_bin[i]
                        if printOnScreen :
                            print(message)

                        countStep = countStep + 1

                        self.messageRaisedFlyBy.emit(message)
                        self.processAdvancesFlyBy.emit(int(countStep*100/totStep))

                        #abort check breakpoint inside trajectory loop
                        QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
                        if  self.isAborting() :
                            self.__printXMLTreeToFile(inputTree, inputFlyByFile)
                            self.finishedFlyBy.emit()
                            return

                    #outside for loop        
                    #put the trajectory status to True
                    core.write_header_spacecraft(
                        trajectory.impact,trajectory.phase,trajectory.altitude, trajectory.velocity, 
                        trajectory.area, chunk.number_of_bins, size_bins, impact_mass_per_bin, 
                        total_mass, dumpResultsFolder)

                    trajectory.status = True
                    inputElement.find(trajectory.name).find("Status").text = 'True'

                    message = "...trajectory {} Done: total impact mass calculated {} Kg".format(trajectory.name, "{:e}".format(total_mass))
                    if printOnScreen :
                        print(message)
                    self.messageRaisedFlyBy.emit(message)
                        
                except:
                    message="backEnd.FlyByProcessWorker: ...trajectory {} processing Failed!".format(trajectory.name)
                    if printOnScreen:
                        print(message)

                    #emitting stuff
                    self.messageRaisedFlyBy.emit(message)
                    countStep = countStep - countStep%chunk.number_of_bins + chunk.number_of_bins
                    self.processAdvancesFlyBy.emit(int(countStep*100/totStep))

                    #put the trajectory status to false
                    trajectory.status = False
                    inputElement.find(trajectory.name).find("Status").text = 'False'

                    continue

            #abort check breakpoint inside trajectory loop
            QCoreApplication.processEvents(QEventLoop.AllEvents, 3000)
            if  self.isAborting() :
                self.__printXMLTreeToFile(inputTree, inputFlyByFile)
                self.finishedFlyBy.emit()
                return
        ##############################################################

        #follow up operations. 
        #visit all false status trajectories and clean up their folders.
        #if at least one folder survived put the general status to true.
        #alter inputFlyBy.xml with final status and rewrite it.

        countT = len(listTrajectories)
        #reuse already open inputElement of inputTree
        for trajectory in listTrajectories:

            if not trajectory.status:
                shutil.rmtree(os.path.join(flybyfolder, trajectory.name))
                countT = countT - 1

            statXML = inputElement.find(trajectory.name).find("Status")
            statXML.text = str(trajectory.status)

        #print inputTree again on file
        self.__printXMLTreeToFile(inputTree, inputFlyByFile)

        #update __status
        if countT > 0:
            self.__status = 0
            message = "...successfully processed {} trajectories on {}".format(countT, time.strftime("%Y%m%d-%H%M%S"))
        else:
            self.__status = 1
            message = "...trajectories processing Failed on {}".format(time.strftime("%Y%m%d-%H%M%S"))

        if printOnScreen:
            print(message)
        
        self.messageRaisedFlyBy.emit(message)
        self.processAdvancesFlyBy.emit(int(countStep*100/totStep))

        # emit finishedFlyBy
        self.finishedFlyBy.emit()
        return 

