from PySide2.QtWidgets import QWidget, QToolBox
from PySide2.QtWidgets import QPushButton, QGroupBox, QVBoxLayout
from PySide2.QtWidgets import QHBoxLayout, QFrame, QCheckBox
from PySide2.QtWidgets import QLabel, QSpinBox, QDoubleSpinBox
from PySide2.QtWidgets import QGridLayout, QComboBox, QSlider


from PySide2.QtCore import Qt

import xml.etree.ElementTree as ET

class ComaParameters(QWidget):
    '''!
        Collection of input widgets to instrument the coma simulation. The following class is 
        organized in a foldable form contaning 7 sub-sections marked with name suffix "Input" 
        Each sub-section can be expanded through selection, then filled, once at the time. 
        Subsections available are referred to : 
        - input to set the particle size distribution (ParticleSizeDistroInput)
        - input to define dust scattering properties (ScatteringInput)
        - input to define the dust production rate (GlobalDustProductionInput)
        - input to define Outflow velocity condition (OutflowVelocityInput)
        - input to define the Day/Night asymmetry condition (DayNightAsymmetryInput)
        - input to define the Rotational variability (RotationalPeriodInput)
        - input to define the computational Mesh Resolution (GridResolutionInput)

    '''

    def __init__(self):
        '''!
            Constructor.
        '''
        super().__init__()

        # set up Coma Simulation parameters
        ## widget form to setup the Particle Size Distribution set of inputs
        self.particleSizeDistro = ParticleSizeDistroInput()

        ## widget form to setup the Scattering Properties set of inputs
        self.scattering = ScatteringInput()

        ## widget form to setup the Dust Production Rate set of inputs
        self.dustProduction = GlobalDustProductionInput()

        ## widget form to setup the Dust Outflow Velocity set of inputs
        self.outflowVelocity = OutflowVelocityInput()

        ## widget form to setup the Day/Night Asymmetry set of inputs
        self.dayNightAsymmetry = DayNightAsymmetryInput()

        ## widget form to setup the Rotational Variability set of inputs
        self.rotationalVariability = RotationalPeriodInput()

        ## widget form to setup the Grid Resolution set of inputs
        self.gridResolution = GridResolutionInput()

        ## Define the foldable object (Toolbox), containing the 7 subsections. See general class doxy
        self.comaToolbox = QToolBox()
        self.comaToolbox.addItem(
            self.particleSizeDistro, "Particle Size Distribution")
        self.comaToolbox.addItem(self.scattering, "Scattering Properties")
        self.comaToolbox.addItem(self.dustProduction, "Dust production")
        self.comaToolbox.addItem(self.outflowVelocity, "Dust outflow velocity")
        self.comaToolbox.addItem(self.dayNightAsymmetry, "Day/Night Asymmetry")
        self.comaToolbox.addItem(
            self.rotationalVariability, "Rotational variability")
        self.comaToolbox.addItem(self.gridResolution, "Grid Resolution")

        ## Button for run coma. The button is simply a redundancy that triggers the Run Coma Simulation action in Run menu
        self.runComaButton = QPushButton("Run Coma")
        self.runComaButton.setToolTip("Start cometary simulation")
        ## Button for abort coma simulation. The button is simply a redundancy that triggers the Stop Coma Simulation action in Run menu
        self.stopComaButton = QPushButton("Stop")
        self.stopComaButton.setToolTip("Abort cometary simulation")

        self.stopComaButton.setEnabled(False)

        # Internal layout to hold together executable buttons
        layout_buttons = QHBoxLayout()
        layout_buttons.setContentsMargins(2, 2, 2, 2)
        layout_buttons.addWidget(self.runComaButton)
        layout_buttons.addWidget(self.stopComaButton)

        # Internal widget to host the layout_buttons
        widget_buttons = QWidget()
        widget_buttons.setLayout(layout_buttons)

        # Internal layout to hold together coma input forms and the buttons compound widget
        layout_coma = QVBoxLayout()
        layout_coma.setContentsMargins(2, 2, 2, 2)
        layout_coma.addWidget(self.comaToolbox)
        layout_coma.addSpacing(75)
        # Internal line to separate coma form from buttons
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Raised)
        layout_coma.addWidget(line)
        layout_coma.addSpacing(25)
        layout_coma.addWidget(widget_buttons)
        self.setLayout(layout_coma)

        self.dayNightAsymmetry.chooseDNA.stateChanged.connect(self.__rotVariabilityControllerSlot)



    def __rotVariabilityControllerSlot(self):
        '''!
            internal slot to enable/disable the RotationalVariability slot of Coma Parameters form
            Suitable to be connected to a checkbox signal, like the DayNightAsimmetry slot main checkbox
        '''
        if self.dayNightAsymmetry.chooseDNA.isChecked():
            self.rotationalVariability.setEnabled(True)
        else:
            self.rotationalVariability.chooseRotation.setCheckState(Qt.Unchecked)
            self.rotationalVariability.setEnabled(False)

    def setDefaults(self):
        '''!
            Set input sub-sections widgets of the master class to their pre-defined default value
        '''
        self.particleSizeDistro.setDefaults()
        self.scattering.setDefaults()
        self.dustProduction.setDefaults()
        self.outflowVelocity.setDefaults()
        self.dayNightAsymmetry.setDefaults()
        self.rotationalVariability.setDefaults()
        self.gridResolution.setDefaults()

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.particleSizeDistro.setReadOnly(flag)
        self.scattering.setReadOnly(flag)
        self.dustProduction.setReadOnly(flag)
        self.outflowVelocity.setReadOnly(flag)
        self.dayNightAsymmetry.setReadOnly(flag)
        self.rotationalVariability.setReadOnly(flag)
        self.gridResolution.setReadOnly(flag)

    def dumpToXML(self, xmlElement):
        '''!
            Dump sub-sections input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("ComaParameters")
        if section is None:
            section = ET.SubElement(xmlElement, "ComaParameters")

        self.particleSizeDistro.dumpToXML(section)
        self.scattering.dumpToXML(section)
        self.dustProduction.dumpToXML(section)
        self.outflowVelocity.dumpToXML(section)
        self.dayNightAsymmetry.dumpToXML(section)
        self.rotationalVariability.dumpToXML(section)
        self.gridResolution.dumpToXML(section)

    def restoreFromXML(self, xmlElement):
        '''!
            Restore sub-sections input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # check if the element node exists in the tree chunk
        section = xmlElement.find("ComaParameters")
        if section is None:
            return False

        flag = bool(True)
        flag = (flag and self.particleSizeDistro.restoreFromXML(section))
        flag = (flag and self.scattering.restoreFromXML(section))
        flag = (flag and self.dustProduction.restoreFromXML(section))
        flag = (flag and self.outflowVelocity.restoreFromXML(section))
        flag = (flag and self.dayNightAsymmetry.restoreFromXML(section))
        flag = (flag and self.rotationalVariability.restoreFromXML(section))
        flag = (flag and self.gridResolution.restoreFromXML(section))

        return flag

    def compareWithXML(self, xmlElement):
        '''!
            Compare sub-sections input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("ComaParameters")
        if section is None:
            return False

        flag = bool(True)
        flag = (flag and self.particleSizeDistro.compareWithXML(section))
        flag = (flag and self.scattering.compareWithXML(section))
        flag = (flag and self.dustProduction.compareWithXML(section))
        flag = (flag and self.outflowVelocity.compareWithXML(section))
        flag = (flag and self.dayNightAsymmetry.compareWithXML(section))
        flag = (flag and self.rotationalVariability.compareWithXML(section))
        flag = (flag and self.gridResolution.compareWithXML(section))

        return flag


class ParticleSizeDistroInput(QWidget):
    '''!
        Collecting input to build the particle size distribution required to run the coma simulation.
    '''

    def __init__(self):
        '''!
            Constructor.
        '''
        super().__init__()

        ## Top check box to save the choice of the User (automatical default vs custom built. See ComMoDE Model documentation).
        self.distroType = QCheckBox("Default size distribution")
        self.distroType.setToolTip(
            "enabled -> proceed with the default size distribution, disabled(expert mode)->build a custom size distribution")
        ## Internal item for grouping the input needed to build a custom size distribution
        self.expertGroup = QGroupBox("Build a custom size distribution")

        # tag along changing state of check box
        self.distroType.stateChanged.connect(self.enableExpert)

        # fill the expert group

        ## Input widget to define the total number of bins in the distribution - value limits must be within [10,80]
        self.sizeBins = QSpinBox()
        self.sizeBins.setRange(10, 80)
        self.sizeBins.setSingleStep(1)
        labelSizeBins = QLabel("Number of size bins")
        labelSizeBins.setToolTip(
            "[-] : specifies the number of logarithmic size bins")

        ## Input widget to define the exponent k describing the the dust particle size distribution ejected from the nucleus nd(a) ∝ a^−κ. Value limits must be within [3.1, 4-5]
        self.powerLaw = QDoubleSpinBox()
        self.powerLaw.setRange(3.1, 4.5)
        self.powerLaw.setSingleStep(0.1)
        labelPowerLaw = QLabel("Power law index")
        labelPowerLaw.setToolTip(
            "[-] : specifies the exponent κ describing the dust particle size distribution ejected from the nucleus nd(a) ∝ a^−κ.")

        labelSizeMiddle = QLabel("10^")
        labelSizeMiddle.setAlignment(Qt.AlignRight | Qt.AlignCenter)
        labelSizeMiddle.setFixedWidth(40)
        labelSizeMiddle2 = QLabel("10^")
        labelSizeMiddle2.setAlignment(Qt.AlignRight | Qt.AlignCenter)
        labelSizeMiddle2.setFixedWidth(40)

        ## Input widget to define the smallest size of the distribution. The number is intended as the exponent x to use in expression 10^x to define the size in [m]. Value limits must be within [-7,-5]
        self.minSize = QSpinBox()
        self.minSize.setRange(-7, -5)
        self.minSize.setSingleStep(1)
        labelMinSize = QLabel("Smallest size")
        labelMinSize.setToolTip(
            "[-] : specifies the smallest considered particle size. Value entry corresponds to the exponent x to use in expression 10^x to create the size in [m]")

        ## Input widget to define the largest size of the distribution. The number is intended as the exponent x to use in expression 10^x to define the size in [m]. Value limits must be within [-3,-1]
        self.maxSize = QSpinBox()
        self.maxSize.setRange(-3, -1)
        self.maxSize.setSingleStep(1)
        labelMaxSize = QLabel("Largest size")
        labelMaxSize.setToolTip(
            "[-] : specifies the largest considered particle size. Value entry corresponds to the exponent x to use in expression 10^x to create the size in [m]")

        # create the expert group layout
        layout_expert = QGridLayout()
        layout_expert.setVerticalSpacing(2)
        layout_expert.addWidget(labelSizeBins, 0, 0)
        layout_expert.addWidget(self.sizeBins, 0, 2)
        layout_expert.addWidget(labelPowerLaw, 1, 0)
        layout_expert.addWidget(self.powerLaw, 1, 2)
        layout_expert.addWidget(labelMinSize, 2, 0)
        layout_expert.addWidget(labelSizeMiddle, 2, 1)
        layout_expert.addWidget(self.minSize, 2, 2)
        layout_expert.addWidget(labelMaxSize, 3, 0)
        layout_expert.addWidget(labelSizeMiddle2, 3, 1)
        layout_expert.addWidget(self.maxSize, 3, 2)

        self.expertGroup.setLayout(layout_expert)

        # create the overall layout
        layout = QVBoxLayout()
        layout.setContentsMargins(2, 2, 2, 2)
        layout.addWidget(self.distroType)
        layout.addSpacing(3)
        layout.addWidget(self.expertGroup)
        self.setLayout(layout)

        self.setDefaults()

    def enableExpert(self):
        '''!
            internal slot to enable building a custom size distribution (expert mode)
        '''
        if self.distroType.isChecked():
            self.expertGroup.setEnabled(False)
            self.setDefaults()
        else:
            self.expertGroup.setEnabled(True)

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.distroType.setCheckState(Qt.Checked)
        self.sizeBins.setValue(18)
        self.powerLaw.setValue(3.8)
        self.minSize.setValue(-7)
        self.maxSize.setValue(-1)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.sizeBins.setReadOnly(flag)
        self.powerLaw.setReadOnly(flag)
        self.minSize.setReadOnly(flag)
        self.maxSize.setReadOnly(flag)
        # check box don't have read only revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.distroType.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("SizeDistribution")
        if section is None:
            section = ET.SubElement(xmlElement, "SizeDistribution")

        checkdefault = section.find("default")
        if checkdefault is None:
            checkdefault = ET.SubElement(section, "default")
        checkdefault.text = str(self.distroType.isChecked())

        nbins = section.find("NumberOfSizeBins")
        if nbins is None:
            nbins = ET.SubElement(section, "NumberOfSizeBins")
        nbins.text = str(self.sizeBins.value())

        plaw = section.find("PowerLawIndex")
        if plaw is None:
            plaw = ET.SubElement(section, "PowerLawIndex")
        plaw.text = "{:.2f}".format(self.powerLaw.value())

        minsize = section.find("SmallestSize")
        if minsize is None:
            minsize = ET.SubElement(section, "SmallestSize")
        minsize.text = str(self.minSize.value())

        maxsize = section.find("LargestSize")
        if maxsize is None:
            maxsize = ET.SubElement(section, "LargestSize")
        maxsize.text = str(self.maxSize.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("SizeDistribution")
        if section is None:
            return False

        checkdefault = section.find("default")
        if not (checkdefault is None):
            s = checkdefault.text.strip()
            if s == "True":
                self.distroType.setCheckState(Qt.Checked)
            else:
                self.distroType.setCheckState(Qt.Unchecked)

        nbins = section.find("NumberOfSizeBins")
        if not (nbins is None):
            self.sizeBins.setValue(int(nbins.text))

        plaw = section.find("PowerLawIndex")
        if not (plaw is None):
            self.powerLaw.setValue(float(plaw.text))

        minsize = section.find("SmallestSize")
        if not (minsize is None):
            self.minSize.setValue(int(minsize.text))

        maxsize = section.find("LargestSize")
        if not (maxsize is None):
            self.maxSize.setValue(int(maxsize.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("SizeDistribution")
        if section is None:
            return False
        # compare the checkBox distroType, if False compare all the form of custom distro
        innerCheck = str(self.distroType.isChecked())

        checkdefault = section.find("default")
        if checkdefault is None:
            return False

        if checkdefault.text.strip() != innerCheck:
            return False

        if (innerCheck == "False"):
            nbins = section.find("NumberOfSizeBins")
            plaw = section.find("PowerLawIndex")
            minsize = section.find("SmallestSize")
            maxsize = section.find("LargestSize")

            if nbins is None or plaw is None or minsize is None or maxsize is None:
                return False

            if str(self.sizeBins.value()) != nbins.text.strip():
                return False
            if "{:.2f}".format(self.powerLaw.value()) != plaw.text.strip():
                return False
            if str(self.minSize.value()) != minsize.text.strip():
                return False
            if str(self.maxSize.value()) != maxsize.text.strip():
                return False

        # otherwise
        return True


class ScatteringInput(QWidget):
    '''!
        Collecting input to define dust scattering properties required by the coma simulation.
    '''

    def __init__(self):
        '''!
            Constructor.
        '''
        super().__init__()

        ## widget to collect the heliocentric distance input - value limits within [0.5,2.0] AU
        self.helioDist = QDoubleSpinBox()
        self.helioDist.setRange(0.5, 2.0)
        self.helioDist.setSingleStep(0.01)
        labelHelioDist = QLabel("Heliocentric distance")
        labelHelioDist.setToolTip(
            "[AU] : specifies the distance between the sun and the comet")

        ## widget to collect the phase angle input (to be used in conjunction with ScatteringInput.helioDist) - value limits within [0.0,180.0] deg
        self.phaseAngle = QSpinBox()
        self.phaseAngle.setRange(0, 180)
        self.phaseAngle.setSingleStep(1)
        labelPhaseAngle = QLabel("Phase angle")
        labelPhaseAngle.setToolTip("[deg] : specifies the phase angle")

        ## widget to define the target dust material : enlist all available choices (water ice - 0, enstatite - 1, astro-silicate - 2, troilite - 3)
        self.materials = QComboBox()
        self.materials.addItem("water ice")
        self.materials.addItem("enstatite")
        self.materials.addItem("astro-silicate")
        self.materials.addItem("troilite")
        labelMaterials = QLabel("Material")
        labelMaterials.setToolTip(" : pick up a dust material from list")

        # create the expert group layout
        layout = QGridLayout()
        layout.setVerticalSpacing(2)
        layout.addWidget(labelMaterials, 0, 0)
        layout.addWidget(self.materials, 0, 1)
        layout.addWidget(labelHelioDist, 1, 0)
        layout.addWidget(self.helioDist, 1, 1)
        layout.addWidget(labelPhaseAngle, 2, 0)
        layout.addWidget(self.phaseAngle, 2, 1)

        self.setLayout(layout)

        self.setDefaults()

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.materials.setCurrentIndex(0)  # move it to water ice
        self.helioDist.setValue(0.5)
        self.phaseAngle.setValue(0)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.helioDist.setReadOnly(flag)
        self.phaseAngle.setReadOnly(flag)

        # combobox don't have read only revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.materials.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("ScatteringProperties")
        if section is None:
            section = ET.SubElement(xmlElement, "ScatteringProperties")

        mat = section.find("Material")
        if mat is None:
            mat = ET.SubElement(section, "Material")
        mat.text = self.materials.currentText()

        heliodist = section.find("HeliocentricDistance")
        if heliodist is None:
            heliodist = ET.SubElement(section, "HeliocentricDistance")
        heliodist.text = "{:.2f}".format(self.helioDist.value())

        pangle = section.find("PhaseAngle")
        if pangle is None:
            pangle = ET.SubElement(section, "PhaseAngle")
        pangle.text = str(self.phaseAngle.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("ScatteringProperties")
        if section is None:
            return False

        mat = section.find("Material")
        if not (mat is None):
            s = mat.text.strip()
            if s == "enstatite":
                self.materials.setCurrentIndex(1)
            elif s == "astro-silicate":
                self.materials.setCurrentIndex(2)
            elif s == "troilite":
                self.materials.setCurrentIndex(3)
            else:
                self.materials.setCurrentIndex(0)

        heliodist = section.find("HeliocentricDistance")
        if not (heliodist is None):
            self.helioDist.setValue(float(heliodist.text))

        pangle = section.find("PhaseAngle")
        if not (pangle is None):
            self.phaseAngle.setValue(int(pangle.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("ScatteringProperties")
        if section is None:
            return False

        mat = section.find("Material")
        hdist = section.find("HeliocentricDistance")
        pangle = section.find("PhaseAngle")

        if mat is None or hdist is None or pangle is None:
            return False
        if mat.text.strip() != self.materials.currentText():
            return False
        if "{:.2f}".format(self.helioDist.value()) != hdist.text.strip():
            return False
        if str(self.phaseAngle.value()) != pangle.text.strip():
            return False

        # otherwise
        return True


class GlobalDustProductionInput(QWidget):
    '''!
        Collecting input to define the dust production rate required by the coma simulation.
    '''

    def __init__(self):
        '''!
            Constructor
        '''
        super().__init__()

        ## widget to collect the dust production rate input - value limits within [1000,40000] kg/s
        self.dustProdRate = QDoubleSpinBox()
        self.dustProdRate.setRange(1000, 40000)
        self.dustProdRate.setSingleStep(1.0)
        ## (internal) label attached to dust production rate input
        self.labelDustProdRate = QLabel("Dust production rate")
        self.labelDustProdRate.setToolTip(
            "[kg/s] : specifies the global dust production rate of the comet")

        ## widget to collect the Afrho input - value limits within [25,300] m
        self.afrho = QDoubleSpinBox()
        self.afrho.setRange(25.0, 300.0)
        self.afrho.setSingleStep(1.0)

        ## widget to check Afrho as method to pass the production rate input
        self.chooseAfrho = QCheckBox('Use Af\u03C1')
        self.chooseAfrho.setToolTip(
            '[m] : enabled-> use Af\u03C1 instead, to define the comet production rate')

        # (internal) define internal layout to draw internal input wdiget
        layout_standard = QGridLayout()
        layout_standard.setVerticalSpacing(2)
        layout_standard.addWidget(self.labelDustProdRate, 0, 0)
        layout_standard.addWidget(self.dustProdRate, 0, 1)
        layout_standard.addWidget(self.chooseAfrho, 1, 0)
        layout_standard.addWidget(self.afrho, 1, 1)

        self.setLayout(layout_standard)

        self.chooseAfrho.stateChanged.connect(self.enableAfrho)

        self.setDefaults()

    def enableAfrho(self):
        '''!
            Internal slot method to enable/disable Afrho alternative to dust production rate on class chooseAfhro check box
        '''
        if self.chooseAfrho.isChecked():
            self.afrho.setEnabled(True)
            self.dustProdRate.setEnabled(False)
            self.labelDustProdRate.setEnabled(False)
        else:
            self.afrho.setEnabled(False)
            self.dustProdRate.setEnabled(True)
            self.labelDustProdRate.setEnabled(True)

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.chooseAfrho.setCheckState(Qt.Checked)  # just to enable the change of state
        self.chooseAfrho.setCheckState(Qt.Unchecked)
        self.afrho.setValue(25.0)
        self.dustProdRate.setValue(1000.0)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.afrho.setReadOnly(flag)
        self.dustProdRate.setReadOnly(flag)

        # checkbox don't have read only revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.chooseAfrho.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("DustProduction")
        if section is None:
            section = ET.SubElement(xmlElement, "DustProduction")

        afrhochoice = section.find("UseAfrho")
        if afrhochoice is None:
            afrhochoice = ET.SubElement(section, "UseAfrho")
        afrhochoice.text = str(self.chooseAfrho.isChecked())

        rate = section.find("DustProductionRate")
        if rate is None:
            rate = ET.SubElement(section, "DustProductionRate")
        rate.text = "{:.2f}".format(self.dustProdRate.value())

        afrhovalue = section.find("Afrho")
        if afrhovalue is None:
            afrhovalue = ET.SubElement(section, "Afrho")
        afrhovalue.text = "{:.2f}".format(self.afrho.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("DustProduction")
        if section is None:
            return False

        useafrho = section.find("UseAfrho")
        if not (useafrho is None):
            s = useafrho.text.strip()
            if s == "True":
                self.chooseAfrho.setCheckState(Qt.Checked)
            else:
                self.chooseAfrho.setCheckState(Qt.Unchecked)

        rate = section.find("DustProductionRate")
        if not (rate is None):
            self.dustProdRate.setValue(float(rate.text))

        afrhoval = section.find("Afrho")
        if not (afrhoval is None):
            self.afrho.setValue(float(afrhoval.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("DustProduction")
        if section is None:
            return False

        innerCheck = str(self.chooseAfrho.isChecked())

        useafrho = section.find("UseAfrho")
        if useafrho is None:
            return False
        if useafrho.text.strip() != innerCheck:
            return False

        if (innerCheck == "False"):
            rate = section.find("DustProductionRate")

            if rate is None:
                return False

            if "{:.2f}".format(self.dustProdRate.value()) != rate.text.strip():
                return False
        else:
            afrhoval = section.find("Afrho")

            if afrhoval is None:
                return False

            if "{:.2f}".format(self.afrho.value()) != afrhoval.text.strip():
                return False

        # otherwise
        return True


class OutflowVelocityInput(QWidget):
    '''!
        Collecting input to define the dust outflow velocity condition required by the coma simulation.
    '''

    def __init__(self):
        '''!
            Constructor.
        '''
        super().__init__()

        ## widget to collect the gas production rate input - limits within [1000, 40000] Kg/s
        self.gasProdRate = QDoubleSpinBox()
        self.gasProdRate.setRange(1000, 40000)
        self.gasProdRate.setSingleStep(1.0)
        labelGasProdRate = QLabel("Gas production rate")
        labelGasProdRate.setToolTip("[Kg/s] : specifies the gas production ")

        ## widget to enable further inputs to fill (expert mode) for the class
        self.enableExpertWidget = QCheckBox("Enable other options...")

        layout_standard = QGridLayout()
        layout_standard.setVerticalSpacing(2)
        layout_standard.addWidget(labelGasProdRate, 0, 0)
        layout_standard.addWidget(self.gasProdRate, 0, 1)
        layout_standard.addWidget(self.enableExpertWidget, 2, 0)

        standardWidget = QWidget()
        standardWidget.setLayout(layout_standard)

        ## widget to collect the nucleus radius input - limits within [1000, 30000] m
        self.nucleusRadius = QDoubleSpinBox()
        self.nucleusRadius.setRange(1000, 30000)
        self.nucleusRadius.setSingleStep(1.0)
        labelNucleusRadius = QLabel("Radius of comet nucleus")
        labelNucleusRadius.setToolTip(
            "[m] : specifies the radius of the comet")

        ## widget to collect the nucleus surface temperature - limits within [280, 340] K
        self.nucleusTemperature = QDoubleSpinBox()
        self.nucleusTemperature.setRange(280.0, 340.0)
        self.nucleusTemperature.setSingleStep(1.0)
        labelNucleusTemperature = QLabel("Nucleus surface temperature")
        labelNucleusTemperature.setToolTip(
            "[K] : specifies the temperature on the surface of the nucleus")

        ## widget to collect the bulk dust density of the dust particles - limits within [400, 800] Kg/m^3
        self.dustBulkDensity = QDoubleSpinBox()
        self.dustBulkDensity.setRange(400.0, 800.0)
        self.dustBulkDensity.setSingleStep(1.0)
        labelDustBulkDensity = QLabel("Bulk dust density")
        labelDustBulkDensity.setToolTip(
            "[kg/m^3] : specifies the bulk dust density of the prototype dust particles")

        # (internal) layout to displace the expert mode input widgets
        layout_expert = QGridLayout()
        layout_expert.setVerticalSpacing(2)

        layout_expert.addWidget(labelNucleusRadius, 0, 0)
        layout_expert.addWidget(self.nucleusRadius, 0, 1)
        layout_expert.addWidget(labelNucleusTemperature, 1, 0)
        layout_expert.addWidget(self.nucleusTemperature, 1, 1)
        layout_expert.addWidget(labelDustBulkDensity, 2, 0)
        layout_expert.addWidget(self.dustBulkDensity, 2, 1)

        ## (internal) wrapping widget to host layout_expert
        self.expertWidget = QWidget()
        self.expertWidget.setLayout(layout_expert)

        # final layout hosting both standard and expert sub layouts
        final_layout = QVBoxLayout()
        final_layout.setContentsMargins(2, 2, 2, 2)
        final_layout.addWidget(standardWidget)
        final_layout.addSpacing(3)
        final_layout.addWidget(self.expertWidget)
        self.setLayout(final_layout)

        # connecting standard with expert
        self.enableExpertWidget.stateChanged.connect(self.enableExpertGroup)

        self.setDefaults()

    def enableExpertGroup(self):
        '''!
            Internal slot to enable all the other options in expert mode
        '''
        if self.enableExpertWidget.isChecked():
            self.expertWidget.setEnabled(True)
        else:
            self.expertWidget.setEnabled(False)
            self.nucleusRadius.setValue(5000.0)
            self.nucleusTemperature.setValue(314.0)
            self.dustBulkDensity.setValue(533.0)


    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default values
        '''
        self.gasProdRate.setValue(1000.0)
        self.enableExpertWidget.setCheckState(
            Qt.Checked)  # just to enable the change state
        self.enableExpertWidget.setCheckState(Qt.Unchecked)
        self.nucleusRadius.setValue(5000.0)
        self.nucleusTemperature.setValue(314.0)
        self.dustBulkDensity.setValue(533.0)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.gasProdRate.setReadOnly(flag)
        self.nucleusRadius.setReadOnly(flag)
        self.nucleusTemperature.setReadOnly(flag)
        self.dustBulkDensity.setReadOnly(flag)

        # checkbox don't have read only revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.enableExpertWidget.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("DustOutflowVelocity")
        if section is None:
            section = ET.SubElement(xmlElement, "DustOutflowVelocity")

        rate = section.find("GasProductionRate")
        if rate is None:
            rate = ET.SubElement(section, "GasProductionRate")
        rate.text = "{:.2f}".format(self.gasProdRate.value())

        expertchoice = section.find("expert")
        if expertchoice is None:
            expertchoice = ET.SubElement(section, "expert")
        expertchoice.text = str(self.enableExpertWidget.isChecked())

        radius = section.find("NucleusRadius")
        if radius is None:
            radius = ET.SubElement(section, "NucleusRadius")
        radius.text = "{:.2f}".format(self.nucleusRadius.value())

        temperature = section.find("NucleusSurfaceTemperature")
        if temperature is None:
            temperature = ET.SubElement(section, "NucleusSurfaceTemperature")
        temperature.text = "{:.2f}".format(self.nucleusTemperature.value())

        density = section.find("DustBulkDensity")
        if density is None:
            density = ET.SubElement(section, "DustBulkDensity")
        density.text = "{:.2f}".format(self.dustBulkDensity.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("DustOutflowVelocity")
        if section is None:
            return False

        expert = section.find("expert")
        if not (expert is None):
            s = expert.text.strip()
            if s == "True":
                self.enableExpertWidget.setCheckState(Qt.Checked)
            else:
                self.enableExpertWidget.setCheckState(Qt.Unchecked)

        rate = section.find("GasProductionRate")
        if not (rate is None):
            self.gasProdRate.setValue(float(rate.text))

        radius = section.find("NucleusRadius")
        if not (radius is None):
            self.nucleusRadius.setValue(float(radius.text))

        temperature = section.find("NucleusSurfaceTemperature")
        if not (temperature is None):
            self.nucleusTemperature.setValue(float(temperature.text))

        density = section.find("DustBulkDensity")
        if not (density is None):
            self.dustBulkDensity.setValue(float(density.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("DustOutflowVelocity")
        if section is None:
            return False

        rate = section.find("GasProductionRate")
        if rate is None:
            return False
        if "{:.2f}".format(self.gasProdRate.value()) != rate.text.strip():
            return False

        innerCheck = str(self.enableExpertWidget.isChecked())

        expert = section.find("expert")
        if expert is None:
            return False
        if expert.text.strip() != innerCheck:
            return False

        if (innerCheck == "True"):
            rad = section.find("NucleusRadius")
            temp = section.find("NucleusSurfaceTemperature")
            dens = section.find("DustBulkDensity")

            if rad is None or temp is None or dens is None:
                return False

            if "{:.2f}".format(self.nucleusRadius.value()) != rad.text.strip():
                return False
            if "{:.2f}".format(self.nucleusTemperature.value()) != temp.text.strip():
                return False
            if "{:.2f}".format(self.dustBulkDensity.value()) != dens.text.strip():
                return False

        # otherwise
        return True


class DayNightAsymmetryInput(QWidget):
    '''!
        Collecting input to define the Day/Night asymmetry optional condition for the coma simulation.
    '''

    def __init__(self):
        '''!
            Constructor
        '''
        super().__init__()

        ## widget check box to enable the asymmetry of Day and Night inside the coma simulation
        self.chooseDNA = QCheckBox("Use Day/Night side asymmetry ")
        self.chooseDNA.setToolTip(
            "[-] : enable day- to night-side asymmetry. In the isotropic case the dust densities can be calculated analytically which decreases the running time by orders of magnitude")

        layout_standard = QGridLayout()
        layout_standard.setVerticalSpacing(2)
        layout_standard.addWidget(self.chooseDNA, 0, 0)

        standardWidget = QWidget()
        standardWidget.setLayout(layout_standard)

        ## widget to choose the function to model emission angle of dust production rate and outflow velocity - function available are cosine, gaussian and half-maxwell
        self.angularDepLaw = QComboBox()
        self.angularDepLaw.addItem("cosine")
        self.angularDepLaw.addItem("gaussian")
        self.angularDepLaw.addItem("half-maxwell")

        ## (internal) widget to attach a Label to the angularDepLaw member
        self.labelAngularDep = QLabel("Angular asymmetry function")
        self.labelAngularDep.setToolTip(
            "[-] : choose function to model the emission angle dependence of the outflow velocity and the dust production rate")

        ## widget to collect cosine function amplitude in case an asymmetry function of type cosine is chosen - limits within [0,1]
        self.cosineAmp = QDoubleSpinBox()
        self.cosineAmp.setRange(0.0, 1.0)
        self.cosineAmp.setSingleStep(0.01)

        ## (internal) widget to attach a Label to the cosineAmp member
        self.labelCosineAmp = QLabel("Amplitude of cosine")
        self.labelCosineAmp.setToolTip(
            "[-] : specifies the amplitude of the cosine used to introduce an emission angle dependence. At the maximum amplitude the dust production rate at the anti-solar point is 0. This input parameter can only be set if the angular asymmetry function is the cosine ")

        layout_standard.addWidget(self.labelAngularDep, 0, 1)
        layout_standard.addWidget(self.angularDepLaw, 1, 0)
        layout_standard.addWidget(self.labelCosineAmp, 2, 0)

        layout_expert = QGridLayout()
        layout_expert.setVerticalSpacing(2)

        layout_expert.addWidget(self.labelAngularDep, 0, 0)
        layout_expert.addWidget(self.angularDepLaw, 0, 1)
        layout_expert.addWidget(self.labelCosineAmp, 1, 0)
        layout_expert.addWidget(self.cosineAmp, 1, 1)

        ## (internal) wrapping widget to host layout_expert
        self.expertWidget = QWidget()
        self.expertWidget.setLayout(layout_expert)

        ## (internal) final collective layout widget
        final_layout = QVBoxLayout()
        final_layout.setContentsMargins(2, 2, 2, 2)
        final_layout.addWidget(standardWidget)
        final_layout.addSpacing(3)
        final_layout.addWidget(self.expertWidget)
        self.setLayout(final_layout)

        self.chooseDNA.stateChanged.connect(self.enableDNA)
        self.angularDepLaw.currentIndexChanged.connect(self.enableCosine)

        self.setDefaults()

    def enableDNA(self):
        '''!
            internal slot to enable the the day/nightside asimmetry expert options
        '''
        if self.chooseDNA.isChecked():
            self.expertWidget.setEnabled(True)
        else:
            self.expertWidget.setEnabled(False)

    def enableCosine(self, index):
        '''!
            Internal slot to enable cosine amplitude widget, if cosine function is chosed as angular asymmetry function
        '''
        if self.angularDepLaw.currentIndex() != 0:
            self.cosineAmp.setEnabled(False)
            self.labelCosineAmp.setEnabled(False)
        else:
            self.cosineAmp.setEnabled(True)
            self.labelCosineAmp.setEnabled(True)

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.chooseDNA.setCheckState(Qt.Checked)  # activate manually the change item of the checkbox
        self.chooseDNA.setCheckState(Qt.Unchecked)
        # activate manually the change item of the combobox
        self.angularDepLaw.setCurrentIndex(1)
        self.angularDepLaw.setCurrentIndex(0)

        self.cosineAmp.setValue(0.0)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.cosineAmp.setReadOnly(flag)

        # checkbox/combobox don't have read only revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.chooseDNA.setEnabled(enable)
        self.angularDepLaw.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("DayNightAsymmetry")
        if section is None:
            section = ET.SubElement(xmlElement, "DayNightAsymmetry")

        choose = section.find("active")
        if choose is None:
            choose = ET.SubElement(section, "active")
        choose.text = str(self.chooseDNA.isChecked())

        function = section.find("AngularAsymmetryFunction")
        if function is None:
            function = ET.SubElement(section, "AngularAsymmetryFunction")
        function.text = self.angularDepLaw.currentText()

        cosamp = section.find("CosineAmplitude")
        if cosamp is None:
            cosamp = ET.SubElement(section, "CosineAmplitude")
        cosamp.text = "{:.2f}".format(self.cosineAmp.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("DayNightAsymmetry")
        if section is None:
            return False

        active = section.find("active")
        if not (active is None):
            s = active.text.strip()
            if s == "True":
                self.chooseDNA.setCheckState(Qt.Checked)
            else:
                self.chooseDNA.setCheckState(Qt.Unchecked)

        function = section.find("AngularAsymmetryFunction")
        if not (function is None):
            s = function.text.strip()
            if s == "gaussian":
                self.angularDepLaw.setCurrentIndex(1)
            elif s == "half-maxwell":
                self.angularDepLaw.setCurrentIndex(2)
            else:
                self.angularDepLaw.setCurrentIndex(0)

        cosamp = section.find("CosineAmplitude")
        if not (cosamp is None):
            self.cosineAmp.setValue(float(cosamp.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("DayNightAsymmetry")
        if section is None:
            return False

        innerCheck = str(self.chooseDNA.isChecked())

        active = section.find("active")
        if active is None:
            return False
        if active.text.strip() != innerCheck:
            return False

        if (innerCheck == "True"):
            function = section.find("AngularAsymmetryFunction")

            if function is None:
                return False
            if self.angularDepLaw.currentText() != function.text.strip():
                return False

            if function.text.strip() == "cosine":
                cosamp = section.find("CosineAmplitude")
                if cosamp is None:
                    return False
                if "{:.2f}".format(self.cosineAmp.value()) != cosamp.text.strip():
                    return False

        # otherwise
        return True


class RotationalPeriodInput(QWidget):
    '''!
        Collecting additional input to take into comet nucleus rotation for calculating the dust production rate.
    '''

    def __init__(self):
        '''!
            Constructor.
        '''
        super().__init__()

        ## checkbox widget input: if enabled, the dust production rate will vary due to the rotation of the comet nucleus
        self.chooseRotation = QCheckBox("Use Rotational variability")
        self.chooseRotation.setToolTip(
            "enabled -> dust production rate will vary due to the rotation of the comet nucleus")

        ## widget input that specifies the rotational period of the comet nucleus - limits within [0,50] hh
        self.rotationPeriod = QDoubleSpinBox()
        self.rotationPeriod.setRange(0.0, 50.0)
        self.rotationPeriod.setSingleStep(1.0)

        ## (internal) label attached to rotationPeriod widget
        self.rotationPeriodLabel = QLabel("Rotational period")
        self.rotationPeriodLabel.setToolTip(
            "[hh] : specifies the rotational period of the comet nucleus")

        ## widget input that specifies the amplitude of the rotational variability relative to the average dust production rate - limits within [0,100] %
        self.rotAmplitudeVal = QSpinBox()
        self.rotAmplitudeVal.setFixedWidth(80)
        self.rotAmplitudeVal.setRange(0, 100)
        self.rotAmplitudeVal.setSingleStep(1)

        ## (internal) label attached to rotAmplitudeVal / rotAmplitudeSlider widgets
        self.labelRotAmplitude = QLabel("Amplitude of rotational variability")
        self.labelRotAmplitude.setToolTip(
            "[%] : specifies the amplitude of the dust production rate variability due to rotation, relative to the average dust production rate")

        ## slider widget paired with rotAmplitudeVal input. Value of % amplitude can be alternatively specified moving the slider or inserted manually from rotAmplitudeVal.
        self.rotAmplitudeSlider = QSlider()
        self.rotAmplitudeSlider.setOrientation(Qt.Horizontal)
        self.rotAmplitudeSlider.setRange(0, 100)
        self.rotAmplitudeSlider.setValue(0)
        self.rotAmplitudeSlider.setSingleStep(1)
        self.rotAmplitudeSlider.setTracking(True)

        layout_standard = QGridLayout()
        layout_standard.setVerticalSpacing(2)
        layout_standard.setHorizontalSpacing(20)

        layout_standard.addWidget(self.chooseRotation, 0, 0)
        layout_standard.addWidget(self.rotationPeriodLabel, 1, 0)
        layout_standard.addWidget(self.rotationPeriod, 1, 1)
        layout_standard.addWidget(self.labelRotAmplitude, 2, 0)
        layout_standard.addWidget(self.rotAmplitudeSlider, 2, 1)
        layout_standard.addWidget(self.rotAmplitudeVal, 2, 2)

        self.setLayout(layout_standard)

        self.chooseRotation.stateChanged.connect(self.enableRotationPeriod)
        self.rotAmplitudeVal.valueChanged.connect(self.moveSlider)
        self.rotAmplitudeSlider.valueChanged.connect(self.changeBoxValue)

        self.setDefaults()

    def enableRotationPeriod(self):
        '''!
            Internal slot to enable rotation variability expert group
        '''
        if self.chooseRotation.isChecked():
            self.rotationPeriod.setEnabled(True)
            self.rotationPeriodLabel.setEnabled(True)
            self.labelRotAmplitude.setEnabled(True)
            self.rotAmplitudeVal.setEnabled(True)
            self.rotAmplitudeSlider.setEnabled(True)

        else:
            self.rotationPeriodLabel.setEnabled(False)
            self.rotationPeriod.setEnabled(False)
            self.labelRotAmplitude.setEnabled(False)
            self.rotAmplitudeVal.setEnabled(False)
            self.rotAmplitudeSlider.setEnabled(False)

    def moveSlider(self, value):
        '''!
            Internal slot to change amplitude slider position
            @param value new value to position the rotAmplitudeSlider (must be within [0,100])
        '''
        self.rotAmplitudeSlider.setValue(value)

    def changeBoxValue(self, value):
        '''!
            Internal slot to change amplitude box value
            @param value new Amplitude value to assign into the relative spin box widget rotAmplitudeVal (must be within [0,100])
        '''
        self.rotAmplitudeVal.setValue(value)

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.chooseRotation.setCheckState(
            Qt.Checked)  # activate manually the change item of the combobox
        self.chooseRotation.setCheckState(Qt.Unchecked)
        self.rotationPeriod.setValue(0.0)
        # activate manually the mutual tracking amplitude val/slider
        self.rotAmplitudeVal.setValue(1)
        self.rotAmplitudeVal.setValue(0)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.rotationPeriod.setReadOnly(flag)
        self.rotAmplitudeVal.setReadOnly(flag)

        # checkbox/slider don't have read only revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.chooseRotation.setEnabled(enable)
        self.rotAmplitudeSlider.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("RotationalVariability")
        if section is None:
            section = ET.SubElement(xmlElement, "RotationalVariability")

        choose = section.find("active")
        if choose is None:
            choose = ET.SubElement(section, "active")
        choose.text = str(self.chooseRotation.isChecked())

        period = section.find("RotationalPeriod")
        if period is None:
            period = ET.SubElement(section, "RotationalPeriod")
        period.text = "{:.2f}".format(self.rotationPeriod.value())

        rotAmplitude = section.find("RotationalAmplitude")
        if rotAmplitude is None:
            rotAmplitude = ET.SubElement(section, "RotationalAmplitude")
        rotAmplitude.text = str(self.rotAmplitudeVal.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("RotationalVariability")
        if section is None:
            return False

        active = section.find("active")
        if not (active is None):
            s = active.text.strip()
            if s == "True":
                self.chooseRotation.setCheckState(Qt.Checked)
            else:
                self.chooseRotation.setCheckState(Qt.Unchecked)

        period = section.find("RotationalPeriod")
        if not (period is None):
            self.rotationPeriod.setValue(float(period.text))

        amp = section.find("RotationalAmplitude")
        if not (amp is None):
            self.rotAmplitudeVal.setValue(int(amp.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("RotationalVariability")
        if section is None:
            return False

        innerCheck = str(self.chooseRotation.isChecked())

        active = section.find("active")
        if active is None:
            return False
        if active.text.strip() != innerCheck:
            return False

        if (innerCheck == "True"):
            period = section.find("RotationalPeriod")
            amp = section.find("RotationalAmplitude")

            if period is None or amp is None:
                return False
            if "{:.2f}".format(self.rotationPeriod.value()) != period.text.strip():
                return False
            if str(self.rotAmplitudeVal.value()) != amp.text.strip():
                return False

        # otherwise
        return True


class GridResolutionInput(QWidget):
    '''!
        Collecting input to define the computational mesh resolution of the coma simulation.
    '''

    def __init__(self):
        '''!
            Constructor
        '''
        super().__init__()

        ## widget input to define the maximum cometocentric distance - limits within [10000, 100000] Km
        self.cometoDist = QDoubleSpinBox()
        self.cometoDist.setRange(10000.0, 100000.0)
        self.cometoDist.setSingleStep(1.0)
        labelCometoDist = QLabel("Maximum distance")
        labelCometoDist.setToolTip(
            "[Km] : specifies the maximum cometocentric distance")

        ## widget input to define the sidelength of the mesh cells - limits within [10, 100] Km
        self.resolution = QSpinBox()
        self.resolution.setRange(10, 100)
        self.resolution.setSingleStep(1)
        labelResolution = QLabel("Resolution")
        labelResolution.setToolTip(
            "[Km] : : specifies the sidelength of the grid cells")

        # create thegroup layout
        layout = QGridLayout()
        layout.setVerticalSpacing(2)
        layout.addWidget(labelCometoDist, 0, 0)
        layout.addWidget(self.cometoDist, 0, 1)
        layout.addWidget(labelResolution, 1, 0)
        layout.addWidget(self.resolution, 1, 1)

        self.setLayout(layout)

        self.setDefaults()

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.cometoDist.setValue(10000.0)
        self.resolution.setValue(10)

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.cometoDist.setReadOnly(flag)
        self.resolution.setReadOnly(flag)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("GridResolution")
        if section is None:
            section = ET.SubElement(xmlElement, "GridResolution")

        dist = section.find("MaximumDistance")
        if dist is None:
            dist = ET.SubElement(section, "MaximumDistance")
        dist.text = "{:.2f}".format(self.cometoDist.value())

        res = section.find("Resolution")
        if res is None:
            res = ET.SubElement(section, "Resolution")
        res.text = str(self.resolution.value())

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        section = xmlElement.find("GridResolution")
        if section is None:
            return False

        dist = section.find("MaximumDistance")
        if not (dist is None):
            self.cometoDist.setValue(float(dist.text))

        res = section.find("Resolution")
        if not (res is None):
            self.resolution.setValue(int(res.text))

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("GridResolution")
        if section is None:
            return False

        dist = section.find("MaximumDistance")
        res = section.find("Resolution")

        if dist is None or res is None:
            return False
        if "{:.2f}".format(self.cometoDist.value()) != dist.text.strip():
            return False
        if str(self.resolution.value()) != res.text.strip():
            return False

        # otherwise
        return True
