from PySide2.QtWidgets import QWidget
from PySide2.QtWidgets import QPushButton, QGroupBox, QVBoxLayout
from PySide2.QtWidgets import QHBoxLayout, QFrame, QMessageBox
from PySide2.QtWidgets import QLabel, QDoubleSpinBox, QInputDialog, QLineEdit
from PySide2.QtWidgets import QGridLayout, QGroupBox, QListWidget


from PySide2.QtCore import Qt, Signal

import xml.etree.ElementTree as ET

class Trajectory():
    '''!
        Worker class collecting User Input needed to build a spacecraft trajectory
    '''  
    def __init__(self):
        '''!
            Constructor
        '''
        # set up Trajectory parameters
        # ---------------------------------------------------------------------------------------------

        ## Spacecraft velocity [Km/s]
        self.velocity = float(10.0)
        # ---------------------------------------------------------------------------------------------

        ## Phase Angle of Approach [deg] 
        self.phaseAngle = float(-90.0)
        # ---------------------------------------------------------------------------------------------

        ## Impact parameter [Km] 
        self.impact = float(100.0)
        # ---------------------------------------------------------------------------------------------

        ## Altitude Angle [deg] 
        self.altitudeAngle = float(-90.0)
        # ---------------------------------------------------------------------------------------------

        ## Spacecraft frontal area [m^2]
        self.SCArea = float(0.01)
        # ---------------------------------------------------------------------------------------------

        ## Internal member to track down the status of trajectories (already run or not)
        self.status = False

        ## Internal member to name the current trajectory
        self.name = ""
        
        self.setDefaults()

    def setDefaults(self):
        '''!
            Set input of the class to their pre-defined default value
        '''
        self.velocity = float(10.0)
        self.phaseAngle = float(-90.0)
        self.impact = float(100.0)
        self.altitudeAngle = float(-90.0)
        self.SCArea = float(0.01)
        self.status = False
        self.name = ""

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement root instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node exists in the tree chunk
        trajectorySection = ET.SubElement(xmlElement, self.name)

        # fill velocity leaf
        scvelocity = trajectorySection.find("SpacecraftVelocity")
        if scvelocity is None:
            scvelocity = ET.SubElement(trajectorySection, "SpacecraftVelocity")
        scvelocity.text = "{:.2f}".format(self.velocity)

        # fillpahse angle leaf
        scpangle = trajectorySection.find("PhaseAngleOfApproach")
        if scpangle is None:
            scpangle = ET.SubElement(trajectorySection, "PhaseAngleOfApproach")
        scpangle.text = "{:.2f}".format(self.phaseAngle)

        # fill impact leaf
        scimpact = trajectorySection.find("ImpactParameter")
        if scimpact is None:
            scimpact = ET.SubElement(trajectorySection, "ImpactParameter")
        scimpact.text = "{:.2f}".format(self.impact)

        # fill altitude leaf
        scaltitude = trajectorySection.find("AltitudeAngle")
        if scaltitude is None:
            scaltitude = ET.SubElement(trajectorySection, "AltitudeAngle")
        scaltitude.text = "{:.2f}".format(self.altitudeAngle)

        # fill area leaf
        scarea = trajectorySection.find("SpacecraftFrontalArea")
        if scarea is None:
            scarea = ET.SubElement(trajectorySection, "SpacecraftFrontalArea")
        scarea.text = "{:.2f}".format(self.SCArea)

        # fill internal status leaf
        stleaf = trajectorySection.find("Status")
        if stleaf is None:
            stleaf = ET.SubElement(trajectorySection, "Status")
        stleaf.text = str(self.status)

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance of Trajectory kind
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        self.name = xmlElement.tag

        # check velocity
        scvelocity = xmlElement.find("SpacecraftVelocity")
        if not (scvelocity is None):
            self.velocity = float(scvelocity.text)
        else : 
            return False

        # check phase angle
        scpangle = xmlElement.find("PhaseAngleOfApproach")
        if not (scpangle is None):
            self.phaseAngle = float(scpangle.text)
        else : 
            return False

        # check impact
        scimpact = xmlElement.find("ImpactParameter")
        if not (scimpact is None):
            self.impact = float(scimpact.text)
        else : 
            return False

        # check altitude angle
        scaltitude = xmlElement.find("AltitudeAngle")
        if not (scaltitude is None):
            self.altitudeAngle = float(scaltitude.text)
        else : 
            return False

        # check sc area
        scareaItem = xmlElement.find("SpacecraftFrontalArea")
        if not (scareaItem is None):
            self.SCArea = float(scareaItem.text)
        else : 
            return False

        # check status
        stleaf = xmlElement.find("Status")
        if not (stleaf is None):
            self.status = (stleaf.text.strip() == 'True')
        else : 
            return False

        return True

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        scvelocity = xmlElement.find("SpacecraftVelocity")
        if scvelocity is None:
            return False
        if "{:.2f}".format(self.velocity) != scvelocity.text.strip():
            return False

        scpangle = xmlElement.find("PhaseAngleOfApproach")
        if scpangle is None:
            return False
        if "{:.2f}".format(self.phaseAngle) != scpangle.text.strip():
            return False

        scimpact = xmlElement.find("ImpactParameter")
        if scimpact is None:
            return False
        if "{:.2f}".format(self.impact) != scimpact.text.strip():
            return False

        scaltitude = xmlElement.find("AltitudeAngle")
        if scaltitude is None:
            return False
        if "{:.2f}".format(self.altitudeAngle) != scaltitude.text.strip():
            return False

        scareaItem = xmlElement.find("SpacecraftFrontalArea")
        if scareaItem is None:
            return False
        if "{:.2f}".format(self.SCArea) != scareaItem.text.strip():
            return False

        stleaf = xmlElement.find("Status")
        if stleaf is None:
            return False
        if str(self.status) != stleaf.text.strip():
            return False

        # otherwise
        return True

class TrajectoryWidget(QGroupBox):
    ## custom signal to enable/disable tehe widget if no valid trajectory is present
    validTrajectory = Signal(bool)
    
    '''!
        Widget class collecting User Input needed to build a spacecraft trajectory. It will push collected data
        directly to an instance of Trajectory class connected to the widget through method setTrajectory
    '''  
    def __init__(self):
        '''!
            Constructor
        '''
        super().__init__()

        # default Trajectory parameters
        # ---------------------------------------------------------------------------------------------

        ## Spacecraft velocity [Km/s] input handler - limit values allowed must be within [10,80] Km/s
        self.velocity = QDoubleSpinBox()
        self.velocity.setRange(10.0, 80.0)
        self.velocity.setSingleStep(1.0)
        labelVelocity = QLabel("Spacecraft Velocity")
        labelVelocity.setToolTip("[Km/s] : specifies the magnitude of the spacecraft velocity relative to the comet nucleus")
        # ---------------------------------------------------------------------------------------------

        ## Phase Angle of Approach [deg] input handler - limit values allowed must be within [-90,90] deg
        self.phaseAngle = QDoubleSpinBox()
        self.phaseAngle.setRange(-90.0, 90.0)
        self.phaseAngle.setSingleStep(1.0)
        labelPhaseAngle = QLabel("Phase Angle of Approach")
        labelPhaseAngle.setToolTip("[deg] : specifies the phase angle of approach")
        # ---------------------------------------------------------------------------------------------

        ## Impact parameter [Km] input handler - limit values allowed must be within [100, 5000] Km
        self.impact = QDoubleSpinBox()
        self.impact.setRange(100, 5000)
        self.impact.setSingleStep(1.0)
        labelImpact = QLabel("Impact Parameter")
        labelImpact.setToolTip("[Km] : specifies cometocentric distance of the spacecraft at closest approach")
        # ---------------------------------------------------------------------------------------------

        ## Altitude Angle [deg] input handler - limit values allowed must be within [-90, 90] deg
        self.altitudeAngle = QDoubleSpinBox()
        self.altitudeAngle.setRange(-90.0, 90.0)
        self.altitudeAngle.setSingleStep(1.0)
        labelAltitudeAngle = QLabel("Altitude angle")
        labelAltitudeAngle.setToolTip("[deg] : specifies the phase angle of the spacecraft at the point it crosses the sun-comet line")
        # ---------------------------------------------------------------------------------------------

        ## Spacecraft frontal area [m^2] input handler - limit values allowed must be within [0.01, 10] m^2
        self.SCArea = QDoubleSpinBox()
        self.SCArea.setRange(0.01, 10.0)
        self.SCArea.setSingleStep(0.01)
        labelSCArea = QLabel("Spacecraft frontal area")
        labelSCArea.setToolTip("[m^2] : specifies the frontal area of the spacecraft, impacted by dust")
        # ---------------------------------------------------------------------------------------------

        # Layout form to arrange nicely the input widgets
        layout_standard = QGridLayout()
        layout_standard.setVerticalSpacing(2)
        layout_standard.addWidget(labelVelocity, 0, 0)
        layout_standard.addWidget(self.velocity, 0, 1)
        layout_standard.addWidget(labelPhaseAngle, 1, 0)
        layout_standard.addWidget(self.phaseAngle, 1, 1)
        layout_standard.addWidget(labelImpact, 2, 0)
        layout_standard.addWidget(self.impact, 2, 1)
        layout_standard.addWidget(labelAltitudeAngle, 3, 0)
        layout_standard.addWidget(self.altitudeAngle, 3, 1)
        layout_standard.addWidget(labelSCArea, 4, 0)
        layout_standard.addWidget(self.SCArea, 4, 1)

        self.setLayout(layout_standard)

        ##internal instance of current Trajectory object to modify. 
        self.__trajectory = None

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.velocity.setReadOnly(flag)
        self.phaseAngle.setReadOnly(flag)
        self.impact.setReadOnly(flag)
        self.altitudeAngle.setReadOnly(flag)
        self.SCArea.setReadOnly(flag)

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.velocity.setValue(10)
        self.phaseAngle.setValue(-90)
        self.impact.setValue(100)
        self.altitudeAngle.setValue(-90)
        self.SCArea.setValue(0.01)

    def setTrajectory(self, trajectory):
        '''!
            Set the current trajectory to be managed by the widget
            @param trajectory : instance of Trajectory class type
        '''
        self.__trajectory = trajectory

    def pushData(self):
        '''!
            Write data collected by the graphical widget directly into the Trajectory object previously paired to the class
        '''
        if not self.__trajectory : 
            return
        #compare data with inner data of __trajectory
        if "{:.2f}".format(self.velocity.value()) != "{:.2f}".format(self.__trajectory.velocity) :
            self.__trajectory.velocity = self.velocity.value()
            self.__trajectory.status = False

        if "{:.2f}".format(self.phaseAngle.value()) != "{:.2f}".format(self.__trajectory.phaseAngle) :
            self.__trajectory.phaseAngle = self.phaseAngle.value()
            self.__trajectory.status = False

        if "{:.2f}".format(self.impact.value()) != "{:.2f}".format(self.__trajectory.impact) :
            self.__trajectory.impact = self.impact.value()
            self.__trajectory.status = False

        if "{:.2f}".format(self.altitudeAngle.value()) != "{:.2f}".format(self.__trajectory.altitudeAngle) :
            self.__trajectory.altitudeAngle = self.altitudeAngle.value()
            self.__trajectory.status = False

        if "{:.2f}".format(self.SCArea.value()) != "{:.2f}".format(self.__trajectory.SCArea) :
            self.__trajectory.SCArea = self.SCArea.value()
            self.__trajectory.status = False

    def pullData(self):
        '''!
            Read data collected into the Trajectory object previously paired to the class, and re-expose them into graphical widgets
        '''
        if self.__trajectory : 
            self.velocity.setValue(self.__trajectory.velocity)
            self.phaseAngle.setValue(self.__trajectory.phaseAngle)
            self.impact.setValue(self.__trajectory.impact)
            self.altitudeAngle.setValue(self.__trajectory.altitudeAngle)
            self.SCArea.setValue(self.__trajectory.SCArea)
            self.setTitle(self.__trajectory.name)
            self.validTrajectory.emit(True)
        else:
            self.setDefaults()
            self.setTitle("")
            self.validTrajectory.emit(False)
class FlyByParameters(QWidget):
    '''!
        Collection of input widgets to instrument coma post processing analysis on defined spacecraft fly-by trajectories 
    '''
    def __init__(self):
        '''!
            Constructor.
        '''
        super().__init__()

        ## widget that display all available trajectories created in the form
        self.listTrajectories = QListWidget()
        self.setToolTip("Display all created trajectories. Click on one item to access/edit its contents into the form below")

        ## Command button to add/create a new trajectory to evaluate in fly-by post-processing
        self.addTrajectoryButton = QPushButton("+")
        self.addTrajectoryButton.setToolTip("Add a new trajectory to evaluate")

        ## Command button to remove/stash a previosly created trajectory from fly-by final post-processing
        self.removeTrajectoryButton = QPushButton("-")
        self.removeTrajectoryButton.setToolTip("Remove currently selected trajectory")

        ## Form Widget to edit/visualize currently selected trajectory constitutive parameters
        self.form = TrajectoryWidget()


        ## Run button for post-processing stage. The button is simply a redundancy that triggers the Run S/C fly-by mission action in Run menu
        self.runFlyByButton = QPushButton("Run S/C Fly-By Missions")
        self.runFlyByButton.setToolTip("Evaluate all brand-new created / modified trajectories")

        ## Stop button for post-processing stage. The button is simply a redundancy that triggers the Stop S/C fly-by mission action in Run menu
        self.stopFlyByButton = QPushButton("Stop Missions")
        self.stopFlyByButton.setToolTip("Abort fly-by missions job")
        self.stopFlyByButton.setEnabled(False)

        #connections
        self.addTrajectoryButton.clicked.connect(self.addTrajectoryAction)
        self.listTrajectories.currentRowChanged.connect(self.switchTrajectoryAction)
        self.removeTrajectoryButton.clicked.connect(self.removeTrajectoryAction)
        self.form.validTrajectory.connect(self.__disableForm)

        self.form.setTrajectory(None)
        self.form.pullData()

        ## Internal, list of all possible trajectories to evaluate in this fly-by post-processing session
        self.trajectories = []

        # put input widgets in a clean layout.
        
        # collect controllers for multiple trajectories management
        __controllerWidget = QWidget()

        # Layout form to arrange nicely the input widgets
        __layout_controller = QGridLayout()
        __layout_controller.setVerticalSpacing(2)
        __layout_controller.addWidget(self.addTrajectoryButton, 0, 0)
        __layout_controller.addWidget(QWidget(), 0, 1)
        __layout_controller.addWidget(self.removeTrajectoryButton, 1, 0)
        __layout_controller.addWidget(QWidget(), 1, 1)
        __layout_controller.addWidget(QWidget(), 2, 0)
        __layout_controller.addWidget(QWidget(), 2, 1)

        __controllerWidget.setLayout(__layout_controller)

        __layout_header = QHBoxLayout()
        __layout_header.setContentsMargins(2, 2, 2, 2)
        __layout_header.addWidget(self.listTrajectories)
        __layout_header.addWidget(__controllerWidget)
        
        __header = QWidget()
        __header.setLayout(__layout_header)

        # collect buttons to run and stop
        __buttonWidget = QWidget()

        # Layout form to arrange nicely the buttons
        __layout_buttons = QHBoxLayout()
        __layout_buttons.setContentsMargins(2, 2, 2, 2)
        __layout_buttons.addWidget(self.stopFlyByButton)
        __layout_buttons.addWidget(self.runFlyByButton)
        
        __buttonWidget.setLayout(__layout_buttons)

        # Master layout of the class to help auto sizing of the standardWidget and run button.
        layout_flyby = QVBoxLayout()
        layout_flyby.setContentsMargins(2, 2, 2, 2)
        layout_flyby.addWidget(__header)
        layout_flyby.addWidget(self.form)
        layout_flyby.addSpacing(5)
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Raised)
        layout_flyby.addWidget(line)
        layout_flyby.addSpacing(5)
        layout_flyby.addWidget(__buttonWidget)
        self.setLayout(layout_flyby)


    def addTrajectoryAction(self) :
        '''!
            Action to add a new trajectory with a custom unique name into the self.listTrajectories widget and self.trajectories internal list
        '''
        text = "Name the trajectory (name should be unique and not already assigned to other trajectories"
        name, ok = QInputDialog.getText(self, "Insert trajectory Name", text, QLineEdit.Normal, "trajectory_xxx")
        name = name.strip()
        if not ok or not name:
            QMessageBox.warning(self, "No valid trajectory name inserted","Try again inserting a valid trajectory name.")
            return

        #check if name start with a number, in that case prepend a t charachter on it
        if str(name)[0].isdigit() :
            name = "t"+name

        #check if name is already list as item in self.listTrajectories
        listqitems = self.listTrajectories.findItems(str(name), Qt.MatchExactly)
        if len(listqitems) > 0:
            QMessageBox.warning(self, "Trajectory name already in use!","Try again inserting a different trajectory name.")
            return

        #append in self.trajectories
        self.trajectories.append(Trajectory())
        self.trajectories[-1].name = name           

        #append it in the list
        self.listTrajectories.addItem(name) 
        self.listTrajectories.setCurrentRow(self.listTrajectories.count()-1) #this will trigger signal of currentRowChanged

    def switchTrajectoryAction(self, newRow) :
        '''!
            Action to switch active trajectory in the current form, suitable to be connected with QListWidget currentRowChanged signal
            @param newRow: integer identifying the row of the current active trajectory item in the self.listTrajectories/self.trajectories lists
        '''
        if newRow < 0:
            return
        
        # store data from form widget to the old connected item of self.trajectories
        self.form.pushData() 
        #change the trajectory
        self.form.setTrajectory(self.trajectories[newRow]) 
        #restore data in the form from the new connected trajectory item
        self.form.pullData()

    def removeTrajectoryAction(self) :
        '''!
            Action to remove an active trajectory from the lists self.listTrajectories/self.trajectories
        '''
        rowToDelete = self.listTrajectories.currentRow()
        if rowToDelete < 0:
            return

        #remove item form self.listTrajectories first
        self.listTrajectories.takeItem(rowToDelete) 
        # remove item from self.trajectories
        self.trajectories.pop(rowToDelete) 

        #unconnect the form and call pull data who will restore defaults
        self.form.setTrajectory(None)
        self.form.pullData()

    def __disableForm(self, flag):
        '''!
            Internal utility to enable/disable the trajectory form if a valid/unvalid trajectory is created.
            @param flag, true to enable the form, false to disable it
        '''
        self.form.setEnabled(flag)

    def setDefaults(self):
        '''!
            Set input widgets of the class to their pre-defined default value
        '''
        self.trajectories.clear()
        self.listTrajectories.clear()
        self.form.setTrajectory(None)
        self.form.pullData()

    def setReadOnly(self, flag):
        '''!
            Method to set all the graphical widgets in the form visible to the User, but not interactable
            @param flag : (boolean) if true put all the graphical widgets in the form read only, otherwise make them interactable
        '''
        self.form.setReadOnly(flag)
        # for other stuffs revert to old good enabled/disabled
        enable = str(flag) == 'False'
        self.listTrajectories.setEnabled(enable)
        self.addTrajectoryButton.setEnabled(enable)
        self.removeTrajectoryButton.setEnabled(enable)

    def dumpToXML(self, xmlElement):
        '''!
            Dump input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        #force the last active item to push its data
        self.form.pushData()

        # check if the element node FlyByTrajectories exists in the tree chunk
        flybySection = xmlElement.find("FlyByTrajectories")
        if flybySection is None:
            flybySection = ET.SubElement(xmlElement, "FlyByTrajectories")

        for item in self.trajectories :
            item.dumpToXML(flybySection)

    def restoreFromXML(self, xmlElement):
        '''!
            Restore input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # calling preventively setDefaults in case of errors.
        self.setDefaults()

        # check if the element node FlyByTrajectories exists in the tree chunk
        flybySection = xmlElement.find("FlyByTrajectories")
        if flybySection is None:
            return False

        check = bool(True)
        for child in flybySection:         
            data = Trajectory()
            check = check and data.restoreFromXML(child)
            if check: 
                self.trajectories.append(data)
                self.listTrajectories.addItem(data.name)

        if not check :
            self.setDefaults()
        
        if len(self.trajectories) > 0:
            self.listTrajectories.setCurrentRow(0)

        return check

    def compareWithXML(self, xmlElement):
        '''!
            Compare input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        # check if the element node FlyByTrajectories exists in the tree chunk
        flybySection = xmlElement.find("FlyByTrajectories")
        if flybySection is None:
            return False

        if len(self.trajectories) != len(flybySection.getchildren()) :
            return False            
        
        count = int(0)
        check = bool(True)
        for child in flybySection:         
            check = check and self.trajectories[count].compareWithXML(child)
            count = count + 1
            
        return check
    
    
