from PySide2.QtWidgets import QPlainTextEdit, QPushButton, QWidget
from PySide2.QtWidgets import QVBoxLayout, QHBoxLayout,QFileDialog

class LoggerWidget(QWidget):
    '''!
       Widget to expose verbose/textual feedback of actions done inside the GUI
       context.
       The class has an internal instance log(QPlainTextEdit) to visualize the messages 
       plus two buttons to clear the logger and dump its content to file onto filesystem
    '''
    def __init__(self, message):
        '''!
            Constructor
            @param message initial message to display in the logger
        '''
        super().__init__()

        ##QPlainTextEdit log object to write message with appendPlainText method
        self.log = QPlainTextEdit()
        self.log.setReadOnly(True) #put the class in readonly mode

        ## Button to clear contents of self.log instance
        self.clearButton = QPushButton("Clear")
        self.clearButton.setToolTip("Clear contents of the logger")
        
        ## Button to clear contents of self.log instance
        self.dumpToFileButton = QPushButton("Dump to file")
        self.dumpToFileButton.setToolTip("Dump contents of the current logger to file")

        #edit connections
        self.clearButton.clicked.connect(self.clear)
        self.dumpToFileButton.clicked.connect(self.dumpToFile)
        
        ##Internal instance to hold the initial message
        self.__initialMessage = message

        #define layouts
        buttons_layout = QHBoxLayout()
        buttons_layout.setContentsMargins(1, 1, 1, 1)
        buttons_layout.addWidget(self.clearButton)
        buttons_layout.addWidget(self.dumpToFileButton)

        dummyWidget = QWidget()
        dummyWidget.setLayout(buttons_layout)

        final_layout = QVBoxLayout()
        final_layout.setContentsMargins(1, 1, 1, 1)
        final_layout.addWidget(self.log)
        final_layout.addWidget(dummyWidget)

        self.setLayout(final_layout)

        #print the initial message
        self.log.setPlainText(self.__initialMessage)
    
    def clear(self):
        '''!
            Clear contents of log member
        '''
        self.log.clear()
        self.log.setPlainText(self.__initialMessage)

    def dumpToFile(self):
        '''!
            Dump contents of log member to file chosen by user. A dialog to choose the file will appear on GUI.
        '''
        filename = QFileDialog.getSaveFileName(self, "Dump logger to file", 'commode.log', "Text Files (*.log *.txt *.dat)")

        if filename[0] :
            with open(str(filename[0]), 'w') as file : 
                file.write(self.log.toPlainText())

    def appendPlainText(self, message):
        '''!
            append a message to the logger. Any new message will be appended in a new line
            @param message target message to append
        '''
        self.log.appendPlainText(message)
    

