from PySide2.QtWidgets import QTabWidget

from PySide2.QtGui import QIcon

from frontEnd.ComaParameters import ComaParameters
from frontEnd.FlyByParameters import FlyByParameters

import xml.etree.ElementTree as ET


class ParamTabWidget(QTabWidget):
    '''!
        Top widget form to host inputs needed to run coma simulations and fly-by mission post-processing. The widget is organized
        in dedicated "tabs" according to which task the inputs refer to.
    '''

    def __init__(self, iconaddress):
        '''!
            Constructor.
            @param iconaddress full path where the assets resources can be found
        '''
        super().__init__()
        ## input tab form dedicated to inputs needed by ComMoDE comet simulation
        self.comaParameters = ComaParameters()
        ## input tab form dedicated to inputs needed by ComMoDE Spacecraft fly-by trajectories analysis
        self.flybyParameters = FlyByParameters()
        self.addTab(self.comaParameters, QIcon(
            iconaddress+"setting.png"), "Coma Simulation")
        self.addTab(self.flybyParameters, QIcon(
            iconaddress+"satellite2.png"), "S/C fly-by trajectories")

    def setDefaults(self):
        '''!
            Set tabs and their input sub-sections widgets of the master class to their pre-defined default value
        '''
        self.comaParameters.setDefaults()
        self.flybyParameters.setDefaults()

    def dumpComaToXML(self, xmlElement):
        '''!
            Dump Coma Form and its sub-sections input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("ComaForm")
        if section is None:
            section = ET.SubElement(xmlElement, "ComaForm")

        self.comaParameters.dumpToXML(section)

    def dumpFlyByToXML(self, xmlElement):
        '''!
            Dump FlyBy Trajectories Form and its sub-sections input widgets contents into a Xml Element section istance
            @param xmlElement instance to dump input widgets contents (xml.etree.ElementTree Element)
        '''
        # check if the element node SizeDistribution exists in the tree chunk
        section = xmlElement.find("FlyByForm")
        if section is None:
            section = ET.SubElement(xmlElement, "FlyByForm")

        self.flybyParameters.dumpToXML(section)

    def restoreComaFromXML(self, xmlElement):
        '''!
            Restore Coma Fors and its sub-sections input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # check if the element node exists in the tree chunk
        section = xmlElement.find("ComaForm")
        if section is None:
            return False

        flag = self.comaParameters.restoreFromXML(section)
        return flag

    def restoreFlyByFromXML(self, xmlElement):
        '''!
            Restore FlyBy Trajectories and its sub-sections input widgets values from a XML Element instance
            @param xmlElement instance to restore input widgets contents (xml.etree.ElementTree Element)
            @return True if restoring was successfull or False if something went wrong
        '''
        # check if the element node exists in the tree chunk
        section = xmlElement.find("FlyByForm")
        if section is None:
            return False

        flag = self.flybyParameters.restoreFromXML(section)
        return flag

    def compareComaWithXML(self, xmlElement):
        '''!
            Compare Coma form and its sub-sections input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("ComaForm")
        if section is None:
            return False

        flag = self.comaParameters.compareWithXML(section)
        return flag

    def compareFlyByWithXML(self, xmlElement):
        '''!
            Compare FlyBy Trajectories form and its sub-sections input widgets values with those provided into a XML Element instance
            @param xmlElement instance to compare data with (xml.etree.ElementTree Element)
            @return True if the values hold by the class matching those of the XML Element, False otherwise.
        '''
        section = xmlElement.find("FlyByForm")
        if section is None:
            return False

        flag = self.flybyParameters.compareWithXML(section)
        return flag

