from PySide2.QtCore import QObject, Signal
import os

class StatusProject(QObject):
    '''!
        Utility class to track down the status of coma simulation and fly-by trajectories post-processing runs. 
        Specifically, it expose two booleans validComa and validFlyBy that can be True or False depending on the 
        outcome of the runs.
    '''
    ## Signal emitted everytime validComa variable changes.
    comaChanged = Signal()
    ## Signal emitted everytime validFlyBy variable changes.
    flyByChanged = Signal()
    
    def __init__(self):
        '''!
            Constructor
        '''
        super(StatusProject, self).__init__()

        ## Private validComa bool variable: set to true everytime coma simulation is run successfully
        self.__validComa = False
        ## Private validFlyBy bool variable: set to true everytime trajectories post-processing is run successfully
        self.__validFlyBy = False
        
    def updateComaStatus(self, flag) :
        '''!
            Method to update the __validComa member
            @param flag valid boolean 
        '''
        self.__validComa = flag 
        self.comaChanged.emit()

    def updateFlyByStatus(self, flag) :
        '''!
            Method to update the __validFlyBy member
            @param flag valid boolean 
        '''
        self.__validFlyBy = flag 
        self.flyByChanged.emit()

    def writeOnFile(self, dirpath) :
        '''!
            Print current values of __validComa and __validFlyBy to a .status_project file located in the dirpath root directory
            @param dirpath full path to directory where the .status_project file will be written.
        '''
        if dirpath and os.path.isdir(dirpath) :
            with open(os.path.join(dirpath, ".status_project"), 'w') as statfile : 
                statfile.write(str(self.__validComa)+"\n")
                statfile.write(str(self.__validFlyBy)+"\n")

    def restoreFromFile(self, dirpath) :
        '''!
            Restore current values of __validComa and __validFlyBy from a .status_project file located in the dirpath root directory
            @param dirpath full path to directory where the .status_project file is.
            @return boolean True/False if restoring was successfull
        '''            
        statfile = os.path.join(dirpath, ".status_project")
        if not os.path.exists(statfile):
            return False
 
        file = open(statfile, 'r') 
        lines = file.readlines()
        lines = [line.strip() for line in lines] 
       
        self.updateComaStatus(str(lines[0]) == 'True')
        self.updateFlyByStatus(str(lines[1]) == 'True')

    def getComaStatus(self) :
        '''!
            @return value of internal member __validComa
        '''            
        return self.__validComa

    def getFlyByStatus(self) :
        '''!
            @return value of internal member __validFlyBy
        '''            
        return self.__validFlyBy
