import os, random

import numpy as np
import pandas 

from PySide2.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QGridLayout, QListWidget,QAbstractItemView
from PySide2.QtWidgets import QAction, QComboBox, QLabel, QPushButton, QMessageBox
from PySide2.QtWidgets import QDialog, QColorDialog, QCheckBox, QScrollArea
from PySide2.QtGui import QIcon, QColor
from PySide2.QtCore import Qt, QSize, QObject, Slot, Signal

import matplotlib 
matplotlib.use("Qt5Agg")
from matplotlib.backends.backend_qt5agg import (FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
from matplotlib.figure import Figure

# adjusting css of matplotlib -> should be moved to css definition
matplotlib.rcParams["text.color"] = '#eff0f1'
matplotlib.rcParams["legend.labelcolor"] = '#eff0f1'
matplotlib.rcParams["legend.facecolor"] = '#31363b'
matplotlib.rcParams["axes.labelcolor"] = '#eff0f1'
matplotlib.rcParams["axes.edgecolor"] = '#76797C'
matplotlib.rcParams["axes.titlecolor"] = '#eff0f1'
matplotlib.rcParams["axes.facecolor"] = '#232629'
matplotlib.rcParams["figure.facecolor"] = '#31363b'
matplotlib.rcParams["grid.linestyle"] = ':'
matplotlib.rcParams["xtick.color"] = '#eff0f1'
matplotlib.rcParams["ytick.color"] = '#eff0f1'
#####################################################################

class AxisInfoStruct(QObject):
    '''!
        C-struct like container to host options to define cartesian plot axis
    '''
    def __init__(self):
        '''!
            Constructor
        '''
        QObject.__init__(self)

        ## name identifying axis would be x or y
        self.name = str('')
        ## string to identify the data name attached to the axis
        self.dataname = str('Density')
        ## index to identify the data name attached to the axis
        self.dataindex = int(0)
        ## boolean to identify if the axis is logarithmic or not.
        self.logarithmic = bool(False)

class AxisDialog(QDialog):
    '''!
        Utility dialog to select the axis data type to visualize on canvas 2D and if its logorathmic or not.
    '''
    def __init__(self, parent, axisInfoObject, tooltiplist):
        '''!
            Constructor
            @param parent QPushButton class from which the current dialog depends
            @param axisInfoObject AxisInfoStruct instance from which reading initial options and store the new modified ones.
            @param tooltiplist list of tooltips to attach to the internal list of 6 data available, e.g Density, Fluc, Fluence, R, S and time.
        '''
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle("ComMoDE : Setup "+axisInfoObject.name+" axis options")
        self.setFixedSize(300,200)
        
        ## store internal instance of the PlotCanvas2D parent class
        self.axisInfo = axisInfoObject
        
        ## Drop down menu object listing all possible quantities representable on axis : namely dust_density, flux, fluence, R and S coordinates, time CA
        self.dataChoice = QComboBox()
        self.dataChoice.addItem("Density")
        self.dataChoice.addItem("Flux")
        self.dataChoice.addItem("Fluence")
        self.dataChoice.addItem("R")
        self.dataChoice.addItem("S")
        self.dataChoice.addItem("Time")

        #tooltipping entries to push more information
        self.dataChoice.setItemData(0, tooltiplist[0], Qt.ToolTipRole)
        self.dataChoice.setItemData(1, tooltiplist[1], Qt.ToolTipRole)
        self.dataChoice.setItemData(2, tooltiplist[2], Qt.ToolTipRole)
        self.dataChoice.setItemData(3, tooltiplist[3], Qt.ToolTipRole)
        self.dataChoice.setItemData(4, tooltiplist[4], Qt.ToolTipRole)
        self.dataChoice.setItemData(5, tooltiplist[5], Qt.ToolTipRole)

        # set the list entry to the the value corresponding inside self.axisInfo.
        itemFound = self.dataChoice.findText(self.axisInfo.dataname)
        self.dataChoice.setCurrentIndex(itemFound)

        ## Checkbox to define the current axis as logarithmic or regular spaced
        self.isLog = QCheckBox("Logarithmic")
        self.isLog.setToolTip("enabled -> define the axis as logarithmic. BEWARE:any non-positive data will be automatically excluded from the drawing.")

        # check/uncheck according to the value reported in self.axisInfo.
        if self.axisInfo.logarithmic : 
            self.isLog.setCheckState(Qt.Checked)
        else : 
            self.isLog.setCheckState(Qt.Unchecked)


        #create layout of accept/reject buttons
        ## Accept button widget, to accept modifications
        self.acceptButton = QPushButton("Ok")
        ## Cancel button widget, to refuse modifications
        self.rejectButton = QPushButton("Cancel")

        self.acceptButton.released.connect(self.storeAndAccept)
        self.rejectButton.released.connect(self.reject) 

         ## internal layout for accept/cancel buttons
        __layout_buttons = QHBoxLayout()
        __layout_buttons.addWidget(self.acceptButton)
        __layout_buttons.addWidget(self.rejectButton)
        # internal instance of QWdiget holding the __layout_buttons layout
        __buttonsWidget = QWidget()
        __buttonsWidget.setLayout(__layout_buttons)

        ## internal final_layout to attach to the dialog
        __final_layout = QVBoxLayout()
        __final_layout.addWidget(self.dataChoice)
        __final_layout.addWidget(self.isLog)
        __final_layout.addWidget(__buttonsWidget)
        self.setLayout(__final_layout)

    def storeAndAccept(self):
        '''!
            internal execution command, connected to any choice made in modal dialog
            It will store the selected data name and if its logarithmic or not inside the internal instance of self.axisInfo.
            an emit a signal of sizeBin value changed
        '''
        self.axisInfo.dataname = self.dataChoice.currentText()
        self.axisInfo.dataindex = self.dataChoice.currentIndex()
        if  self.isLog.isChecked() :
            self.axisInfo.logarithmic = True
        else:
            self.axisInfo.logarithmic = False

        self.accept()    

class CustomToolbar(NavigationToolbar):
    '''!
        Build the custom Home/Pan/Zoom/Save toolbar attached onto the visualization canvas, 
        adding some more actions to it: 
        1) 2 drop down menus to choose the x-y data to visualize in 2D environment 
        2) 1 drop down ment to select the current active size bin to visualize
        3) Action to open a sub-widget to handle multiple trajectories results
        3) Visualize/Hide the 2D grid onto the 2D environment
    '''
    ## list of default controllers/actions inherited from matplotlib base class NavigationToolbar
    toolitems = [t for t in NavigationToolbar.toolitems if
                 t[0] in ('Pan', 'Zoom', 'Save')]
    
    axesChanged = Signal()
    def __init__(self, iconaddress, *args, **kwargs):
        '''!
            Constructor. 
            @param iconaddress path location of icons resources used.
            @param args not used (legacy of base NavigationToolbar class)
            @param kwargs not used (legacy of base NavigationToolbar class)
        '''
        super(CustomToolbar, self).__init__(*args, **kwargs)

        #change icons to pre-existent action items
        newIcons = {
            "Pan"  : QIcon(iconaddress+"four-arrows.png"),
            "Zoom" : QIcon(iconaddress+"zoomsel.png"),
            "Save" : QIcon(iconaddress+"Snapshot.png")
        }

        #track first and last action object to add separators onto it.
        for action in self.actions() :
            if action.text() in newIcons:
                action.setIcon(newIcons.get(action.text(), QIcon()))
                if(action.text() == "Save"):
                    lastAction = action
                if(action.text() == "Pan"):
                    firstAction = action

        firstAction = self.insertSeparator(firstAction)
        lastAction = self.insertSeparator(lastAction)


        self.dataTips = []
        self.dataTips.append("Dust Density [m^-3]")
        self.dataTips.append("Dust Flux [m^-2 s^-1]")
        self.dataTips.append("Dust Fluence [m^-2]")
        self.dataTips.append("R coordinate [m]")
        self.dataTips.append("S coordinate [m]")
        self.dataTips.append("Time To Closest Approach [s]")

        ## internal Widget carrying label + button to enable the dialog to edit X Axis options
        self.__XDataWidget = QWidget()
        ## data object carrying the options attached to the axis
        self.XAxisInfo = AxisInfoStruct()
        self.XAxisInfo.name = "x"
        ## internal label object of x axis
        __xlabel = QLabel("x: ")
        ## Button to access x axis options (quantity to visualize and if it's logarithmic or not)
        self.XAxisButton = QPushButton(self.XAxisInfo.dataname)
        self.XAxisButton.setToolTip("Click to access the dialog to change options related to x axis")
        
        ## internal layout of __XDataWidget
        __layoutX = QHBoxLayout()
        __layoutX.addWidget(__xlabel)
        __layoutX.addWidget(self.XAxisButton)
        self.__XDataWidget.setLayout(__layoutX)

        ## internal Widget carrying label + button to enable the dialog to edit Y Axis options
        self.__YDataWidget = QWidget()
        ## data object carrying the options attached to the axis
        self.YAxisInfo = AxisInfoStruct()
        self.YAxisInfo.name = "y"
        ## internal label object of y axis
        __ylabel = QLabel("y: ")
        ## Button to access y axis options (quantity to visualize and if it's logarithmic or not)
        self.YAxisButton = QPushButton(self.YAxisInfo.dataname)
        self.YAxisButton.setToolTip("Click to access the dialog to change options related to y axis")
        
        ## internal layout of __YDataWidget
        __layoutY = QHBoxLayout()
        __layoutY.addWidget(__ylabel)
        __layoutY.addWidget(self.YAxisButton)
        self.__YDataWidget.setLayout(__layoutY)
        
        ## Drop down menu object listing data to fix to enable visualization. If Switch mode is off, the menu will retain all possible bin distribution sizes to fix, otherwise it it's on, all possible trajectories to fix.
        self.DataChoice = QPushButton()

        ## Action to toggle on/off the visualization mode to be used. If OFF, it will let to analyze data from multiple trajectories, once the distribution bin size is fixed. If Off, it will let to analize data from multiple bin size, once a trajectory is fixed.
        self.switchVizMode = QAction(QIcon(iconaddress+"switch-vector.png"), "Toggle on/off to change plot mode" )
        self.switchVizMode.setToolTip("Toggle on/off to change plot mode. On: fix trajectory and analyze multiple size bin data. Off: fix bin size, and analyze multiple trajectory data")
        self.switchVizMode.setCheckable(True)

        ## inserting x,y and bin widgets to appear as Bin, x, y in the toolbar
        firstAction = self.insertWidget(firstAction, self.__YDataWidget)
        firstAction = self.insertWidget(firstAction, self.__XDataWidget)
        firstAction = self.insertWidget(firstAction, self.DataChoice)

        # internal empty object just to pad switchVizMode from DataChoice
        __internalPad = QWidget()
        __internalPad.setFixedWidth(15)
        firstAction = self.insertWidget(firstAction, __internalPad)

        # insert the switch action in the toolbar
        firstAction = self.insertAction(firstAction, self.switchVizMode)

        ## Action object to reset current view to host completely data in the figure
        self.resetAction = QAction(QIcon(iconaddress+"reset-icon-white.png"), "Reset home view")
        self.resetAction.setToolTip("Reset the axis to view all data in the figure ")
        
        self.insertAction(lastAction, self.resetAction)

        ## Action object to enable/disable visualization of data from multiple trajectories
        self.multiMissionsManagerAction = QAction(QIcon(iconaddress+"lineschart.png"), "Managing visualization of multiple data")
        self.multiMissionsManagerAction.setToolTip("Managing visualization of multiple data")
        self.insertAction(self.resetAction, self.multiMissionsManagerAction)

        ## Action object to enable/disable visualization of grid in the plot canvas
        self.gridToggleAction = QAction(QIcon(iconaddress+"grid.png"), "Toggle grid on/off")
        self.gridToggleAction.setToolTip("Toggle grid on/off")
        self.gridToggleAction.setCheckable(True)
        
        self.insertAction(self.multiMissionsManagerAction, self.gridToggleAction)

        # finally connect Axis push button to action opening dialogs.
        self.XAxisButton.clicked.connect(self.openXAxisInfoDialog)
        self.YAxisButton.clicked.connect(self.openYAxisInfoDialog)
    
    
    def openXAxisInfoDialog(self):
        '''!
        Open a modal dialog to modify the visualization option of the X axis in the cartesian plot.
        The user can modify the type of data to visualize in X axis and if the current axis logarithmically spaced or not. 
        '''    
        dialog = AxisDialog(self, self.XAxisInfo, self.dataTips)
        response = dialog.exec_()
        if response :
            self.XAxisButton.setText(self.XAxisInfo.dataname)
            self.axesChanged.emit()

    def openYAxisInfoDialog(self):
        '''!
        Open a modal dialog to modify the visualization option of the Y axis in the cartesian plot.
        The user can modify the type of data to visualize in X axis and if the current axis logarithmically spaced or not. 
        '''    
        dialog = AxisDialog(self, self.YAxisInfo, self.dataTips)
        response = dialog.exec_()
        if response :
            self.YAxisButton.setText(self.YAxisInfo.dataname)
            self.axesChanged.emit()

class TrajectoryVisualizationStruct(QObject):
    '''!
        C-struct like container to define significant data for a single trajectory during visualization
    '''
    def __init__(self):
        '''!
            Constructor
        '''
        QObject.__init__(self)

        ## name identifying trajectory
        self.name = str('')
        ## path to folder where txt files data from ComMoDe fly-by post-processing are collected
        self.folder = str('')
        ## hex string defining the color associated to the trajectory
        self.color = str('')
        ## boolean to mark if the trajectory is enabled/disabled for visualization
        self.status = bool(False)

    def getActiveBinIndexFile(self, binIndex):
        '''!
            Return the trajectory file "dust_density_along_trajectoryX.txt" path associated to the bin index X passed 
            in argument. The search will be done in self.folder of the class. 
            @param binIndex integer represent the current active size bin index to search
            @return data file of trajctory associated to the bin if found, None otherwise
        '''
        if not self.folder:
            return None

        name = "dust_density_along_trajectory"+str(binIndex)+".txt"
        filepath = os.path.join(self.folder, name) 

        if not os.path.exists(filepath) : 
            return None
        
        return filepath

class BinSizeVisualizationStruct(QObject):
    '''!
        C-struct like container to define significant data for a bin size during mulitple visualization
    '''
    def __init__(self):
        '''!
            Constructor
        '''
        QObject.__init__(self)

        ## stringfied value of the bin size
        self.name = str('')
        ## hex string defining the color associated to the bin size
        self.color = str('')
        ## boolean to mark if the bin size is enabled/disabled for visualization
        self.status = bool(False)

class PlotCanvas2D(QWidget):
    '''! 
        Create a standard 2D plot environment, for cartesian representation of data. 
        It embeds a command toolbar inside of class type CustomToolbar.
        Data to visualize are directly parsed from trajectory files located in 
        a standard ComMoDe project folder ->fly-by results sub-folder (use method parseTrajectories). 
    '''
    ## custom signal of the class to track down when a new active bin size is chosen
    activeFixedDataChanged = Signal()

    def __init__(self, iconaddress):
        '''!
            Constructor
             @param iconaddress path location of icons resources used.
        '''    
        super().__init__()

        ## internal member to store path to icons resources
        self.__iconaddress = iconaddress
        ## internal layout of the canvas
        __layout = QVBoxLayout(self)

        ## Area to draw 2D cartesian plots
        self.figure = Figure(dpi=90, layout='tight')
        ## Canvas Widget hosting the figure member drawings.
        self.canvas = FigureCanvas(self.figure)
        ## Custom toolbar object attached to visualization environment, collecting all of its controllers (see class CustomToolbar).    
        self.tbar = CustomToolbar(self.__iconaddress, self.canvas, self)
       
        __layout.addWidget(self.tbar)
        __layout.addWidget(self.canvas)
        self.setLayout(__layout)

        ## Axes of the cartesian plot, defined in self.figure member 
        self.axes = self.figure.subplots()
        self.axes.grid(False)

        ## Member to store path of ComMode active project folder
        self.projectFolder = ''

        ## List of all trajectories available.Each trajectory data is stored in an list item of type TrajectoryVisualizationStruct
        self.trajectories = []

        ## List of possible bin sizes available. Meta data info (color, isActive, bin size value) are collected in a item of type BinSizeVisualizationStruct
        self.binlist = []

        ## List of all raw data associated to trajectories according to the active bin size selected by the User, or to bin sizes of a single trajectory selected by the User. List item type should be pandas.DataFrame
        self.rawdata = []
        # note for devs: to save memory in multi data visualization the active trajectories/binsizes selected with The MultiDataDialog 
        # will be equipped with raw data (pandas DataFrame). This job is done by the method uploadRawData.
        # to keep the indexing of rawdata list synchronized with trajectories/bin size list, the i-th disabled trajectoried will be put on None.

        ## internal member carrying the index of fixed data (trajectory or bin size) according to visualizing mode(bin size for multiple trajectories, and trajectory for multiple bin sizes)       
        self.__activeFixedDataIndex = 0

        # fill trajectories and bin list
        self.parseTrajectories()

        # refresh plot
        self.updateFigure()

        #attach slot to custom actions of the toolbar (the other inherited actions are already handled)
        self.tbar.resetAction.triggered.connect(self.refreshView)
        self.tbar.gridToggleAction.triggered.connect(self.gridToggleSlot)
        self.tbar.axesChanged.connect(self.updateFigure)
        self.tbar.axesChanged.connect(self.updateFigure)
        self.activeFixedDataChanged.connect(self.refreshData)
        self.tbar.multiMissionsManagerAction.triggered.connect(self.openMultipleDataDialog)
        self.tbar.DataChoice.clicked.connect(self.openDataToFixDialog)
        self.tbar.switchVizMode.triggered.connect(self.changeVisualizationMode)


    def changeVisualizationMode(self):
        '''!
            According to the state of switchVizMode object in the toolbar, if switch is toggled on: 
            1) Class will be in multiple bin sizes data visualization mode, for a fixed trajectory. 
            2) The DataChoice widget will let the user to fix the trajectory from those available.
            3) the MultipleMissionManager will let the User to visualize data from N different bin sizes available for the trajectory. 

            Otherwise, if switch is toggled off: 
            1) Class will be in multiple trajectories data visualization mode, for a fixed bin size. 
            2) The DataChoice widget will let the user to fix the bin size from those available.
            3) the MultipleMissionManager will let the User to visualize data from N different trajectories. 
        '''

        #reshape raw data to the correct list used e.g. binsizes or trajectories.
        self.rawdata.clear()

        if self.tbar.switchVizMode.isChecked() :
            self.tbar.DataChoice.setText("traj: ")
            self.tbar.DataChoice.setToolTip("Fix a trajectory from those available. Multiple bin sizes data referring to this trajectory will be drawn")
            self.rawdata = [None]*len(self.binlist)
        else :
            self.tbar.DataChoice.setText("bin: ")
            self.tbar.DataChoice.setToolTip("Fix a bin [m] from cometary dust size distribution. Multiple trajectories data for this bin will be drawn")
            self.rawdata = [None]*len(self.trajectories)


        #this will force to update the raw data from files.
        self.setActiveFixedDataIndex(0)
        #there's no concurrent access in visualizing while the trajectories data are being processed. So it's safe to leave the chunk
        # of data to visualize (trajectories or bin sizes) with the previous active status.

    def refreshView(self):
        '''!
            simply call a self.canvas.show()
        ''' 
        self.axes.autoscale(True)
        self.canvas.draw()
        self.axes.autoscale(False)

    def gridToggleSlot(self):
        '''!
            Instructions to enable/hide the grid in 2D cartesian graph 
        ''' 
        if self.tbar.gridToggleAction.isChecked() :
            self.axes.grid(True)
        else:
            self.axes.grid(False)
        self.canvas.draw()

    def cleanTrajectories(self):
        '''! 
            Clear up internal members trajectories and rawdata, empty the list of sizebins, put the activeFixedDataIndex to 0
        '''
        self.binlist.clear()
        self.__activeFixedDataIndex = 0
        self.trajectories.clear()
        self.rawdata.clear()

    def parseTrajectories(self):
        '''! 
            Scan projectFolder and recover information on: 
            - active size bins from coma simulation. Fill the members with BinSizeVisualizationStructItemType. 
            - active trajectory sub-folders in flyby-results folder of the project. Fill the member trajectories
            with item of type TrajectoryVisualizationStruct.
            - reshape also the rawdata list, which will have size as trajectories/binsizes according to the visualization mode
            - if trajectories are present, according to visualization mode, fix the first bin size and enable the first trajectory 
            of the pool, uploading its raw data OR fix the first trajectory and enable the first bin size of the pool, uploading its 
            raw data
        '''
        if not self.projectFolder: 
            self.cleanTrajectories()
            return

        try:
            #get the particle bins first.    
            comaheader = os.path.join(self.projectFolder, "coma-results", "header.txt")
            with open(comaheader, 'r') as file:
                rawlines = file.read().split('\n')
            
            sizebinstring = ""
            for item in rawlines:
                if "particle_size_bins[m]:" in item:
                    sizebinstring = item

            __list = sizebinstring.replace("particle_size_bins[m]:", "").replace(",", " ").strip().split()
            
            ## filling binsizes metadata
            for item in __list :
                bdata = BinSizeVisualizationStruct()
                bdata.name = "{:.4E}".format(float(item))
                bdata.status = False
                bdata.color = "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])
                self.binlist.append(bdata)

            ## filling trajectories meta data
            flybyfolder = os.path.join(self.projectFolder, "flyby-results")

            tfolderslist = os.listdir(flybyfolder)
            for item in tfolderslist:
                tdata = TrajectoryVisualizationStruct()
                tdata.name = item
                tdata.folder= os.path.join(flybyfolder, item)
                tdata.status = False 
                tdata.color = "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])
                self.trajectories.append(tdata)

            # reshape rawdata
            self.rawdata.clear()
            if self.tbar.switchVizMode.isChecked() :
                self.rawdata = [None]*len(self.binlist)
            else :
                self.rawdata = [None]*len(self.trajectories)

            #if the list are not empty, position the first item as "visualizable"
            if  len(self.trajectories) > 0 :
                self.trajectories[0].status = True 
 
            if  len(self.binlist) > 0 :
                self.binlist[0].status = True 
                
            #fixing the Fixed Data to the first item if any
            #this will emit signal and force the upload of raw data
            self.setActiveFixedDataIndex(0)

        except: 
            self.cleanTrajectories()
            QMessageBox.critical("PlotCanvas2D Impossible to parse trajectories data", "Cannot parse trajectories data from flyby-results of project folder {}".format(self.projectFolder))


    def uploadRawData(self):
        '''!
            According the visulization mode, the method will parse the right dust_density_along_trajectoryX.txt files 
            whose data need to be visualized. 
            RawData can be driven by request to visualize per bin sizes or per trajectories. 
            So the method will scan the trajectories available. 
            If the user is in  bin sizes mode:
            1) the method will retrieve the target trajectory index from self.__activeFixedDataIndex, and 
            access data in self.trajectories list.
            2) reshape the self.rawdata as the number of bin sizes currently available (self.binlist size)
            3) for bin sizes marked as active, use pandas to parse data from the file dust_density_along_trajectoryX.dat, 
            where X is the positional index of i-th binsize in the self.binlist size. 
            4) Unactive bin sizes will be associated to None rawdata entries
            On the other hand, If the user is in trajectories mode:
            1) the method will retrieve the target binsize index from self.__activeFixedDataIndex, and 
            access data in self.binlist list.
            2) reshape the self.rawdata as the number of trajectories currently available (self.trajectories size)
            3) for trajectories marked as active, use pandas to parse data from the file dust_density_along_trajectoryX.dat, 
            where X is the self.__activeFixedDataIndex to access the self.binlist. 
            4) Unactive trajectories will be associated to None rawdata entries
            
            The method will consider reshaped self.rawdata with self.trajectories/self.binlist, according to visualization mode.
            The reshaping is done typically in parseTrajectories/cleanTrajectories and in changeVisualization mode.
        '''    
        count = int(0)
        
        if self.tbar.switchVizMode.isChecked() :
            for b in self.binlist :
                # trajectory is fixed, looping on binsizes
                filepath = self.trajectories[self.__activeFixedDataIndex].getActiveBinIndexFile(count)
                if b.status and filepath :
                    self.rawdata[count] = pandas.read_table(filepath, sep=",", header=None).astype(float).dropna()
                    #note for devs, homogenize data in float types, than purge NaNs
                else:
                    self.rawdata[count] = pandas.DataFrame()
                count = count +1
        else :
            # bin size is fixed, looping on trajectories
            for t in self.trajectories :
                filepath = t.getActiveBinIndexFile(self.__activeFixedDataIndex)
                if t.status and filepath :
                    self.rawdata[count] = pandas.read_table(filepath, sep=",", header=None).astype(float).dropna()
                    #note for devs, homogenize data in float types, than purge NaNs
                else:
                    self.rawdata[count] = pandas.DataFrame()
                count = count +1

    def refreshData(self):
        '''!
            Compound method to call uploadRawData and updateFigure methods in one call
        '''
        self.uploadRawData()
        self.updateFigure()

    def openMultipleDataDialog(self):
        '''!
             Open a MultipleDataDialog widget controlling multiple trajectories/binsizes data visualization,
             according to the mode chosen by the class with switchVisualizationMode.
        '''
        dialog = MultipleDataDialog(self, self.__iconaddress)
        response = dialog.exec_()
        if response :
            self.uploadRawData()
            self.updateFigure()

    def openDataToFixDialog(self):
        '''!
             Open a DataToFixDialog widget to choose bin size/trajectory to fix in visualization mode, according to 
             the mode chosen by the class with switchVisualizationMode.
             Closing the modal dialog will automatically fill the activeFixedData index member
        '''
        if self.tbar.switchVizMode.isChecked() :
            #fixing trajectory
            dialog = DataToFixDialog(self, self.trajectories)
        else: 
            #fixing binsize
            dialog = DataToFixDialog(self, self.binlist)

        dialog.exec_()
        
    def getActiveFixedDataIndex(self) :
        '''!
            @return active bin size used to visualize trajectory data
        
        '''
        return self.__activeFixedDataIndex

    def setActiveFixedDataIndex(self, index) :
        '''!
            Set active "fixed data" according to visualization mode (can be bin size or trajectory). Contextually emit a signal of activeFixedDataChanged
            @param index of the new bin size/trajectory to consider as active
        '''

        if self.tbar.switchVizMode.isChecked():
            #working in fixed trajectory and multiple bin size to visualize
            size = len(self.trajectories) 
        else:
            #working in fixed bin size and multiple trajectories to visualize
            size = len(self.binlist)


        if index >= 0 and index <size:
            self.__activeFixedDataIndex = index 
            
            if self.tbar.switchVizMode.isChecked():
                self.tbar.DataChoice.setText("traj: "+ self.trajectories[index].name)
            else: 
                self.tbar.DataChoice.setText("bin: "+ self.binlist[index].name)

            self.activeFixedDataChanged.emit()


    def updateFigure(self):
        '''!
            Redraw Figure in the canvas using information of self.trajectories and self.rawdata already acquired. 
            The method will not parse trajectories, nor refresh raw data to visualize, please use the dedicated methods. 
        ''' 
        self.axes.cla()
        xindex = self.tbar.XAxisInfo.dataindex
        yindex = self.tbar.YAxisInfo.dataindex

        if self.tbar.XAxisInfo.logarithmic : 
            self.axes.set_xscale('log')
        else:
            self.axes.set_xscale('linear')    

        if self.tbar.YAxisInfo.logarithmic : 
            self.axes.set_yscale('log')
        else:
            self.axes.set_yscale('linear')    

        count = int(0)
        for data in self.rawdata:
            if self.tbar.switchVizMode.isChecked() :
                colorData = self.binlist[count].color
                nameData = self.binlist[count].name
            else :
                colorData = self.trajectories[count].color
                nameData = self.trajectories[count].name

            if not data.empty :
                self.axes.plot(data[xindex].to_numpy(), data[yindex].to_numpy(), color=colorData, label=nameData)

            count = count + 1   
        
        ###########################################################
        if count > 0 :
            self.axes.legend()
        
        self.gridToggleSlot()
        __indexX = self.tbar.XAxisInfo.dataindex
        self.axes.set_xlabel(self.tbar.dataTips[__indexX])
        __indexY = self.tbar.YAxisInfo.dataindex
        self.axes.set_ylabel(self.tbar.dataTips[__indexY])

        self.canvas.draw()

class DataToFixDialog(QDialog):
    '''!
        Utility dialog to select the fixed bin size or trajectory for the current visualization mode
    '''
    def __init__(self, parent, datalist):
        '''!
            Constructor
            @param parent PlotCanvas2D class from which the current dialog depends
            @param datalist list of BinSizeVisualizationStruct or TrajectoryVisualization struct to populate QListWidget self.list member
        '''
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle("ComMoDE : Choose the Data to fix (bin size or trajectory)")
        self.setFixedSize(300,400)
        
        ## store internal instance of the PlotCanvas2D parent class
        self.canvas2D = parent
        
        ## list of checkbox Widgets associated to each entry of datalist in input
        self.list = QListWidget()   
        self.list.setSelectionMode(QAbstractItemView.SingleSelection)

        __formattedlist = []
        for item in datalist:
            __formattedlist.append(item.name)

        self.list.insertItems(0,__formattedlist)
        self.list.setCurrentRow(self.canvas2D.getActiveFixedDataIndex())
 
        #create layout of accept/reject buttons
        ## Accept button widget, to accept modifications
        self.acceptButton = QPushButton("Ok")
        ## Cancel button widget, to refuse modifications
        self.rejectButton = QPushButton("Cancel")

        self.acceptButton.released.connect(self.storeAndAccept)
        self.rejectButton.released.connect(self.reject) 

         ## internal layout for accept/cancel buttons
        __layout_buttons = QHBoxLayout()
        __layout_buttons.addWidget(self.acceptButton)
        __layout_buttons.addWidget(self.rejectButton)
        # internal instance of QWdiget holding the __layout_buttons layout
        __buttonsWidget = QWidget()
        __buttonsWidget.setLayout(__layout_buttons)

        ## internal final_layout to attach to the dialog
        __final_layout = QVBoxLayout()
        __final_layout.addWidget(self.list)
        __final_layout.addWidget(__buttonsWidget)
        self.setLayout(__final_layout)

    def storeAndAccept(self):
        '''!
            internal execution command, connected to any selection of item in self.list
            It will store the selected item of the list in internal member activeFixedDataIndex of PlotCanvas2D
            an emit a signal of activeFixedData value changed
        '''
        row = self.list.currentRow()
        self.canvas2D.setActiveFixedDataIndex(row)
        self.accept()

class MultipleDataDialog(QDialog):
    '''!
        Utility dialog to control visualization of data from multiple trajectories/bin sizes available in ComMoDe.
        The user have the possibility to: 
        - enable/disable the visualization of a particular item (trajectory or bin size)
        - change its color
    '''
    def __init__(self, parent, iconaddress):
        '''!
            Constructor
            @param parent PlotCanvas2D class from which the current dialog depends
            @param iconaddress path location of icons resources used. 
        '''
        super().__init__(parent)
        self.setModal(True)
        
        ## store pointer to the right list of objects (trajectories or binsize) form PlotCanvas2D parent class
        if parent.tbar.switchVizMode.isChecked():
            self.canvas2DList = parent.binlist
            self.setWindowTitle("ComMoDE : Multiple Data Visualization Manager (bin size [m])")
        else:
            self.canvas2DList = parent.trajectories
            self.setWindowTitle("ComMoDE : Multiple Data Visualization Manager (trajectories)")
        
        ## list of checkbox Widgets associated to each list item. Keep indexing of parent list
        self.tcheckArray = []
        ## list of QColor object associated to each list item. Keep indexing of parent list
        self.tcolorArray = []
        ## list of Color change button widgets associated to each list item. Keep indexing of parent list 
        self.tchangecolorArray = []
        
        ## internal layout for displace the item dependent widgets tcheckArray,tcolorArray,tchangecolorArray
        __layout_traj = QGridLayout()

        for i in range(len(self.canvas2DList)) : 
            self.tcheckArray.append(QCheckBox(self.canvas2DList[i].name))
            if self.canvas2DList[i].status :
                self.tcheckArray[i].setCheckState(Qt.Checked)

            __color = QColor()
            __color.setNamedColor(self.canvas2DList[i].color)
            self.tcolorArray.append(__color)

            __coloredbutton = QPushButton()
            __coloredbutton.setIcon(QIcon(iconaddress+"colorpalette.png"))
            __coloredbutton.setIconSize(QSize(20,20))
            __coloredbutton.setToolTip("Change color of " + self.canvas2DList[i].name)
            self.tchangecolorArray.append(__coloredbutton)

            __layout_traj.addWidget(self.tcheckArray[i], i%10, int(2*(i//10)))
            __layout_traj.addWidget(self.tchangecolorArray[i], i%10, int(2*(i//10) + 1))

            self.tcheckArray[i].stateChanged.connect(lambda state = 1, index = i: self.checkValidItem(index))
            self.tchangecolorArray[i].clicked.connect(lambda checked = True, index = i: self.changeColor(index))

        #internal instance of QWidget holding the __layout_traj layout
        __trajGridWidget = QWidget()
        #internal instance to make __trajGridWidget contents scrollable
        __trajGridScroll = QScrollArea()
        __trajGridScroll.setWidgetResizable(True)
        __trajGridScroll.setWidget(__trajGridWidget)
        __trajGridWidget.setLayout(__layout_traj)

        #create layout of accept/reject buttons
        ## Accept button widget, to accept modifications
        self.acceptButton = QPushButton("Ok")
        ## Cancel button widget, to refuse modifications
        self.rejectButton = QPushButton("Cancel")

        self.acceptButton.released.connect(self.storeAndAccept)
        self.rejectButton.released.connect(self.reject) 

        ## internal layout for accept/cancel buttons
        __layout_buttons = QHBoxLayout()
        __layout_buttons.addWidget(self.acceptButton)
        __layout_buttons.addWidget(self.rejectButton)
        # internal instance of QWdiget holding the __layout_buttons layout
        __buttonsWidget = QWidget()
        __buttonsWidget.setLayout(__layout_buttons)

        ## internal final_layout to attach to the dialog
        __final_layout = QVBoxLayout()
        __final_layout.addWidget(__trajGridScroll)
        __final_layout.addWidget(__buttonsWidget)
        self.setLayout(__final_layout)

    def checkValidItem(self, index):
        '''!
            Control the total number of items enabled for concurrent visualization do not exceed 10.
            if it does, raise a warning and unvalidate the check of the checkbox marked with index index
            @param index of the item checkbox
        '''
        count = int(0)
        for itemcbox in self.tcheckArray :
            if itemcbox.isChecked() : 
                count = count + 1    

        if count > 10 and self.tcheckArray[index].isChecked() :
            text = "No more than 10 concurrent items are allowed for visualization. Please consider to disable some items first, before enabling further ones"
            QMessageBox.warning(self, "Cannot select item", text)
            self.tcheckArray[index].setCheckState(Qt.Unchecked) 

    def changeColor(self, index):
        '''!
            Change the color associated to a particular item identified by index integer
            @param index trajectory identifier in the list tcolorArray          
        '''
        __newColor = QColorDialog.getColor(self.tcolorArray[index], self)
        hexstring1 = __newColor.name()
        hexstring2 = self.tcolorArray[index].name()
        if hexstring1 != hexstring2 : 
            self.tcolorArray[index] = __newColor

    def storeAndAccept(self):
        '''!
            Execution command, make the modifications done to the dialog effective, and 
            return feedback to structures of the parent PlotCanvas2D class.
            In particular feedback info are: 
            - the trajectory/bin size item must be visualized or not
            - the color to associate to the data
        '''
        for i in range(len(self.tcheckArray)) :
            self.canvas2DList[i].status = str(self.tcheckArray[i].isChecked()) == 'True'
            self.canvas2DList[i].color = self.tcolorArray[i].name()

        self.accept()